/*
**  CXSC is a C++ library for eXtended Scientific Computing (V 2.0_beta3)
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2003 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany
**
**  This program 'Approx-Error' is free software; you can redistribute it 
**  and/or modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
****************************************************************************
**  The program     'Approx-Error'     calculates an upper bound of the   **
**  absolute and relative approximation error by rational approximation.  **
**  f(x) is approximated by the fraction      P_M(x-x1)/Q_N(x-x1)         **
**  Kraemer|Blomquist, 17.01.05;                                          **
****************************************************************************
*/

#include <iostream>      // For cout, cerr, ...
#include "litaylor.hpp"  // For l_itaylor type
#include <imath.hpp>     // For power(...)

using namespace std;
using namespace cxsc;
using namespace taylor;

const int Z1 = 1;   // Number of subintervals for the remainder term
const int Z2 = 50;  // Each subinterval will again be subdivided by
                    // Z2 new subintervals. The expansion point for
		    // all these Z2 subintervalls is x0.

// *************************************************************************
// *****   The function f(x) to be approximated must be defined here:   ****
// *************************************************************************
l_itaylor STD(const l_itaylor& x)
{   // Definition of the function to be approximated 
    // by a rational function
    l_itaylor t,t1,x2;
    l_interval c = 4*li_pi4();  // c = pi;
    l_interval a14 = l_interval(16384);  // a14 = 2^(14)
    l_interval ak;

    for (int k=1;k<=14;k++) a14 = a14 / (2*k+1);
    // a14 = 2^(14) / (1*3*5* ... *29);
    x2 = sqr(x);  // x2 is a VARIABLE of type l_itaylor
    t1 = 2*exp(-x2)/sqrt(c);
    t = const_l_itaylor(get_order(x),a14); // CONST of type l_itaylor!
    ak = a14; // Start value of the recursion formula
    for (int k=13;k>=0;k--) // Horner's scheme
    {
	ak = ak*(2*k+3)/2;  // Recursion formula for ak
	t = t*x2 + ak;
    }
    return t*t1;
}
// *************************************************************************

l_itaylor TAYLOR_STD(const l_interval& x0, int ord)
// Calculating the staggered Taylor coefficients of the function defined
// in STD(...) for the expansion interval x0 up to order ord.
// Diam(x0) should be as small as possible to avoid overestimations!
{
    l_itaylor arg = var_l_itaylor(ord, x0);
    return STD(arg);
}

interval POLY_HORN(const interval& arg, const ivector& coef)
// POLY_HORN evaluates a polynomial of degree Ub(coef) for the interval
// argement arg. The polynomial coefficients are given by the ivector coef
// with components of type interval and indices from 0,1,... up to Ub(coef) 
{
    interval t;
    t = coef[Ub(coef)]; // polynom. coefficient with maximal index Ub(coef)
    for (int k=Ub(coef)-1; k>=0; k--) // Horner's scheme
	t = t*arg + coef[k];
    return t;
}

void REST_TEIL(const interval& x, int K, real& RN, real& fx)
// Output values:
// RN: Upper bound of all absolute values of the (K+1)-th Taylor
//     coefficient of STD(...) for all arguments of the subinterval x.
// fx: Lower bound of all absolute function values of STD(...) for all 
//     arguments of the subinterval x.
// Input values:
// x:  A subinterval of type interval;
// K:  The degree of Taylor expansion of STD(...).
{
    l_interval x_li;
    l_itaylor t;
    interval u;

    x_li = x;
    t = TAYLOR_STD(x_li,K+1);

    u = get_j_coef(t, K+1); // u: Inclusion of the (K+1)-th Taylor coeffic.  
    RN = Sup(abs(u));

    u = get_j_derive(t,0);
    fx = Inf(abs(u));
} // REST_TEIL

void TAU(const interval& x, int K, int Nr, real& tau_abs, real& tau_rel)
// Output:
// tau_abs: Upper bound of the absolute remainder values over interval x
//          for calculating the absolute error.
// tau_rel: Upper bound of the absolute remainder values over interval x
//          for calculating the relative error.
// Input:
// x :      The considered interval; its mid-point is the 
//          Taylor expansion point x0.
// K :      The degree of Taylor expansion
// Nr:      The number of subintervals to avoide overestimations by 
//          calculating the Taylor-coefficients and the Lower bound
//          of the absolute values of f(x); Nr = 1,2,3, ...
{
    real delta,c1,c2;
    interval xt; // xt is one of the Nr subintervals

    tau_abs = 0;  tau_rel = MaxReal;
    delta = diam(x) / Nr;
    Sup(xt) = Inf(x);

    for (int k=1; k<=Nr; k++)
    {
	Inf(xt) = Sup(xt);

	Sup(xt) = (k==Nr) ? Sup(x) : Inf(xt) + delta;

	REST_TEIL(xt,K,c1,c2);
	if (c1>tau_abs) tau_abs = c1;
	if (c2<tau_rel) tau_rel = c2;
    }
    c1 = divu(diam(x),2);
    xt = power(interval(c1),K+1);

    tau_abs = mulu(tau_abs,Sup(xt));
    tau_rel = (tau_rel==0) ? MaxReal : divu(tau_abs,tau_rel);

} // TAU

void ERROR_TEIL(const interval& x,  // Considered subinterval
                const interval& x0, // Expansion interval with diam=0 for
                                    // the Taylor expansion of f(x);
                                    // (x-x0) is the polynomial argument.
		const ivector& z,   // Polynomial coefficients of the 
                                    // numerator. These coefficients are
                                    // calculated in staggered Arithmetic
                                    // with regard to x0.
		const ivector& b,   // Polynomial coefficients of the 
                                    // denominator. These coefficients are
                                    // calculated in staggered Arithmetic
                                    // with regard to x0.
		const ivector& sN,  // K+1 coefficients of the Taylor
                                    // polynomial for the approximation of
                                    // f(x) with regard to x0.
                const real& tau_abs,// Upper bound for the absolute values
                                    // of the remainder term with regard to x.
// Output:
		real& abs_,         // Upper bound of the absolute 
                                    // approximation error of the first
                                    // error term.
		real& rel_)         // Upper bound of the relative 
                                    // approximation error of the first
                                    // error term.                             
{
    interval Zae,Ne1,Ne2,c;
    c = interval(-tau_abs,+tau_abs);
    Zae = POLY_HORN(x-x0,z);
    Ne1 = POLY_HORN(x-x0,b);
    Ne2 = Ne1 * (POLY_HORN(x-x0,sN)+c);
    if(0 <= Ne1) {abs_ = MaxReal; rel_ = MaxReal;}
    else {
	abs_ = Sup(abs(Zae/Ne1));
	rel_ = (0 <= Ne2) ? MaxReal : Sup(abs(Zae/Ne2));
    }
} // ERROR_TEIL

void SUMMAND1(
//  Input:
    const interval& x,   // considered subinterval
    const interval& x0,  // Expansion interval with diam(x0)=0 for the
                         // Taylor expansion of f(x)
                         // (x-x0) is the polynomial argument.
    const ivector& z,    // Polynomial coefficients of the numerator.
                         // These coefficients are calculated in staggered 
                         // Arithmetic with regard to x0.
    const ivector& b,    // Polynomial coefficients of the denominator.
                         // These coefficients are calculated in staggered 
                         // Arithmetic with regard to x0.
    const ivector& sN,   // K+1 coefficients of the Taylor polynomial for 
                         // the approximation of f(x) with regard to x0.
          int Nr,        // Number of subintervals in which x will be
                         // subdivided.
    const real& tau_abs, // Upper bound of the absolute values of the 
                         // remainder term with regard to the subinterval x
// Output:
          real& abs_e,   // Upper bound of the absolute approximation error
                         // with regard to the first error term. 
          real& rel_e)   // Upper bound of the relative approximation error
                         // with regard to the first error term. 
{
    real delta,c1,c2;
    interval xt;  // one of the Nr subintervals

    abs_e = 0;   rel_e = 0;
    delta = diam(x) / Nr;
    Sup(xt) = Inf(x);
    for (int k=1; k<=Nr; k++) {
	Inf(xt) = Sup(xt);
	Sup(xt) = (k==Nr) ? Sup(x) : Inf(xt) + delta;
	ERROR_TEIL(xt,x0,z,b,sN,tau_abs,c1,c2);
	if (c1>abs_e) abs_e = c1;
	if (c2>rel_e) rel_e = c2;
    } 
} // SUMMAND1

l_itaylor POLY_ZAEHLER(
    const l_itaylor& x,  
    const l_interval& x1, // expansion point x1.
    const l_ivector& ci)  // Staggered inclusions of the numerator 
                          // polynomial coefficients with regard to
                          // the expansion point x1.
{
    l_itaylor t = const_l_itaylor(get_order(x),ci[Ub(ci)]);
    for (int k=Ub(ci)-1; k>=0; k--) {
	t = t * (x-x1) + ci[k];   // Horner's scheme
    }
    return t;
} // POLY_ZAEHLER

l_itaylor TRANS_Z(
    const l_interval x0,  // Evaluation point
          int n,
    const l_interval& x1, // expansion point x1
    const l_ivector& ci)
{
    l_itaylor arg = var_l_itaylor(n,x0);
    return POLY_ZAEHLER(arg,x1,ci);
} // TRANS_Z

l_itaylor POLY_NENNER(
    const l_itaylor& x,  
    const l_interval& x1, // expansion point x1.
    const l_ivector& di)  // Staggered inclusions of the numerator 
                          // polynomial coefficients with regard to
                          // the expansion point x1.
{
    l_itaylor t = const_l_itaylor(get_order(x),di[Ub(di)]);
    for (int k=Ub(di)-1; k>=0; k--) {
	t = t * (x-x1) + di[k];   // Horner's scheme
    }
    return t;
}

l_itaylor TRANS_N(
    const l_interval x0,  // Evaluation point
          int n,
    const l_interval& x1, // expansion point x1
    const l_ivector& di)
{
    l_itaylor arg = var_l_itaylor(n,x0);
    return POLY_NENNER(arg,x1,di);
} // TRANS_N
 
void Approx_Fehler(
// This function calculates the abs. and rel. error bounds for one
// of the Z subintervals of ab.
// Input:
    const interval& intv, // The considered subinterval
          int K,          // Degree of Taylor polynomial (with mid-point
                          // of intv) for the approximation of f(x)
          int P,          // Pecision parameter for litaylor
          int M,          // Degree of the numerator polynomial of the 
                          // rational approximation
          int N,          // Degree of the denominator polynomial of the 
                          // rational approximation
    const l_ivector& ci,  // Staggered inclusions of the numerator 
                          // polynomial coefficients with regard to
                          // the expansion point x1.
    const l_ivector& di,  // Staggered inclusions of the denominator 
                          // polynomial coefficients with regard to
                          // the expansion point x1.
    const l_interval& x1, // The coefficients ci,di (for example 
                          // calculated with Mathematica) are related to 
                          // the expansion point x1.
// Output:
          real& abs_err,  // Upper bound of the absolute approx. error 
          real& rel_err)  // Upper bound of the absolute approx. error  
{
    stagprec = P;
    real tau_abs,tau_rel;
    int Bh;
    l_interval x0  = l_interval(mid(intv));
    l_ivector  a0  = l_ivector(0,M);
    l_ivector  b0  = l_ivector(0,N);
    l_ivector  s0  = l_ivector(0,K+1);
    l_ivector  z0  = l_ivector(0,K+N);
    l_itaylor  tz;
    l_itaylor  tn;
    l_itaylor  t0;
    ivector    sN  = ivector(0,K);
    ivector    z   = ivector(0,K+N);
    ivector    b   = ivector(0,N);
    interval x_Null;
    if (x0==x1) { a0 = ci;  b0 = di;}
    else {
	tz = TRANS_Z(x0,Ub(a0),x1,ci);
	for (int j=0; j<=Ub(a0); j++) 
	    a0[j] = get_j_coef(tz,j);
	tn = TRANS_N(x0,Ub(b0),x1,di);
	for (int j=0; j<=Ub(b0); j++) 
	    b0[j] = get_j_coef(tn,j);
    }
    t0 = TAYLOR_STD(x0,Ub(s0));
    for (int j=0; j<=Ub(s0); j++)
	s0[j] = get_j_coef(t0,j); 
    for (int j=0; j<=Ub(sN); j++) 
	sN[j] = s0[j];
// Now the polynomial coefficients z[j] of the numerator for calculating
// the absolute or relative approximation error are computed:

// Polynomial multiplication:

    for (int k=0; k<=Ub(b0); k++) {
	z0[k] = 0;
	for (int j=0; j<=k; j++)
	    z0[k] = z0[k] + b0[k-j]*s0[j];
    }
    for (int k=Ub(b0)+1; k<=Ub(s0)-1; k++) {
	z0[k] = 0;
	for (int j=0; j<=Ub(b0); j++) 
	    z0[k] = z0[k] + b0[Ub(b0)-j]*s0[k-Ub(b0)+j];
    }
    Bh = Ub(s0)-1+Ub(b0);
    for (int k=Ub(s0); k<=Bh; k++) {
	z0[k] = 0;
	for (int j=0; j<=Bh-k; j++)
	    z0[k] = z0[k] + b0[Ub(b0)-j]*s0[k-Ub(b0)+j];
    }  // End of polynomial multiplication

// Polynomial subtraction:
    for (int k=0; k<=Ub(a0); k++)
	z0[k] = z0[k] - a0[k];
    for (int k=0; k<=Ub(z); k++)
	z[k] = z0[k];

    TAU(intv,K,Z1,tau_abs,tau_rel);

    for (int k=0; k<=Ub(b); k++)
	b[k] = b0[k];
    x_Null = x0;

    SUMMAND1(intv,x_Null,z,b,sN,Z2,tau_abs,abs_err,rel_err);

    if (abs_err==MaxReal || tau_abs==MaxReal) abs_err = MaxReal;
    else abs_err = addu(abs_err,tau_abs);

    if (rel_err==MaxReal || tau_rel==MaxReal) rel_err = MaxReal;
    else rel_err = addu(rel_err,tau_rel);
} // Approx_Fehler

void Haupt(const interval& ab,  // ab: Approximation interval 
           int Z,               // Z:  Number of subintervals
           int K,               // K:  Degree of Taylor polynomial for f(x)
           int M,               // M:  Degree of numerator polynomial p
           int N,               // N:  Degree of denominator polynomial q
           const real& x1,      // x1: Expansion point of rat. approximation
           const rvector& p,    // Polynomial coefficients of the numerator
	   const rvector& q,    // Polynomial coefficients of the denominator
	         real& abs_err, // abs_err: Absolute error bound
	         real& rel_err) // rel_err: Relative error bound
{
    int P;        // P: actual stagprec value
    interval xt;  // xt: One of the Z subintervals
    real delta,   // Step size
	 ca,cr;   // Absolute and relative error bound
    l_interval x1l = l_interval(x1);
    l_ivector ci = l_ivector(0,M);
    l_ivector di = l_ivector(0,N);

    P = stagprec;

    for (int k=0; k<=M; k++)
	ci[k] = l_interval(p[k]);
    for (int k=0; k<=N; k++)
	di[k] = l_interval(q[k]);

    abs_err = 0;  rel_err = 0;
    delta = diam(ab) / Z;
    Sup(xt) = Inf(ab); 
    for (int k=1; k<=Z; k++)
    {
	Inf(xt) = Sup(xt);
	if (k==Z) Sup(xt) = Sup(ab);
	else Sup(xt) = Inf(xt) + delta;

	Approx_Fehler(xt,K,P,M,N,ci,di,x1l,ca,cr);

	if (ca>abs_err) abs_err = ca;
	if (cr>rel_err) rel_err = cr;
    }
} // Haupt()

int main()
{
    string str = "***********************************";
    string strleer = "    ";
    char Zeichen = char(247);
    int M,   // Polynomial degree of p_M(x-x1)
        N,   // Polynomial degree of q_N(x-x1)
        Z,   // Number of subintervals in the approximation interval ab
        K,   // Degree of Taylor expansion of f(x)
	max, // max = max{M,N}
        P;   // stagprec value
    double dbl; // Will be only used by function fscanf(...)
    real x1; // Expansion point for the given rational approximation
             // f(x) approx p_M(x-x1) / q_N(x-x1);
             // p_M(x-x1) := p[0] + ... + p[M]*(x-x1)^M;
             // q_N(x-x1) := q[0] + ... + q[N]*(x-x1)^N;
             // The polynomial degrees M,N, the expansion point x1 and 
             // the coefficients p[i],q[j] are stored in this sequence
             // in decimal form in the file APPR-Dat.txt which must be
             // located in the same directory where this program will be
             // started. 
    real abs_err, // Absolute error bound
	 rel_err; // Relative error bound

    interval ab;  // ab: Approximation interval
 
    FILE *fp;

    if ( (fp = fopen("APPR-Dat.txt", "rb")) == 0)
    {
	cerr << "Error by opening file" << endl;
	exit(1);
    }
    cout << endl;
    cout << strleer+str+str << endl;
    cout << strleer+"*   "+
         "                                                                 *" 
         << endl;
    cout << strleer+"*   "+
         "This program calculates guaranted absolute and relative error    *" 
         << endl;
    cout << strleer+"*   "+
         "bounds for the rational approximation                            *"
         << endl;
    cout << strleer+"*   "+
         "                 f(x) "+char(247)+" P_M(x-x1) / Q_N(x-x1).       *"
         << endl;
    cout << strleer+"*   "+
         "f(x), defined by standard functions, must be implemented in      *"
         << endl;
    cout << strleer+"*   "+
         "the internal function STD(...).                                  *"
         << endl;
    cout << strleer+"*   "+
         "The polynomial degrees M>0, N>=0 of the polynomials P_M(x-x1),   *"
         << endl;
    cout << strleer+"*   "+
         "Q_N(x-x1), the expansion point x1, which must be representable   *"
         << endl;
    cout << strleer+"*   "+
         "in the IEEE-double format, and the polynomial coefficients       *"
         << endl;
    cout << strleer+"*   "+
         "p[j],q[j], j=0,1,...,M|N, must be written in this sequence in    *"
         << endl;
    cout << strleer+"*   "+
         "decimal form to the text-file 'APPR-Dat.txt', which must be      *"
         << endl;
    cout << strleer+"*   "+
         "stored in the same directory, where this program is running.     *"
         << endl;
    cout << strleer+"*   "+
         "By reading the polynomial coefficients with at least 18 decimal  *"
         << endl;
    cout << strleer+"*   "+
         "digits from the text file, these coefficients are rounded to     *"
         << endl;
    cout << strleer+"*   "+
         "the next IEEE-numbers p[j],q[j], which will then be treated as   *"
         << endl;
    cout << strleer+"*   "+
         "the exact polynomial coefficients!    The text file content is   *"
         << endl;
    cout << strleer+"*   "+
         "listed below; to continue press any key!                         *"
         << endl;
    cout << strleer+"*   "+
         "                                                                 *" 
         << endl;
    cout << strleer+str+str << endl;  

    fscanf(fp," %d %d %lf ", &M,&N,&dbl);
    cout << strleer+"*   "+"M = " << M << endl;
    cout << strleer+"*   "+"N = " << N << endl;

    max = M;
    if (N>max) max = N;

    rvector p(0,M);  // Coefficients of the polynomial p_M(x-x1) of order M
    rvector q(0,N);  // Coefficients of the polynomial q_N(x-x1) of order N

    x1 = dbl;  // x1: Expansion point of the given rational approximation
    cout << SetPrecision(18,18) << Scientific
         << strleer+"*   "+"Expansion point x1 = " << x1 << endl; 

    for (int k=0; k<=M; k++)
    {
	fscanf(fp," %lf ", &dbl);  // Rounding the next decimal string in
	p[k] = dbl;                // the file APPR-Dat.txt to the nearest
                                   // IEEE-real number p[k];
	cout << strleer+"*   "+"p[" << k << "] = " << p[k] << endl;
    }
    for (int k=0; k<=N; k++)
    {
	fscanf(fp," %lf ", &dbl);  // Rounding the next decimal string in
	q[k] = dbl;                // the file APPR-Dat.txt to the nearest
                                   // IEEE-real number q[k];
	cout << strleer+"*   "+"q[" << k << "] = " << q[k] << endl;
    }

    fclose(fp); // closing the Text-File APPR-Dat.txt in the current directory
    do {
    cout << "    ***************************************** " << endl;
    cout << "    *   Approximation intervall [a,b] = ?   * " << endl;
    cout << "    ***************************************** " << endl;
    cin >> ab;
    } while (Inf(ab)>Sup(ab));
    cout << "    ************************************************** " << endl;
    cout << "    *   Z: Number of subintervals in [a,b];  Z = ?   * " << endl;
    cout << "    ************************************************** " << endl;
    cin >> Z;
    if (Z<1) Z = 1;
    cout << "    ********************************************** " << endl;
    cout << "    *   f(x): Taylor expansion of degree K = ?   * " << endl;
    cout << "    ********************************************** " << endl;
    cin >> K;
    if (K<=max) K = max + 1;
    cout << "    Used value K = " << K << endl;
    cout << "    ************************************** " << endl;
    cout << "    *   Multiprecision parameter P = ?   * " << endl;
    cout << "    ************************************** " << endl;
    cin >> P;
    if (P<2) P = 2;
    stagprec = P;
    cout << "    Used stagprec value P = " << P << endl << endl;
    cout << "    Please wait some moments ...." << endl << endl;
    Haupt(ab,Z,K,M,N,x1,p,q,abs_err,rel_err);
    cout << SetPrecision(10,10);
    cout << RndUp << "    *   Error bound for the absolute error:  " 
         << abs_err << endl;
    cout << RndUp << "    *   Error bound for the relative error:  " 
         << rel_err << endl << endl;
} // main

