/*
**  CXSC is a C++ library for eXtended Scientific Computing
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2001 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "abs_relh.hpp"
#include "hilfe.hpp"
#include <interval.hpp>
#include <imath.hpp>

using namespace std;
namespace cxsc {

static const real eps_h = comp(0.5,-51);   // eps_h = 2^(-52) 
                                           // for only high accuracy
static const real q_min = -6231120794008786.0 / 8796093022208.0;
    // q_mine: smallest machine number > -1022*ln(2) = -708.396418...
static const real eps_exp = 
    4782516237658342.0 / 20282409603651670423947251286016.0; 
    // relative error bound of the exponential function; 2.357962555295842E-16
static const real eps_ln = 
    5962622775281519.0 / 20282409603651670423947251286016.0; 
    // relative error bound of the logarithm function; 2.9398E-016
static const real eps_lnp1 = 
  5087233976787912.0 / 20282409603651670423947251286016.0; 
  // relative error bound of the function ln(1+x); 2.5082E-016;
static const real eps_acosh = 
  8204234684677101.0 / 5070602400912917605986812821504.0; 
  // relative error bound of the function acosh(x); 1.6180E-015;
static const interval MR_l = interval(succ(-MinReal),MinReal); 


int e_IEEE(const real& x)
// Liefert den Exponenten e(x) in den Darstellungen:
// Normalisierter Bereich:   x = (-1)^s * 1.m * 2^e(x);
// Denormalisierten Bereich: x = (-1)^s * 0.m * 2^(-1022)
// Null:                     0 = (-1)^s * 0.0...0 * 2^(-1022)
{
    int ex = expo(x);
    ex = ex<-1021 ? -1022 : ex--;
    return ex;
} // e_IEEE

int z_trail(const real& x)
// Liefert die Anzahl aufeinanderfolgender Null-Bits am Ende der Mantisse
// von x.       x=0 --> 53;  x=1 --> 52;  x=1.5 --> 51;  x=pred(2) --> 0;
// Vorsicht: Funktioniert zunchst nur auf INTEL-Systemen!!
{
    real a=x, *zr = &a;
    int p=0,p1;
    if (sign(x)==0) p = 53;
    else {
    unsigned int *z1;
    z1 = (unsigned int*) zr;  // der Anfang von a=x wird als unsigned int
                              // interpretiert!
    p = uint_trail(*z1);
    if (p==32)  // Die ersten 32 Mantissen-Bits sind Null
    {
	z1++;  // Zeiger auf den Anfang der 4 letzten Byte von x setzen
	p1 = uint_trail(*z1);
	if (p1>20) p1 = 20;  
	p += p1;
    }
    }
    return p; 
} // z_trail

int z_lead(const real& x)
// Liefert die Anzahl aufeinanderfolgender Null-Bits am Anfang der Mantisse
// von x.    x=0 --> 0;  x=MinReal --> 0;  x=minreal --> 52;
{
    int p=0;  // p=0, falls x im normalisierten Bereich liegt.
    if (sign(x)==0) p = 0;
    else 
    {
	int ex = expo(x);
	if (ex<-1021) // x<>0 liegt jetzt im denormalisierten Bereich
	{
	    p = -1021 - ex;
	}
    }
    return p; // Rckgabe der Anzahl aufeinanderfolgender Null-Bits
              // am Anfang der Mantisse von x.
} // z_lead

int e_IS(const interval& x)
// Realisiert die Funktion eIS(X) von Seite 18 aus Preprint Nr. 98/5
{
    int ex;
    if (diam(x)>0)
    {
	interval ax = abs(x);
	ex = e_IEEE( Inf(ax) );
    } else  // x ist Punktintervall:
    {
	real r = Inf(x);
	ex = e_IEEE(r) + z_trail(r);
    }
    return ex;
} // e_IS

static real abs_sqrt_(const real& a, const real& dela)
// Wird bentigt von abs_sqrt_abs() zur Maximumberechnung
{
    real r,r1;
    interval x;
    if (a<=dela)
    {
	cerr << "abs_sqrt_abs: absolte error doesn't exist!" << endl;
	exit(1);
    }
    r = subd(a,dela); // Positive Differenz im Nenner mglichst klein whlen!
    x = sqrt( interval(r) );
    r = Inf(x);
    times2pown(r,1);   // Schnelle exakte Multiplikation mit 2=2^1
    r1 = divu(dela,r); // r1 = dela/(2*sqrt(a-dela))

    x = sqrt( interval(a) );
    r = addu(r1,Sup(x));  // r = r1 + sqrt(a);
    r = mulu(eps_h,r);    // r = eps_h*( r1 + sqrt(a) )
    r = addu(r1,r); 
    return r;  // Rckgabe von r1 + eps_h*( r1 + sqrt(a))
}

    static real ForMaxS4(const real& epsa, const real& epsb,
                         const real& a, const real& b)
    // (|a|*epsa + |b|*epsb)/|a+b|
    {
	real d1=0, d2;
	if (sign(epsa)!=0 || sign(epsb)!=0)
	{
	    d1 = mulu(abs(a),epsa), d2 = mulu(abs(b),epsb);
	    d1 = addu(d1,d2);  // Numerator ready
	    interval ai = interval(a), bi = interval(b);
	    ai = abs(ai+bi);   // '+' because of sum
	    d2 = Inf(ai);      // Denominator ready
	    if (d2<=0)
	    {
	        cerr << "Realtive error doesn't exist!" << endl;
	        exit(1);
	    } 
	    d1 = divu(d1,d2);
	}
	return d1;
    }

    static real ForMaxD4(const real& epsa, const real& epsb,
                         const real& a, const real& b)
    // (|a|*epsa + |b|*epsb)/|a-b|
    {
	real d1=0, d2;
	if (sign(epsa)!=0 || sign(epsb)!=0)
	{
	    d1 = mulu(abs(a),epsa), d2 = mulu(abs(b),epsb);
	    d1 = addu(d1,d2);  // Numerator ready
	    interval ai = interval(a), bi = interval(b);
	    ai = abs(ai-bi);   // '-' because of difference
	    d2 = Inf(ai);      // Denominator ready
	    if (d2<=0)
	    {
	        cerr << "Realtive error doesn't exist!" << endl;
	        exit(1);
	    } 
	    d1 = divu(d1,d2);
	}
	return d1;
    }

    static real ForMaxS2(const real& dela,  const real& epsb,
                         const interval& A, const real& b)
    // (dela + |b|*epsb)/Inf(|A+b|)
    {
	real d1=0, d2;
	if (sign(dela)!=0 || sign(epsb)!=0)
	{
	    d2 = mulu(abs(b),epsb);
	    d1 = addu(dela,d2);  // Numerator ready
	    interval bi = interval(b);
	    bi = abs(A+bi);     // '+' because of sum
	    d2 = Inf(bi);       // Denominator ready
	    if (d2<=0)
	    {
	        cerr << "Realtive error doesn't exist!" << endl;
	        exit(1);
	    } 
	    d1 = divu(d1,d2);
	}
	return d1;
    }

    static real ForMaxD2(const real& dela,  const real& epsb,
                         const interval& A, const real& b)
    // (dela + |b|*epsb)/Inf(|A-b|)
    {
	real d1=0, d2;
	if (sign(dela)!=0 || sign(epsb)!=0)
	{
	    d2 = mulu(abs(b),epsb);
	    d1 = addu(dela,d2);  // Numerator ready
	    interval bi = interval(b);
	    bi = abs(A-bi);     // '-' because of difference
	    d2 = Inf(bi);       // Denominator ready
	    if (d2<=0)
	    {
	        cerr << "Realtive error doesn't exist!" << endl;
	        exit(1);
	    } 
	    d1 = divu(d1,d2);
	}
	return d1;
    }

    static real ForMaxS3(const real& epsa,  const real& delb,
                         const interval& B, const real& a)
    // (|a|*epsa + delb)/Inf(|a+B|)
    {
	real d1=0, d2;
	if (sign(epsa)!=0 || sign(delb)!=0)
	{
	    d2 = mulu(abs(a),epsa);
	    d1 = addu(delb,d2);  // Numerator ready
	    interval ai = interval(a);
	    ai = abs(ai+B);     // '+' because of sum
	    d2 = Inf(ai);       // Denominator ready
	    if (d2<=0)
	    {
	        cerr << "Realtive error doesn't exist!" << endl;
	        exit(1);
	    } 
	    d1 = divu(d1,d2);
	}
	return d1;
    }

    static real ForMaxD3(const real& epsa,  const real& delb,
                         const interval& B, const real& a)
    // (|a|*epsa + delb)/Inf(|a-B|)
    {
	real d1=0, d2;
	if (sign(epsa)!=0 || sign(delb)!=0)
	{
	    d2 = mulu(abs(a),epsa);
	    d1 = addu(delb,d2);  // Numerator ready
	    interval ai = interval(a);
	    ai = abs(ai-B);     // '-' because of difference
	    d2 = Inf(ai);       // Denominator ready
	    if (d2<=0)
	    {
	        cerr << "Realtive error doesn't exist!" << endl;
	        exit(1);
	    } 
	    d1 = divu(d1,d2);
	}
	return d1;
    }

    static real mul_exakt(const interval& A, const interval& B,
                          const interval& aP)
    // A,B intervals of the disturbed machine values; aP = abs(A*B);
    // Returning d1=0, if multiplication exact, otherwise:  d1=1;
    {
	const real c=0.5;
	const int p=-1021;
	real d1,SA,SB;
	bool blA,blB,blU,blNull,blAPt,blBPt;

	SA = Sup( abs(A) );
	SB = Sup( abs(B) );
	blNull = (A==0 || B==0);
	blAPt = (diam(A)==0);  blBPt = (diam(B)==0);
	blA = (blAPt && mant(SA)==c);
	blB = (blBPt && mant(SB)==c);
	blU = (Inf(aP)<MinReal);
	if (blA && blU)
	    blA = (expo(SA) >= p - e_IS(B));
	if (blB && blU)
	    blB = (expo(SB) >= p - e_IS(A));
	blU = (blAPt && blBPt);
	if (blU)
	    blU = (z_lead(SA)+z_trail(SA)+z_lead(SB)+z_trail(SB) >= 53);
	d1 = (blA || blB || blU || blNull) ? 0 : 1;
	return d1;
    } // mul_exakt


    static real div_exakt(const interval& A, const interval& B, 
                          const interval& aP)
    // A,B: intervalls of the disturbed machine values; aP = abs(A/B);
    // Returning d1=0, if multiplication exact, otherwise: d1=1;
    {
	real d1;
	bool blB,blU,blNull;
	blNull = (A==0);
	d1 = Sup( abs(B) );
	blB = (diam(B)==0 && mant(d1)==0.5);
	blU = (Inf(aP)<MinReal);
	if (blB && blU)
	    blB = (expo(d1) <= e_IS(A)+1023);
	d1 = (blB || blNull) ? 0 : 1;

	return d1;
    } // div_exakt

static interval exp_u(const real& a, const real& epsa)
// calculating an inclusion of u(a); used in function exp_v()
{
    interval z = interval(epsa);
    z = abs(a) * z;
    return z * exp(z);
}

static interval exp_v(const real& a, const real& epsa)
// calculating an inclusion of u(a)+eps(exp)*(u(a)+1);
// used in function rel_exp_rel()
{
    interval t;
    t = exp_u(a,epsa);
    return t + eps_exp*(t+1);
}

static real ln_abs4Max(const real& a, const real& dela)
// For calculating the maximum; used by abs_ln_abs()
{
    interval z;
    real r,r1;
    r = subd(a,dela);
    r = divu(dela,r);
    z = abs( ln(interval(a)) );
    r1 = addu(Sup(z),r);
    r1 = mulu(r1,eps_ln);
    r1 = addu(r1,r);
    return r1;
}

static real lnp1_abs4Max(const real& a, const real& dela)
    // For calculating the maximum; used by abs_lnp1_abs()
{
    interval z;
    real r,r1;
    r1 = addd(1,a);
    r = subd(r1,dela);
    r = divu(dela,r);
    z = abs( lnp1(interval(a)) );
    r1 = addu(Sup(z),r);
    r1 = mulu(r1,eps_lnp1);
    r1 = addu(r1,r);
    return r1;
}

static real abslnp1_4Max( const real& a, const real& epsa, 
                                         const real& one_meps, 
                                         const real& one_peps )
    // For calculating the function h(a); used by abs_lnp1_rel()
{
    real za = mulu(abs(a),epsa),ne;
    ne = (sign(a)>=0) ? muld(a,one_meps) : muld(a,one_peps);
    ne = addd(1,ne);
    za = divu(za,ne); // za: Delta_dat,f
    interval z = abs( lnp1(interval(a)) );
    ne = Sup(z); // ne: upper bound of |ln(1+a)|
    ne = addu(ne,za);
    ne = mulu(eps_lnp1,ne);
    ne = addu(za,ne);
    return ne;
}

static real Delta_4Max( const real& a, const real& epsa, 
                                       const real& one_meps, 
                                       const real& one_peps )
// For calculating the function Delta_dat,f(a); used by abs_lnp1_rel()
{
    real za = mulu(abs(a),epsa),ne;
    ne = (sign(a)>=0) ? muld(a,one_meps) : muld(a,one_peps);
    ne = addd(1,ne);
    za = divu(za,ne); // za: Delta_dat,f
    return za;
}

static real Help_lnp1_1(const real& a, const real& dela)
// For calculating the function eps_dat,f(a) for a>0 or a<0;
// used by function rel_lnp1_abs()
{
    real r,r1;
    interval z;
    r = addd(1,a);  r = subd(r,dela);
    z = abs( lnp1(interval(a)) );
    r1 = Inf(z);
    r = muld(r,r1); // r: denominator must be positive!
    if (sign(r)<=0)
    {
	cerr << "rel_lnp1_abs(): Relative error doesn't exist!" << endl;
	exit(1);
    } 
    r = divu(dela,r);
    return r;
}

// ---------------------------------------------------------------------------
// ------------------- Calculating the absolute errors -----------------------
// ---------------------------------------------------------------------------
// -------------------------- Delta(a), Delta(b) -----------------------------
// ---------------------------------------------------------------------------

void abs_addh1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A+B and an upper bound del of the absolute error by 
               // calculating (a+b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    int e,e1,e2;
    bool bl1,bl2,bl3;
    real d=addu(dela,delb),S,S1,I1,d1; // d = reproduced data error
    R=A+B;
    A1 = A + interval(-dela,dela);
    B1 = B + interval(-delb,delb);
    bl1 = (A1==0 || B1==0);
    bl2 = (!(0<=B1) && A1/B1<=interval(-2,-0.5));
    aR = abs(A1+B1);  S1 = Sup(aR);  I1 = Inf(aR);
    e = e_IEEE(S1);
    e1 = e_IS(A1);  e2 = e_IS(B1); 
    if(e2<e1) e1=e2; // Minimumbildung
    bl3 = (e<=e1);
    if (bl1 || bl2 || bl3) del=d;
    else{
    aR = abs(R);  S = Sup(aR);
    if (I1>=MinReal) // All sums only in the normalized range
    {
	d1 = addu(d,S);
	d1 = mulu(eps_h,d1);  // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (S1<MinReal) // All sums only in the denormalized range U
    {
	del = addu(d,MinReal);
    } else // Sums in normalized and denormalized range --> maximum!:
    {
	del = addu(d,MinReal); // del: absolute error for the underflow range
        d1 = addu(d,S);
	d1 = mulu(eps_h,d1);  // d1:  absolute rounding error
	d1 = addu(d,d1);      // d1:  absolute error for the normalized range
	if (d1>del) del = d1; // calculating the maximum
    } }
} // abs_addh1
     
void abs_subh1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A-B and an upper bound del of the absolute error by 
               // calculating (a-b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    int e,e1,e2;
    bool bl1,bl2,bl3;
    real d=addu(dela,delb),S,S1,I1,d1; // d = reproduced data error
    R=A-B;
    A1 = A + interval(-dela,dela);
    B1 = B + interval(-delb,delb);
    bl1 = (A1==0 || B1==0);
    bl2 = (!(0<=B1) && A1/B1<=interval(0.5,2));
    aR = abs(A1-B1);  S1 = Sup(aR);  I1 = Inf(aR);
    e = e_IEEE(S1);
    e1 = e_IS(A1);  e2 = e_IS(B1); 
    if(e2<e1) e1=e2; // Minimumbildung
    bl3 = (e<=e1);
    if (bl1 || bl2 || bl3) del=d;
    else{
    aR = abs(R);   S = Sup(aR);
    if (I1>=MinReal) // All differences only in the normalized range
    {
	d1 = addu(d,S);
	d1 = mulu(eps_h,d1);  // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (S1<MinReal) // All differences only in the denormalized range U
    {
	del = addu(d,MinReal);
    } else // differences in normalized and denormalized range --> maximum!:
    {
	del = addu(d,MinReal); // del: absolute error for the underflow range
        d1  = addu(d,S);
	d1  = mulu(eps_h,d1);  // d1:  absolute rounding error
	d1  = addu(d,d1);      // d1:  absolute error for the normalized range
	if (d1>del) del = d1;  // calculating the maximum
    } }
} // abs_subh1

void abs_mulh1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A*B and an upper bound del of the absolute error by 
               // calculating (a*b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1;
    real d = mulu(dela,delb);
    real SA=Sup( abs(A) ), SB=Sup( abs(B) );
    real d1 = mulu(SA,delb), d2 = mulu(SB,dela);
    d = addu(d,d1); 
    d = addu(d,d2);  // d: reproduced data error

    A1 = sign(dela)!=0 ? A+interval(-dela,dela) : A;
    B1 = sign(delb)!=0 ? B+interval(-delb,delb) : B;

    R=A*B; 
    interval aR = abs(R);
    real S = Sup(aR);
    aR = abs(A1*B1);  SA = Sup(aR);   SB = Inf(aR);
    
    d1 = mul_exakt(A1,B1,aR); // d1=0, if multiplication exact;

    if (SB>=MinReal) // Products only in the normalized range
    {
	if (d1!=0) // Multiplication not exact:
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_h,d1);
	}
	del = addu(d,d1);
    } else if (SA<MinReal) // Products only in the denormalized range U
    {
	del = d1!=0 ? addu(d,MinReal) : d;
    } else // Products in normalized and denormalized range --> maximum!:
    {
	del = d1!=0 ? addu(d,MinReal) : d;  // del: absolute error for
                                            //      the underflow range
	if (d1!=0) // Multiplication not exact:
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_h,d1); // d1: absolute rounding error
	}
	d2 = addu(d,d1); // d2: absolute error for the normalized range
	if (d2>del) del = d2; // calculating the maximum
    }
} // abs_mulh1

void abs_divh1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A/B and an upper bound del of the absolute error by 
               // calculating (a/b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aQ;
    bool bla = ( sign(dela)!=0 );;
    A1 = bla ? A+interval(-dela,dela) : A;
    B1 = sign(delb)!=0 ? B+interval(-delb,delb) : B;
    if (0<=B1)
    {   
	cerr << "abs_divh1(...): Absolute error doesn't exist." << endl;
	exit(1);
    }
    real S=Inf( abs(B) ), d=0,d1;  // S > 0;
    if (bla || sign(delb)!=0) 
    {
	d1=Sup( abs(A) );
        d1 = mulu(d1,delb);
        d1 = divu(d1,S);
        d1 = addu(d1,dela);  // numerator ready
        S = subd(S,delb);    // denominator ready
        if (S<=0) 
        {   
	    cerr << "abs_divh1(...): Absolute error doesn't exist." << endl;
	    exit(1);
        }
        d = divu(d1,S);  // d: reproduced data error
    }
    R=A/B; // 0 not in B !!
    aQ = abs(R);   S = Sup(aQ);

    aQ = abs(A1/B1); 
    d1 = div_exakt(A1,B1,aQ);  bla = ( sign(d1)!=0 );
    if (Inf(aQ)>=MinReal) // quotients only in the normalized range
    {
	if (bla) // multiplication not exact
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_h,d1);
	}   // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (Sup(aQ)<MinReal) // Products only in the denormalized range U
    {
	del = bla ? addu(d,MinReal) : d;
    } else // Products in normalized and denormalized range --> maximum!:
    {
	del = bla ? addu(d,MinReal) : d; // del: absolute error for 
                                           // the underflow range
	if (bla)  // multiplication not exact
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_h,d1);  // d1:  absolute rounding error
	}
	d1 = addu(d,d1);      // d1: absolute error for the normalized range
	if (d1>del) del = d1; // calculating the maximum 
    }
} // abs_divh1

// ---------------------------------------------------------------------------
// --------------------------- Delta(a), Eps(b) ------------------------------
// ---------------------------------------------------------------------------

void abs_addh2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A+B and an upper bound del of the absolute error by 
               // calculating (a+b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    int e,e1,e2;
    bool bl1,bl2,bl3;
    real d1 = Sup( abs(B) ),d,S,S1,I1;  
    d1 = mulu(d1,epsb);
    d = addu(dela,d1);  // d = reproduced data error
    R=A+B;
    A1 = A + interval(-dela,dela);
    B1 = B + interval(-d1,d1);
    bl1 = (A1==0 || B1==0);
    bl2 = (!(0<=B1) && A1/B1<=interval(-2,-0.5));
    aR = abs(A1+B1);  S1 = Sup(aR);  I1 = Inf(aR);
    e = e_IEEE(S1);
    e1 = e_IS(A1);  e2 = e_IS(B1); 
    if(e2<e1) e1=e2; // Minimumbildung
    bl3 = (e<=e1);
    if (bl1 || bl2 || bl3) del=d;
    else{
    aR = abs(R);  S = Sup(aR);
    if (I1>=MinReal) // All sums only in the normalized range
    {
	d1 = addu(d,S);
	d1 = mulu(eps_h,d1);  // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (S1<MinReal) // All sums only in the denormalized range U
    {
	del = addu(d,MinReal);
    } else // Sums in normalized and denormalized range --> maximum!:
    {
	del = addu(d,MinReal); // del: absolute error for the underflow range
        d1 = addu(d,S);
	d1 = mulu(eps_h,d1);  // d1:  absolute rounding error
	d1 = addu(d,d1);      // d1:  absolute error for the normalized range
	if (d1>del) del = d1; // calculating the maximum
    } }
} // abs_addh2

void abs_subh2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A-B and an upper bound del of the absolute error by 
               // calculating (a-b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    int e,e1,e2;
    bool bl1,bl2,bl3;
    real d1=Sup( abs(B) ),d,S,S1,I1; 
    d1 = mulu(d1,epsb);
    d = addu(dela,d1);  // d = reproduced data error
    R=A-B;
    A1 = A + interval(-dela,dela);
    B1 = B + interval(-d1,d1);
    bl1 = (A1==0 || B1==0);
    bl2 = (!(0<=B1) && A1/B1<=interval(0.5,2));
    aR = abs(A1-B1);  S1 = Sup(aR);  I1 = Inf(aR);
    e = e_IEEE(S1);
    e1 = e_IS(A1);  e2 = e_IS(B1); 
    if(e2<e1) e1=e2; // Minimumbildung
    bl3 = (e<=e1);
    if (bl1 || bl2 || bl3) del=d;
    else{
    aR = abs(R);  S = Sup(aR);
    if (I1>=MinReal) // All differences only in the normalized range
    {
	d1 = addu(d,S);
	d1 = mulu(eps_h,d1);  // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (S1<MinReal) // All differences only in the denormalized range U
    {
	del = addu(d,MinReal);
    } else // differences in normalized and denormalized range --> maximum!:
    {
	del = addu(d,MinReal); // del: absolute error for the underflow range
        d1  = addu(d,S);
	d1  = mulu(eps_h,d1);  // d1:  absolute rounding error
	d1  = addu(d,d1);      // d1:  absolute error for the normalized range
	if (d1>del) del = d1;  // calculating the maximum
    } }
} // abs_subh2

void abs_mulh2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A*B and an upper bound del of the absolute error by 
               // calculating (a*b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    real d = mulu(dela,epsb),d1;
    real SA=Sup( abs(A) ), SB=Sup( abs(B) );
    d1 = mulu(SA,epsb);
    d = addu(d,d1); 
    d = addu(d,dela);
    d = mulu(SB,d);  // d: reproduced data error

    R=A*B;
    interval aR = abs(R),A1,B1;
    real S = Sup(aR);

    A1 = sign(dela)!=0 ? A+interval(-dela,dela) : A;
    d1 = Sup( abs(B) );  d1 = mulu(d1,epsb);
    B1 = sign(epsb)!=0 ? B+interval(-d1,d1) : B;
    aR = abs(A1*B1);  SA = Sup(aR);  SB = Inf(aR);

    d1 = mul_exakt(A1,B1,aR); // d1=0, if multiplication exact;

    if (SB>=MinReal) // Products only in the normalized range
    {
	if (d1!=0) // Multiplication not exact:
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_h,d1);
	}
	del = addu(d,d1);
    } else if (SA<MinReal) // Products only in the denormalized range U
    {
	del = d1!=0 ? addu(d,MinReal) : d;
    } else // Products in normalized and denormalized range --> maximum!:
    {
	del = d1!=0 ? addu(d,MinReal) : d;  // del: absolute error for
                                            //      the underflow range
	if (d1!=0) // Multiplication not exact:
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_h,d1); // d1: absolute rounding error
	}
	d1 = addu(d,d1); // d1: absolute error for the normalized range
	if (d1>del) del = d1; // calculating the maximum
    }
} // abs_mulh2

void abs_divh2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A/B and an upper bound del of the absolute error by 
               // calculating (a/b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    real d=0,d1,S,IB;
    S=Sup( abs(A) );  IB=Inf( abs(B) );
    bool d_inexact;
    A1 = sign(dela)!=0 ? A+interval(-dela,dela) : A;
    d1 = Sup( abs(B) );  d1 = mulu(d1,epsb);
    B1 = sign(epsb)!=0 ? B+interval(-d1,d1) : B;
    if (0<=B1)
    {   
	cerr << "abs_divh2(...): Absolute error doesn't exist." << endl;
	exit(1);
    }  
    if (sign(dela)!=0 || sign(epsb)!=0) 
    {
        d1 = mulu(S,epsb);
        d1 = addu(d1,dela);  // numerator ready
        S = subd(1,epsb);   
	S = muld(IB,S);    // denominator ready
        if (S<=0) 
        {   
	    cerr << "abs_divh2(...): Absolute error doesn't exist." << endl;
	    exit(1);
        }
        d = divu(d1,S);  // d: reproduced data error
    }
    R=A/B;  // 0 not in B !!
    aR = abs(R);  S=Sup(aR);
    aR = abs(A1/B1);
    d1 = div_exakt(A1,B1,aR);  d_inexact = ( sign(d1)!=0 );

    if (Inf(aR)>=MinReal) // quotients only in the normalized range
    {
	if (d_inexact)
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_h,d1);
	}   // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (Sup(aR)<MinReal) // quotients only in the denormalized range U
    {
	del = d_inexact ? addu(d,MinReal) : d;
    } else // Quotients in normalized and denormalized range --> maximum!:
    {
	del = d_inexact ? addu(d,MinReal) : d; // del: absolute error for 
                                           // the underflow range
	if (d_inexact) 
	{  
	    d1 = addu(d,S);
	    d1 = mulu(eps_h,d1);  // d1:  absolute rounding error
	}
	d1 = addu(d,d1);      // d1: absolute error for the normalized range
	if (d1>del) del = d1; // calculating the maximum 
    }
} // abs_divh2

// ---------------------------------------------------------------------------
// --------------------------- Eps(a), Delta(b) ------------------------------
// ---------------------------------------------------------------------------

void abs_addh3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A+B and an upper bound del of the absolute error by 
               // calculating (a+b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    int e,e1,e2;
    bool bl1,bl2,bl3;
    real d1=Sup( abs(A) ),d,S,S1,I1;  
    d1 = mulu(d1,epsa);
    d = addu(delb,d1);  // d = reproduced data error
    R=A+B;
    A1 = A + interval(-d1,d1);
    B1 = B + interval(-delb,delb);
    bl1 = (A1==0 || B1==0);
    bl2 = (!(0<=B1) && A1/B1<=interval(-2,-0.5));
    aR = abs(A1+B1);  S1 = Sup(aR);  I1 = Inf(aR);
    e = e_IEEE(S1);
    e1 = e_IS(A1);  e2 = e_IS(B1); 
    if(e2<e1) e1=e2; // Minimumbildung
    bl3 = (e<=e1);
    if (bl1 || bl2 || bl3) del=d;
    else{
    aR = abs(R);  S = Sup(aR);
    if (I1>=MinReal) // All sums only in the normalized range
    {
	d1 = addu(d,S);
	d1 = mulu(eps_h,d1);  // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (S1<MinReal) // All sums only in the denormalized range U
    {
	del = addu(d,MinReal);
    } else // Sums in normalized and denormalized range --> maximum!:
    {
	del = addu(d,MinReal); // del: absolute error for the underflow range
        d1 = addu(d,S);
	d1 = mulu(eps_h,d1);  // d1:  absolute rounding error
	d1 = addu(d,d1);      // d1:  absolute error for the normalized range
	if (d1>del) del = d1; // calculating the maximum
    } }
} // abs_addh3

void abs_subh3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A-B and an upper bound del of the absolute error by 
               // calculating (a-b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    int e,e1,e2;
    bool bl1,bl2,bl3;
    real d1=Sup( abs(A) ),S,S1,I1,d; 
    d1 = mulu(d1,epsa);
    d = addu(delb,d1);  // d = reproduced data error
    R=A-B;
    A1 = A + interval(-d1,d1);
    B1 = B + interval(-delb,delb);
    bl1 = (A1==0 || B1==0);
    bl2 = (!(0<=B1) && A1/B1<=interval(0.5,2));
    aR = abs(A1-B1);  S1 = Sup(aR);  I1 = Inf(aR);
    e = e_IEEE(S1);
    e1 = e_IS(A1);  e2 = e_IS(B1); 
    if(e2<e1) e1=e2; // Minimumbildung
    bl3 = (e<=e1);
    if (bl1 || bl2 || bl3) del=d;
    else{
    aR = abs(R);  S = Sup(aR);
    if (I1>=MinReal) // All differences only in the normalized range
    {
	d1 = addu(d,S);
	d1 = mulu(eps_h,d1);  // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (S1<MinReal) // All differences only in the denormalized range U
    {
	del = addu(d,MinReal);
    } else // differences in normalized and denormalized range --> maximum!:
    {
	del = addu(d,MinReal); // del: absolute error for the underflow range
        d1  = addu(d,S);
	d1  = mulu(eps_h,d1);  // d1:  absolute rounding error
	d1  = addu(d,d1);      // d1:  absolute error for the normalized range
	if (d1>del) del = d1;  // calculating the maximum
    } }
} // abs_subh3

void abs_mulh3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A*B and an upper bound del of the absolute error by 
               // calculating (a*b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    real d = mulu(epsa,delb), d1,d2;
    real SA=Sup( abs(A) ), SB=Sup( abs(B) );
    d1 = mulu(SB,epsa);
    d = addu(d,d1); 
    d = addu(d,delb);
    d = mulu(SA,d);  // d: reproduced data error

    R=A*B;
    interval A1,B1,aR = abs(R);
    real S = Sup(aR),I;

    SA = mulu(SA,epsa);
    A1 = sign(epsa)!=0 ? A+interval(-SA,SA) : A;
    B1 = sign(delb)!=0 ? B+interval(-delb,delb) : B;
    aR = abs(A1*B1);  I = Inf(aR);  SA = Sup(aR);

    d1 = mul_exakt(A1,B1,aR);  // d1=0, if multiplication exact;
                                // otherwise: d1=1.
    if (I>=MinReal) // Products only in the normalized range
    {
	if (d1!=0) // Multiplication not exact:
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_h,d1);
	}
	del = addu(d,d1);
    } else if (SA<MinReal) // Products only in the denormalized range U
    {
	del = d1!=0 ? addu(d,MinReal) : d;
    } else // Products in normalized and denormalized range --> maximum!:
    {
	del = d1!=0 ? addu(d,MinReal) : d;  // del: absolute error for
                                            //      the underflow range
	if (d1!=0) // Multiplication not exact:
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_h,d1); // d1: absolute rounding error
	}
	d2 = addu(d,d1); // d2: absolute error for the normalized range
	if (d2>del) del = d2; // calculating the maximum
    }
} // abs_mulh3

void abs_divh3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A/B and an upper bound del of the absolute error by 
               // calculating (a/b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    real IB=Inf( abs(B) ),d=0,d1,SA,S;
    bool bla = (sign(epsa)!=0);
    SA = Sup( abs(A) );    SA = mulu(SA,epsa);
    A1 = bla ? A+interval(-SA,SA) : A;
    B1 = sign(delb)!=0 ? B+interval(-delb,delb) : B;
    if (0<=B1)
    {   
	cerr << "abs_divh3(...): Absolute error doesn't exist." << endl;
	exit(1);
    }
    if (bla || sign(delb)!=0) 
    {
        d1 = divu(delb,IB);
        d1 = addu(d1,epsa);  // numerator ready
        S = subd(IB,delb);   
        if (S<=0) 
        {   
	    cerr << "abs_divh3(...): Absolute error doesn't exist." << endl;
	    exit(1);
        }
        d = divu(d1,S); 
        d = mulu(SA,d); // d: reproduced data error
    }
    R=A/B;
    aR = abs(R);  S = Sup(aR);
    aR = abs(A1/B1);
    d1 = div_exakt(A1,B1,aR);  bla = ( sign(d1)!=0 );

    if (Inf(aR)>=MinReal) // Quotients only in the normalized range
    {
	if (bla)
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_h,d1);
	}   // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (Sup(aR)<MinReal) // Quotients only in the denormal. range U
    {
	del = bla ? addu(d,MinReal) : d;
    } else // Quotients in normalized and denormalized range --> maximum!:
    {
	del = bla ? addu(d,MinReal) : d; // del: absolute error for 
                                         // the underflow range
	if (bla)
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_h,d1);  // d1:  absolute rounding error
	}
	d1 = addu(d,d1);      // d1: absolute error for the normal. range
	if (d1>del) del = d1; // calculating the maximum 
    }
} // abs_divh3

// ---------------------------------------------------------------------------
// ---------------------------- Eps(a), Eps(b) -------------------------------
// ---------------------------------------------------------------------------

void abs_addh4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A+B and an upper bound del of the absolute error by 
               // calculating (a+b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    int e,e1,e2;
    bool bl1,bl2,bl3;
    real SA=Sup( abs(A) ), SB=Sup( abs(B) ),d,d1,S,S1,I1;
    d1 = mulu(SA,epsa);  
    S = mulu(SB,epsb);
    d  = addu(d1,S);  // d = reproduced data error
    R=A+B;
    A1 = A + interval(-d1,d1);
    B1 = B + interval(-S,S);
    bl1 = (A1==0 || B1==0);
    bl2 = (!(0<=B1) && A1/B1<=interval(-2,-0.5));
    aR = abs(A1+B1);  S1 = Sup(aR);  I1 = Inf(aR);
    e = e_IEEE(S1);
    e1 = e_IS(A1);  e2 = e_IS(B1); 
    if(e2<e1) e1=e2; // Minimumbildung
    bl3 = (e<=e1);
    if (bl1 || bl2 || bl3) del=d;
    else{
    aR = abs(R);  S = Sup(aR);
    if (I1>=MinReal) // All sums only in the normalized range
    {
	d1 = addu(d,S);
	d1 = mulu(eps_h,d1);  // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (S1<MinReal) // All sums only in the denormalized range U
    {
	del = addu(d,MinReal);
    } else // Sums in normalized and denormalized range --> maximum!:
    {
	del = addu(d,MinReal); // del: absolute error for the underflow range
        d1 = addu(d,S);
	d1 = mulu(eps_h,d1);  // d1:  absolute rounding error
	d1 = addu(d,d1);      // d1:  absolute error for the normalized range
	if (d1>del) del = d1; // calculating the maximum
    } }
} // abs_addh4

void abs_subh4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A-B and an upper bound del of the absolute error by 
               // calculating (a-b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    int e,e1,e2;
    bool bl1,bl2,bl3;
    real SA=Sup( abs(A) ), SB=Sup( abs(B) ),d,d1,S,S1,I1;
    d1 = mulu(SA,epsa);  
    S = mulu(SB,epsb);
    d  = addu(d1,S);  // d = reproduced data error
    R=A-B;
    A1 = A + interval(-d1,d1);
    B1 = B + interval(-S,S);
    bl1 = (A1==0 || B1==0);
    bl2 = (!(0<=B1) && A1/B1<=interval(0.5,2));
    aR = abs(A1-B1);  S1 = Sup(aR);  I1 = Inf(aR);
    e = e_IEEE(S1);
    e1 = e_IS(A1);  e2 = e_IS(B1); 
    if(e2<e1) e1=e2; // Minimumbildung
    bl3 = (e<=e1);
    if (bl1 || bl2 || bl3) del=d;
    else{
    aR = abs(R);  S = Sup(aR);
    if (I1>=MinReal) // All differences only in the normalized range
    {
	d1 = addu(d,S);
	d1 = mulu(eps_h,d1);  // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (S1<MinReal) // All differences only in the denormalized range U
    {
	del = addu(d,MinReal);
    } else // differences in normalized and denormalized range --> maximum!:
    {
	del = addu(d,MinReal); // del: absolute error for the underflow range
        d1 = addu(d,S);
	d1 = mulu(eps_h,d1);  // d1:  absolute rounding error
	d1 = addu(d,d1);      // d1:  absolute error for the normalized range
	if (d1>del) del = d1; // calculating the maximum
    } }
} // abs_subh4

void abs_mulh4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A*B and an upper bound del of the absolute error by 
               // calculating (a*b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    real d = mulu(epsa,epsb),d1;
    d = addu(d,epsb);
    d = addu(d,epsa); 
    real SA=Sup( abs(A) ), SB=Sup( abs(B) );
    d = mulu(d,SA);
    d = mulu(d,SB);  // d: reproduced data error  

    R=A*B;
    interval A1,B1,aR = abs(R);
    real S = Sup(aR), I;

    SA = mulu(SA,epsa);
    A1 = sign(epsa)!=0 ? A+interval(-SA,SA) : A;
    SB = Sup( abs(B) );  SB = mulu(SB,epsb);
    B1 = sign(epsb)!=0 ? B+interval(-SB,SB) : B;

    aR = abs(A1*B1);   I = Inf(aR);  SA = Sup(aR);

    d1 = mul_exakt(A1,B1,aR); // d1=0, if multiplication exact;
                               // otherwise: d1=1.
    if (I>=MinReal) // All products only in the normalized range
    {
	if (d1!=0) // Multiplication not exact:
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_h,d1);
	}
	del = addu(d,d1);
    } else if (SA<MinReal) // Products only in the denormalized range U
    {
	del = d1!=0 ? addu(d,MinReal) : d;
    } else // Products in normalized and denormalized range --> maximum!:
    {
	del = d1!=0 ? addu(d,MinReal) : d;  // del: absolute error for
                                            //      the underflow range
	if (d1!=0) // Multiplication not exact:
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_h,d1); // d1: absolute rounding error
	}
	d1 = addu(d,d1); // d1: absolute error for the normalized range
	if (d1>del) del = d1; // calculating the maximum
    }
} // abs_mulh4

void abs_divh4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A/B and an upper bound del of the absolute error by 
               // calculating (a/b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    bool bla = ( sign(epsa)!=0 );
    real IB=Inf( abs(B) ), d=0, SA=Sup( abs(A) ),d1,S;
    S = mulu(SA,epsa);
    A1 = bla ? A+interval(-S,S) : A;
    S = Sup( abs(B) );  S = mulu(S,epsb);
    B1 = sign(epsb)!=0 ? B+interval(-S,S) : B;
    if (0<=B1)
    {
	cerr << "abs_divh4(...): Absolute error doesn't exist." << endl;
	exit(1);
    }
    if (bla || sign(epsb)!=0)
    {
        d1 = addu(epsa,epsb);
        d1 = mulu(d1,SA); // numerator ready
        S = subd(1,epsb);  
        if (S<=0) 
        {
	    cerr << "abs_divh4(...): Absolute error doesn't exist." << endl;
	    exit(1);
        }
        S = muld(IB,S); // denominator ready
        d = divu(d1,S);  // d: reproduced data error
    }
    R=A/B;
    aR = abs(R);  S = Sup(aR);
    aR = abs(A1/B1);
    d1 = div_exakt(A1,B1,aR);   bla = ( sign(d1)!=0 );

    if (Inf(aR)>=MinReal) // Quotients only in the normalized range
    {
	if (bla)
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_h,d1); 
	}   // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (Sup(aR)<MinReal) // Quotients only in the denormal. range U
    {
	del = bla ? addu(d,MinReal) : d;
    } else // Quotients in normalized and denormalized range --> maximum!:
    {
	del = bla ? addu(d,MinReal) : d;  // del: absolute error for
                                          // the underflow range
        if (bla)
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_h,d1);  // d1:  absolute rounding error
	}
	d1 = addu(d,d1);      // d1: absolute error for the normalized range
	if (d1>del) del = d1; // calculating the maximum 
    }
} // abs_divh4

// ---------------------------------------------------------------------------
// ------------------- Calculating the relative errors -----------------------
// ---------------------------------------------------------------------------
// -------------------------- Delta(a), Delta(b) -----------------------------
// ---------------------------------------------------------------------------

void rel_addh1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A+B and an upper bound eps of the relative error by 
               // calculating (a+b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    int e,e1,e2;
    bool bl1,bl2,bl3;
    real d=0;  // reproduced relative data error==0, if dela==delb==0;
    R = A+B;
    aR = abs(R);
    real I = Inf(aR),d1,S1,I1;

    if (sign(dela)!=0 || sign(delb)!=0)
    {
	d = addu(dela,delb); // d != 0
        if (sign(I)==0)
        {
	    cerr << "rel_addh1: Relative error doesn't exist!" << endl;
            exit(1);
        }
        d = divu(d,I);  // d: reproduced relative data error
    }
    A1 = A + interval(-dela,dela);
    B1 = B + interval(-delb,delb);
    bl1 = (A1==0 || B1==0);
    bl2 = (!(0<=B1) && A1/B1<=interval(-2,-0.5));
    aR = abs(A1+B1);   S1 = Sup(aR);  I1 = Inf(aR);
    e = e_IEEE(S1);
    e1 = e_IS(A1);  e2 = e_IS(B1); 
    if(e2<e1) e1=e2; // Minimumbildung
    bl3 = (e<=e1);
    if (bl1 || bl2 || bl3) eps=d; 
    else{ 
    if (I1>=MinReal) // sums only in the normalized range
    {
	d1  = addu(1,d);
	d1  = mulu(eps_h,d1); // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (S1<MinReal) // All sums only in the denormalized range U
    {
	d1 = divu(MinReal,I);
	eps = addu(d,d1);
    } else // sums in normalized and denormalized range --> maximum!:
    {
	d1 = divu(MinReal,I);
	eps = addu(d,d1);     // eps:  relative error in U
	d1  = addu(1,d);
	d1  = mulu(eps_h,d1); // d1:  relative rounding error
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    } }
} // rel_addh1

void rel_subh1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A-B and an upper bound eps of the relative error by 
               // calculating (a-b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    int e,e1,e2;
    bool bl1,bl2,bl3;
    real d=0;  // reproduced relative data error==0, if dela==delb==0;
    R = A-B;
    aR = abs(R);
    real I = Inf(aR),d1,S1,I1;
    if (sign(dela)!=0 || sign(delb)!=0)
    {
	d = addu(dela,delb);  // d != 0
        if (sign(I)==0)
        {
	    cerr << "rel_subh1: Relative error doesn't exist!" << endl;
            exit(1);
        }
        d = divu(d,I);  // d: reproduced relative data error
    }
    A1 = A + interval(-dela,dela);
    B1 = B + interval(-delb,delb);
    bl1 = (A1==0 || B1==0);
    bl2 = (!(0<=B1) && A1/B1<=interval(0.5,2));
    aR = abs(A1-B1);  S1 = Sup(aR);  I1 = Inf(aR);
    e = e_IEEE(S1);
    e1 = e_IS(A1);  e2 = e_IS(B1); 
    if(e2<e1) e1=e2; // Minimumbildung
    bl3 = (e<=e1);
    if (bl1 || bl2 || bl3) eps=d;
    else{
    if (I1>=MinReal) // All differences only in the normalized range
    {
	d1  = addu(1,d);
	d1  = mulu(eps_h,d1); // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (S1<MinReal) // All differences only in the denormalized range U
    {
	d1 = divu(MinReal,I);
	eps = addu(d,d1);
    } else // differences in normalized and denormalized range --> maximum!:
    {
	d1  = divu(MinReal,I);
	eps = addu(d,d1);     // eps:  relative error in U
	d1  = addu(1,d);
	d1  = mulu(eps_h,d1); // d1:  relative rounding error
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    } }
} // rel_subh1

void rel_mulh1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A*B and an upper bound eps of the relative error by 
               // calculating (a*b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    R = A*B;
    interval aR = abs(R),A1,B1;
    // real S = Sup(aR);
    real I = Inf(aR), IA = Inf( abs(A) ),IB,d,d1;
    if ( (A==0 && dela==0) || (B==0 && delb==0) ) d = 0;
    else 
    {
	IB = Inf( abs(B) );
	if (dela!=0 && IA==0 || delb!=0 && IB==0 )
	{   
	    cerr << "rel_mulh1(...): Relative error doesn't exist." 
                 << endl;
	    exit(1);
	}
	d  = dela==0 ? 0 : divu(dela,IA);
        d1 = delb==0 ? 0 : divu(delb,IB);
        IA = addu(d1,d);
	d = mulu(d,d1); // 0 * Infinity --> Speicherzugriffsfehler
        d = addu(d,IA); // d:  reproduced relative data error
    }
    A1 = sign(dela)!=0 ? A+interval(-dela,dela) : A;
    B1 = sign(delb)!=0 ? B+interval(-delb,delb) : B;
    aR = abs(A1*B1);  IA = Sup(aR);  IB = Inf(aR);
    d1 = mul_exakt(A1,B1,aR); // d1=0 if multiplication is exact,
                               //      otherwise d1=1;
    if (IB>=MinReal) // All products only in the normalized range
    {
	if (d1!=0)
        {
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1);
        }   // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (IA<MinReal) // All products only in the denormalized range U
    {
	if (d1!=0) d1 = divu(MinReal,I);  // I=0 --> d1 = <+Infinity>
	eps = addu(d,d1);
    } else // products in normalized and denormalized range --> maximum!:
    {
	if (d1!=0)
	{ IB = divu(MinReal,I);  eps = addu(d,IB); }
	else eps = d;  // eps: relative error in U
	if (d1!=0){
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1); // d1:  relative rounding error
	}
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    }
} // rel_mulh1

void rel_divh1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A/B and an upper bound eps of the relative error by 
               // calculating (a/b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aQ;
    A1 = sign(dela)!=0 ? A+interval(-dela,dela) : A;
    B1 = sign(delb)!=0 ? B+interval(-delb,delb) : B;
    if (0<=B1)
    {   
	cerr << "rel_divh1(...): Relative error doesn't exist." << endl;
	exit(1);
    }
    real IB=Inf( abs(B) ),I,d1,d;
    R = A/B;  aQ = abs(R);  I = Inf(aQ);
    bool bla = sign(dela)!=0;
    if (!bla && (A==0 || delb==0)) d = 0;
    else 
    {
	d1 = Inf( abs(A) );
	if (bla && d1==0)
	{   
	    cerr << "rel_divh1(...): Relative error doesn't exist." 
                 << endl;
	    exit(1);
	}
        d  = bla ? divu(dela,d1) : 0;
        d1 = divu(delb,IB);
        d  = addu(d,d1);
        d1 = subd(1,d1);
        if (d1<=0) 
        {
	    cerr << "rel_divh1(...): Relative error doesn't exist." 
		 << endl;
	    exit(1);
        }
        d = divu(d,d1);  // d:  reproduced relative data error
    }
    aQ = abs(A1/B1);
    d1 = div_exakt(A1,B1,aQ);

    if (Inf(aQ)>=MinReal) // All quotients only in the normalized range
    {
	if (d1!=0)
	{
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1);
	}   // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (Sup(aQ)<MinReal) // quotients only in the denormalized range U
    {
	if (d1!=0) d1 = divu(MinReal,I); // I=0 --> d1 = <+Infinity>
	eps = addu(d,d1);
    } else // quotients in normalized and denormalized range --> maximum!:
    {   
	if (d1!=0)
	{
	    IB  = divu(MinReal,I);
	    eps = addu(d,IB);     // eps: relative error in U
	}
	else eps = d;
	if (d1!=0)
	{
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1); // d1:  relative rounding error
	}
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    }
} // rel_divh1

// ---------------------------------------------------------------------------
// --------------------------- Delta(a), Eps(b) ------------------------------
// ---------------------------------------------------------------------------

void rel_addh2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A+B and an upper bound eps of the relative error by 
               // calculating (a+b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    int e,e1,e2;
    bool bl1,bl2,bl3;
    R = A+B;
    aR = abs(R);
    real I = Inf(aR),d,d1,S1,I1;

    d  = ForMaxS2(dela,epsb,A,Inf(B));
    d1 = ForMaxS2(dela,epsb,A,Sup(B));
    if (d1>d) d = d1;  // d: reproduced relative data error
    A1 = A + interval(-dela,dela);
    d1 = Sup( abs(B) );
    d1 = mulu(d1,epsb);
    B1 = B + interval(-d1,d1);
    bl1 = (A1==0 || B1==0);
    bl2 = (!(0<=B1) && A1/B1<=interval(-2,-0.5));
    aR = abs(A1+B1);   S1 = Sup(aR);  I1 = Inf(aR);
    e = e_IEEE(S1);
    e1 = e_IS(A1);  e2 = e_IS(B1); 
    if(e2<e1) e1=e2; // Minimumbildung
    bl3 = (e<=e1);
    if (bl1 || bl2 || bl3) eps=d;
    else{
    if (I1>=MinReal) // sums only in the normalized range
    {
	d1  = addu(1,d);
	d1  = mulu(eps_h,d1); // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (S1<MinReal) // All sums only in the denormalized range U
    {
	d1 = divu(MinReal,I);
	eps = addu(d,d1);
    } else // sums in normalized and denormalized range --> maximum!:
    {
	d1 = divu(MinReal,I);
	eps = addu(d,d1);     // eps:  relative error in U
	d1  = addu(1,d);
	d1  = mulu(eps_h,d1); // d1:  relative rounding error
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    } }
} // rel_addh2

void rel_subh2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A-B and an upper bound eps of the relative error by 
               // calculating (a-b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    int e,e1,e2;
    bool bl1,bl2,bl3;
    R = A-B;
    aR = abs(R);
    real I = Inf(aR),d,d1,S1,I1;

    d  = ForMaxD2(dela,epsb,A,Inf(B));
    d1 = ForMaxD2(dela,epsb,A,Sup(B));
    if (d1>d) d = d1;  // d: reproduced relative data error
    A1 = A + interval(-dela,dela);
    d1 = Sup( abs(B) );
    d1 = mulu(d1,epsb);
    B1 = B + interval(-d1,d1);
    bl1 = (A1==0 || B1==0);
    bl2 = (!(0<=B1) && A1/B1<=interval(0.5,2));
    aR = abs(A1-B1);  S1 = Sup(aR);  I1 = Inf(aR);
    e = e_IEEE(S1);
    e1 = e_IS(A1);  e2 = e_IS(B1); 
    if(e2<e1) e1=e2; // Minimumbildung
    bl3 = (e<=e1);
    if (bl1 || bl2 || bl3) eps=d;
    else{
    if (I1>=MinReal) // differences only in the normalized range
    {
	d1  = addu(1,d);
	d1  = mulu(eps_h,d1); // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (S1<MinReal) // All differences only in the denormalized range U
    {
	d1 = divu(MinReal,I);
	eps = addu(d,d1);
    } else // differences in normalized and denormalized range --> maximum!:
    {
	d1  = divu(MinReal,I);
	eps = addu(d,d1);     // eps:  relative error in U
	d1  = addu(1,d);
	d1  = mulu(eps_h,d1); // d1:  relative rounding error
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    } }
} // rel_subh2

void rel_mulh2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A*B and an upper bound eps of the relative error by 
               // calculating (a*b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    R = A*B;
    interval aR = abs(R),A1,B1;
    real I = Inf(aR), IA = Inf( abs(A) ),d,d1,SA;
    if (A==0 && dela==0 || B==0 && epsb==0) d = 0;
    else if (dela==0) d = epsb;
         else
	 {
	     if (IA==0)
	     {
	     cerr << "rel_mulh2(...): Relative error doesn't exist." 
                  << endl;
	     exit(1);
	     }
	     d = divu(dela,IA);
	     d1 = mulu(d,epsb);
	     d1 = addu(d1,epsb);
	     d = addu(d1,d); 
	 }  // d:  reproduced relative data error
    A1 = sign(dela)!=0 ? A+interval(-dela,dela) : A;
    d1 = Sup( abs(B) );  d1 = mulu(d1,epsb);
    B1 = sign(epsb)!=0 ? B+interval(-d1,d1) : B;
    aR = abs(A1*B1);  SA = Sup(aR);  IA = Inf(aR); 
   
    d1 = mul_exakt(A1,B1,aR);

    if (IA>=MinReal) // All products only in the normalized range
    {
	if (d1!=0)
        {
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1);
        }   // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (SA<MinReal) // All products only in the denormalized range U
    {
	if (d1!=0) d1 = divu(MinReal,I);  // I=0 --> d1 = <+Infinity>
	eps = addu(d,d1);
    } else // products in normalized and denormalized range --> maximum!:
    {
	if (d1!=0)
	{ IA = divu(MinReal,I);  eps = addu(d,IA); }
	else eps = d;  // eps: relative error in U
	if (d1!=0){
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1); // d1:  relative rounding error
	}
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    }
} // rel_mulh2

void rel_divh2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A/B and an upper bound eps of the relative error by 
               // calculating (a/b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    real d,d1,I;
    interval A1,B1,aR;
    bool bla = sign(dela)!=0;
    A1 = bla ? A+interval(-dela,dela) : A;
    d1 = Sup( abs(B) );  d1 = mulu(d1,epsb);
    B1 = sign(epsb)!=0 ? B+interval(-d1,d1) : B;
    if (0<=B1)
    {
	cerr << "rel_divh2(...): Relative error doesn't exist." << endl;
	exit(1);
    }
    if (!bla && (A==0 || epsb==0)) d = 0;
    else
    {
	d1 = Inf( abs(A) );
	if (bla && d1==0)
	{
	    cerr << "rel_divh2(...): Relative error doesn't exist." 
                 << endl;
	    exit(1);
	}
        d  = bla ? divu(dela,d1) : 0;
        d  = addu(d,epsb);
        d1 = subd(1,epsb);
        if (d1<=0) 
        {
	    cerr << "rel_divh2(...): Relative error doesn't exist." << endl;
	    exit(1);
        }
        d = divu(d,d1);  // d:  reproduced relative data error
    }
    R = A/B;
    aR = abs(R);  I = Inf(aR);
    aR = abs(A1/B1);
    d1 = div_exakt(A1,B1,aR);  bla = ( sign(d1)!=0 );

    if (Inf(aR)>=MinReal) // All quotients only in the normalized range
    {
	if (bla)
	{
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1);
	}   // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (Sup(aR)<MinReal) // quotients only in the denormalized range U
    {
	if (bla) d1 = divu(MinReal,I); // I=0 --> d1 = <+Infinity>
	eps = addu(d,d1);
    } else // quotients in normalized and denormalized range --> maximum!:
    {
	if (bla)
	{
	    I  = divu(MinReal,I);
	    eps = addu(d,I);     // eps: relative error in U
	}
	else eps = d;
	if (bla)
	{
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1); // d1:  relative rounding error
	}
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    }
} // rel_divh2

// ---------------------------------------------------------------------------
// --------------------------- Eps(a), Delta(b) ------------------------------
// ---------------------------------------------------------------------------

void rel_addh3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A+B and an upper bound eps of the relative error by 
               // calculating (a+b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    int e,e1,e2;
    bool bl1,bl2,bl3;
    R = A+B;
    aR = abs(R);
    real I = Inf(aR),d,d1,S1,I1;

    d  = ForMaxS3(epsa,delb,B,Inf(A));
    d1 = ForMaxS3(epsa,delb,B,Sup(A));
    if (d1>d) d = d1;  // d: reproduced relative data error
    d1 = Sup( abs(A) );
    d1 = mulu(d1,epsa);
    A1 = A + interval(-d1,d1);
    B1 = B + interval(-delb,delb);
    bl1 = (A1==0 || B1==0);
    bl2 = (!(0<=B1) && A1/B1<=interval(-2,-0.5));
    aR = abs(A1+B1);   S1 = Sup(aR);  I1 = Inf(aR);
    e = e_IEEE(S1);
    e1 = e_IS(A1);  e2 = e_IS(B1); 
    if(e2<e1) e1=e2; // Minimumbildung
    bl3 = (e<=e1);
    if (bl1 || bl2 || bl3) eps=d;
    else{
    if (I1>=MinReal) // sums only in the normalized range
    {
	d1  = addu(1,d);
	d1  = mulu(eps_h,d1); // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (S1<MinReal) // All sums only in the denormalized range U
    {
	d1 = divu(MinReal,I);
	eps = addu(d,d1);
    } else // sums in normalized and denormalized range --> maximum!:
    {
	d1 = divu(MinReal,I);
	eps = addu(d,d1);     // eps:  relative error in U
	d1  = addu(1,d);
	d1  = mulu(eps_h,d1); // d1:  relative rounding error
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    }}
} // rel_addh3

void rel_subh3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A-B and an upper bound eps of the relative error by 
               // calculating (a-b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    int e,e1,e2;
    bool bl1,bl2,bl3;
    R = A-B;
    aR = abs(R);
    real I = Inf(aR),d,d1,S1,I1;

    d  = ForMaxD3(epsa,delb,B,Inf(A));
    d1 = ForMaxD3(epsa,delb,B,Sup(A));
    if (d1>d) d = d1;  // d: reproduced relative data error
    d1 = Sup( abs(A) );
    d1 = mulu(d1,epsa);
    A1 = A + interval(-d1,d1);
    B1 = B + interval(-delb,delb);
    bl1 = (A1==0 || B1==0);
    bl2 = (!(0<=B1) && A1/B1<=interval(0.5,2));
    aR = abs(A1-B1);   S1 = Sup(aR);  I1 = Inf(aR);
    e = e_IEEE(S1);
    e1 = e_IS(A1);  e2 = e_IS(B1); 
    if(e2<e1) e1=e2; // Minimumbildung
    bl3 = (e<=e1);
    if (bl1 || bl2 || bl3) eps=d;
    else{
    if (I1>=MinReal) // differences only in the normalized range
    {
	d1  = addu(1,d);
	d1  = mulu(eps_h,d1); // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (S1<MinReal) // All differences only in the denormalized range U
    {
	d1 = divu(MinReal,I);
	eps = addu(d,d1);
    } else // differences in normalized and denormalized range --> maximum!:
    {
	d1 = divu(MinReal,I);
	eps = addu(d,d1);     // eps:  relative error in U
	d1  = addu(1,d);
	d1  = mulu(eps_h,d1); // d1:  relative rounding error
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    } }
} // rel_subh3

void rel_mulh3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A*B and an upper bound eps of the relative error by 
               // calculating (a*b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    R = A*B;
    interval A1,B1,aR = abs(R);
    real SA, I = Inf(aR), IB = Inf( abs(B) ),d,d1;
    if (A==0 && epsa==0 || B==0 && delb==0) d = 0;
    else if (delb==0) d = epsa;
         else
         {  // delb != 0
	     if (sign(IB)==0)
	     {
	         cerr << "rel_mulh3(...): Relative error doesn't exist." 
                      << endl;
		 exit(1);
	     }
	     d  = divu(delb,IB);
	     d1 = mulu(d,epsa);
	     d1 = addu(d1,epsa);
	     d  = addu(d1,d);   
	 }  // d:  reproduced relative data error

    SA = Sup( abs(A) );  SA = mulu(SA,epsa);
    A1 = sign(epsa)!=0 ? A+interval(-SA,SA) : A;
    B1 = sign(delb)!=0 ? B+interval(-delb,delb) : B;
    aR = abs(A1*B1);  IB = Inf(aR);  SA = Sup(aR);

    d1 = mul_exakt(A1,B1,aR);  // d1=0, if exact multiplication
                               // otherwise d1=1;
    if (IB>=MinReal) // All products only in the normalized range
    {
	if (d1!=0)
        {
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1);
        }   // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (SA<MinReal) // All products only in the denormalized range U
    {
	if (d1!=0) d1 = divu(MinReal,I); // I=0 --> d1 = <+Infinity>
	eps = addu(d,d1);
    } else // products in normalized and denormalized range --> maximum!:
    {
	if (d1!=0)
	{ IB = divu(MinReal,I);  eps = addu(d,IB); }
	else eps = d;  // eps: relative error in U=(-MinReal,+Minreal)
	if (d1!=0){
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1); // d1:  relative rounding error
	}
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    }
} // rel_mulh3

void rel_divh3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A/B and an upper bound eps of the relative error by 
               // calculating (a/b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    real IB = Inf( abs(B) ),d,d1,I;
    bool bla = (sign(epsa)!=0);
    d1 = Sup( abs(A) );  d1 = mulu(d1,epsa);
    A1 = bla ? A+interval(-d1,d1) : A;
    B1 = sign(delb)!=0 ? B+interval(-delb,delb) : B;
    if (0<=B1) 
    {
	cerr << "rel_divh3(...): Relative error doesn't exist." << endl;
	exit(1);
    }
    if (A==0 || (!bla && sign(delb)==0)) d = 0;
    else
    {
        d  = mulu(IB,epsa);
        d  = addu(d,delb);
        d1 = subd(IB,delb);
        if (d1<=0) 
        {
	    cerr << "rel_divh3(...): Relative error doesn't exist." << endl;
	    exit(1);
        }
        d = divu(d,d1);  // d:  reproduced relative data error
    }
    R = A/B;
    aR = abs(R);  I = Inf(aR);
    aR = abs(A1/B1);
    d1 = div_exakt(A1,B1,aR);  bla = ( sign(d1)!=0 );

    if (Inf(aR)>=MinReal) // All quotients only in the normalized range
    {
	if (bla)
	{
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1);
	}   // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (Sup(aR)<MinReal) // quotients only in the denormal. range U
    {
	if (bla) d1 = divu(MinReal,I);
	eps = addu(d,d1);
    } else // quotients in normalized and denormalized range --> maximum!:
    {
	if (bla) 
	{
	    IB  = divu(MinReal,I);
	    eps = addu(d,IB);     // eps: relative error in U
	}
	else eps = d;
	if (bla)
	{
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1); // d1:  relative rounding error
	}
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    }
} // rel_divh3

// ---------------------------------------------------------------------------
// ---------------------------- Eps(a), Eps(b) -------------------------------
// ---------------------------------------------------------------------------

void rel_addh4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A+B and an upper bound eps of the relative error by 
               // calculating (a+b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    int e,e1,e2;
    bool bl1,bl2,bl3;
    real IA = Inf(A), SA = Sup(A), IB = Inf(B), SB = Sup(B),d,d1;
    d  = ForMaxS4(epsa,epsb,IA,IB);
    d1 = ForMaxS4(epsa,epsb,IA,SB);
    if (d1>d) d = d1;
    d1 = ForMaxS4(epsa,epsb,SA,IB);
    if (d1>d) d = d1;
    d1 = ForMaxS4(epsa,epsb,SA,SB);
    if (d1>d) d = d1;  // d: reproduced relative data error

    R = A+B;
    d1 = Sup( abs(A) );
    d1 = mulu(d1,epsa);
    A1 = A + interval(-d1,d1);
    d1 = Sup( abs(B) );
    d1 = mulu(d1,epsb);
    B1 = B + interval(-d1,d1);
    bl1 = (A1==0 || B1==0);
    bl2 = (!(0<=B1) && A1/B1<=interval(-2,-0.5));
    aR = abs(A1+B1);  SA = Sup(aR);  IB = Inf(aR);
    e = e_IEEE(SA);
    e1 = e_IS(A1);  e2 = e_IS(B1); 
    if(e2<e1) e1=e2; // Minimumbildung
    bl3 = (e<=e1);
    if (bl1 || bl2 || bl3) eps=d;
    else
    {
        aR = abs(R);
        real I = Inf(aR);
        if (IB>=MinReal) // sums only in the normalized range
        {
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1); // d1:  relative rounding error
	    eps = addu(d,d1);
        } else if (SA<MinReal) // All sums only in the denormalized range U
        {
	    d1 = divu(MinReal,I);
	    eps = addu(d,d1);
        } else // sums in normalized and denormalized range --> maximum!:
        {
	    d1 = divu(MinReal,I);
	    eps = addu(d,d1);     // eps:  relative error in U
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1); // d1:  relative rounding error
	    d1  = addu(d,d1); // d1:  relative error in the normalized range
	    if (d1>eps) eps = d1; // calculating the maximum 
        }
    }
} // rel_addh4

void rel_subh4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A-B and an upper bound eps of the relative error by 
               // calculating (a-b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    int e,e1,e2;
    bool bl1,bl2,bl3;
    real IA = Inf(A), SA = Sup(A), IB = Inf(B), SB = Sup(B),d,d1;
    d  = ForMaxD4(epsa,epsb,IA,IB);
    d1 = ForMaxD4(epsa,epsb,IA,SB);
    if (d1>d) d = d1;
    d1 = ForMaxD4(epsa,epsb,SA,IB);
    if (d1>d) d = d1;
    d1 = ForMaxD4(epsa,epsb,SA,SB);
    if (d1>d) d = d1;  // d: reproduced relative data error

    R = A-B;
    d1 = Sup( abs(A) );
    d1 = mulu(d1,epsa);
    A1 = A + interval(-d1,d1);
    d1 = Sup( abs(B) );
    d1 = mulu(d1,epsb);
    B1 = B + interval(-d1,d1);
    bl1 = (A1==0 || B1==0);
    bl2 = (!(0<=B1) && A1/B1<=interval(0.5,2));
    aR = abs(A1-B1);  SA = Sup(aR);  IB = Inf(aR);
    e = e_IEEE(SA);
    e1 = e_IS(A1);  e2 = e_IS(B1); 
    if(e2<e1) e1=e2; // Minimumbildung
    bl3 = (e<=e1);
    if (bl1 || bl2 || bl3) eps=d;
    else
    {
        aR = abs(R);
        real I = Inf(aR);
        if (IB>=MinReal) // differences only in the normalized range
        {
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1); // d1:  relative rounding error
	    eps = addu(d,d1);
        } else if (SA<MinReal) // All differcs. only in the denormal. range U
        {
	    d1 = divu(MinReal,I);
	    eps = addu(d,d1);
        } else // differences in normal. and denormalized range --> maximum!:
        {
	    d1 = divu(MinReal,I);
	    eps = addu(d,d1);     // eps:  relative error in U
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1); // d1:  relative rounding error
	    d1  = addu(d,d1); // d1:  relative error in the normalized range
	    if (d1>eps) eps = d1; // calculating the maximum 
        }
    }
} // rel_subh4

void rel_mulh4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A*B and an upper bound eps of the relative error by 
               // calculating (a*b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    R = A*B;
    interval A1,B1, aR = abs(R);
    real SA, I = Inf(aR),d,d1;
    d1 = mulu(epsa,epsb);
    d = addu(d1,epsa);
    d = addu(d,epsb);   // d:  reproduced relative data error
    SA = Sup( abs(A) );  SA = mulu(SA,epsa);
    A1 = sign(epsa)!=0 ? A+interval(-SA,SA) : A;
    d1 = Sup( abs(B) );  d1 = mulu(d1,epsb);
    B1 = sign(epsb)!=0 ? B+interval(-d1,d1) : B;

    aR = abs(A1*B1);  SA = Sup(aR);
    d1 = mul_exakt(A1,B1,aR);
    if (Inf(aR)>=MinReal) // All products only in the normalized range
    {
	if (d1!=0)
        {
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1);
        }   // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (SA<MinReal) // All products only in the denormalized range U
    {
	if (d1!=0) d1 = divu(MinReal,I); // I=0 --> d1 = <+Infinity>
	eps = addu(d,d1);
    } else // products in normalized and denormalized range --> maximum!:
    {
	if (d1!=0)
	{ SA = divu(MinReal,I);  eps = addu(d,SA); }
	else eps = d;  // eps: relative error in U=(-MinReal,+Minreal)
	if (d1!=0){
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1); // d1:  relative rounding error
	}
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    }
} // rel_mulh4

void rel_divh4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A/B and an upper bound eps of the relative error by 
               // calculating (a/b) on the machine with high accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1,aR;
    bool bla;
    real I,d1,d;
    d1 = Sup( abs(A) );  d1 = mulu(d1,epsa);
    A1 = sign(epsa)!=0 ? A+interval(-d1,d1) : A;
    d1 = Sup( abs(B) );  d1 = mulu(d1,epsb);
    B1 = sign(epsb)!=0 ? B+interval(-d1,d1) : B;

    d = addu(epsa,epsb);
    d1 = subd(1,epsb);
    if (d1<=0 || 0<=B1) 
    {
	cerr << "rel_divh4(...): Relative error doesn't exist." << endl;
	exit(1);
    }
    d = A==0 ? 0 : divu(d,d1);  // d:  reproduced relative data error
    R = A/B;
    aR = abs(R);  I = Inf(aR);
    aR = abs(A1/B1);
    d1 = div_exakt(A1,B1,aR);  bla = ( sign(d1)!=0 ); 
    if (Inf(aR)>=MinReal) // All quotients only in the normalized range
    {
	if (bla)
	{
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1);
        }   // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (Sup(aR)<MinReal) // quotients only in the denormal. range U
    {
	if (bla) d1 = divu(MinReal,I);
	eps = addu(d,d1);
    } else // quotients in normalized and denormalized range --> maximum!:
    {
	if (bla)
	{
	    I  = divu(MinReal,I);
	    eps = addu(d,I);     // eps: relative error in U
	} else eps = d;
	if (bla)
	{
	    d1  = addu(1,d);
	    d1  = mulu(eps_h,d1); // d1:  relative rounding error
	}
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    }
} // rel_divh4

//***************************************************************************
//**************** error bounds of the standard functions *******************
//***************************************************************************

//**************************  Square root ***********************************
void abs_sqrt_abs(const interval& A, const real& dela,
		  interval& R, real& abs)
{
    real r;
    abs = abs_sqrt_(Inf(A),dela);
    r   = abs_sqrt_(Sup(A),dela);
    if (r>abs) abs = r;  // Calculating the maximum;
    R = sqrt(A);         // R: inclusion of sqrt(A);
}

void abs_sqrt_rel(const interval& A, const real& epsa,
		  interval& R, real& abs)
{
    real r,r1;
    interval x;
    if (1<=epsa)
    {
	cerr << "abs_sqrt_rel: absolte error doesn't exist!" << endl;
	exit(1);
    }
    r = subd(1,epsa);
    x = sqrt( interval(r) );  // inclusion of sqrt(1-epsa);
    r = Inf(x);
    times2pown(r,1);    // Fast and exact multiplication of r with 2=2^1
    r1 = divu(epsa,r);  // r1 = epsa/( 2*sqrt(1-epsa) )

    r = addu(1,r1);
    r = mulu(eps_h,r);
    r = addu(r,r1);    // r = (r1 + eps_h*(r1+1))

    R = sqrt(A);
    abs = mulu(Sup(R),r);
}

void rel_sqrt_abs(const interval& A, const real& dela,
		  interval& R, real& rel)
{
    real r=Inf(A),r1;
    interval x;
    if (r<=dela)
    {
	cerr << "rel_sqrt_abs: relative error doesn't exist!" << endl;
	exit(1);
    }
    r = divu(dela,r);  // r = dela/Inf(A)
    r1 = subd(1,r);
    x = sqrt( interval(r1) );
    r1 = Inf(x);
    times2pown(r1,1);
    r1 = divu(r,r1);  // r1 = r/(2*sqrt(1-r))

    r = addu(1,r1);
    r = mulu(eps_h,r);
    rel = addu(r1,r);
    R = sqrt(A);
}

void rel_sqrt_rel(const interval& A, const real& epsa,
		  interval& R, real& rel)
{
    real r,r1;
    interval x;
    if (1<=epsa)
    {
	cerr << "rel_sqrt_rel: relative error doesn't exist!" << endl;
	exit(1);
    }
    r = subd(1,epsa);
    x = sqrt( interval(r) );  // inclusion of sqrt(1-epsa);
    r = Inf(x);
    times2pown(r,1);    // Fast and exact multiplication of r with 2=2^1
    r1 = divu(epsa,r);  // r1 = epsa/( 2*sqrt(1-epsa) )

    r = addu(1,r1);
    r = mulu(eps_h,r);
    rel = addu(r,r1);   // r = (r1 + eps_h*(r1+1))

    R = sqrt(A);
}

//***************************************************************************
//*                       Exponential function                              *
//***************************************************************************

void abs_exp_abs(const interval& A, const real& dela,
		 interval& R, real& abs)
{
    interval z=interval(0),z1,At;
    real SA=Sup(A),r;
    At = A + interval(-dela,dela); // At = A_tilde
    R = exp(A);
    if (sign(dela)!=0)
	if (SA+dela<q_min)
	    if (dela<1) z = interval(dela)*MinReal;
	    else {
		z = interval(dela);
		z = exp( ln(z)+z+SA );
	    }
	else z = dela * exp(interval(dela)+SA); 
    // z: inclusion of e^(Sup(A))*dela*e^(dela)

    if (Sup(At)<q_min) r = MinReal; // Undeflow range
    else if (Inf(At)>=q_min)        // normalized range
    {   
	z1 = exp( interval(dela) )*dela + 1;
	z1 = SA + ln(z1);
	if (Sup(z1)>=709) z1 = exp( z1+ln(interval(eps_exp)) );
	else z1 = eps_exp * exp(z1);
	r = Sup(z1);
    }
    else // function values in underflow AND normalized range
    {   
	z1 = exp( interval(dela) )*dela + 1;
	z1 = SA + ln(z1);
	if (Sup(z1)>=709) z1 = exp( z1+ln(interval(eps_exp)) );
	else z1 = eps_exp * exp(z1);
	r = Sup(z1);
	if (MinReal>r) r = MinReal; // calculating the maximum
    }
    abs = addu(Sup(z),r); 
} // abs_exp_abs

void abs_exp_rel(const interval& A, const real& epsa,
		 interval& R, real& abs)
{
    real r;
    interval z=A;
    r = Sup(z);
    r = mulu(r,epsa);
    abs_exp_abs(A,r,R,abs);
} // abs_exp_rel

void rel_exp_abs(const interval& A, const real& dela,
		 interval& R, real& rel)
{
    interval z,z1,At;
    real r;
    At = A + interval(-dela,dela); // At = A_tilde
    R = exp(A);
    z = dela * exp( interval(dela) );
    r = -Inf(A);
    if (Sup(At)<q_min) // Undeflow range
	if (r>709) z1 = exp( interval(r)+ ln(interval(MinReal))); 
	else z1 = MinReal * exp( interval(r) );
    else 
	if (Inf(At)>=q_min) // Normalized range
	    z1 = (z+1)*eps_exp;
	else { // function values in underflow AND normalized range
	if (r>709) z1 = exp( interval(r)+ln(interval(MinReal)) ); 
	else z1 = MinReal * exp( interval(r) ); // z1: Underflow range
	At = (z+1)*eps_exp; // At: Normalized range
	if (Sup(At)>Sup(z1)) z1 = At; // Calculating the maximum
	}
    rel = addu(Sup(z),Sup(z1));
} // rel_exp_abs

void rel_exp_rel(const interval& A, const real& epsa,
		 interval& R, real& rel)
{
    real r;
    interval At,z,z1;
    R = exp(A);
    At = A * (interval(-epsa,epsa)+1);

    if (Sup(At)<q_min) // Underflow range
    {
	r = abs(Inf(A));   z = interval(r);
	if (r>709) z = exp(z+ln(interval(MinReal)));
	else z = MinReal * exp(z);
	z = z + exp_u(Inf(A),epsa);
    } else
	if (Inf(At)>=q_min) // Normalized range
	{
	    z = exp_v(Inf(A),epsa);
	    z1 = exp_v(Sup(A),epsa);
	    if (Sup(z1)>Sup(z)) z = z1; // Calculating the maximum
	}
	else { // Function values in the undeflow AND normalized range
	r = abs(Inf(A));   z = interval(r);
	if (r>709) z = exp(z+ln(interval(MinReal)));
	else z = MinReal * exp(z);
	z = z + exp_u(Inf(A),epsa);

	z1 = exp_v(q_min,epsa);
	At = exp_v(Sup(A),epsa);
	if (Sup(At)>Sup(z1)) z1 = At; // Calculating the maximum
	if (Sup(z1)>Sup(z)) z = z1;   // Calculating the maximum
	}
    rel = Sup(z);
} // rel_exp_rel

//***************************************************************************
//*                        Logarithm function                               *
//***************************************************************************

void abs_ln_abs(const interval& A, const real& dela,
		interval& R, real& abs)
{
    real r;
    r = subd(Inf(A),dela);
    if (sign(r)<=0) 
    {
	cerr << "abs_ln_abs(): Absolute error doesn't exist!" << endl;
	exit(1);
    }
    R = ln(A);
    abs = ln_abs4Max(Inf(A),dela);
    r   = ln_abs4Max(Sup(A),dela);
    if (r>abs) abs = r;
}

void abs_ln_rel(const interval& A, const real& epsa,
		interval& R, real& bnd)
{
    interval z;
    real r,r1;
    r = subd(1.0,epsa);
    r1 = muld(r,Inf(A)); // r1: lower bound of A-tilde, must be positive!
    if (sign(r)<=0 || sign(r1)<=0) 
    {
	cerr << "abs_ln_rel(): Absolute error doesn't exist!" << endl;
	exit(1);
    }
    z = ln(A);  R = z;
    r = divu(epsa,r);
    z = abs(z);
    r1 = addu(r,Sup(z));
    r1 = mulu(r1,eps_ln);
    bnd = addu(r,r1);
}

void rel_ln_abs(const interval& A, const real& dela,
		interval& R, real& rel)
{
    interval At;
    At = A + interval(-dela,dela);
    real r,r1;
    r = subd(Inf(A),dela);

    if (sign(r)<=0 || dela!=0 && 1<=At) 
    {
	cerr << "rel_ln_abs(): Relative error doesn't exist!" << endl;
	exit(1);
    }
    R = ln(A);
    if (dela==0)
	rel = A==interval(1) ? 0 : eps_ln;
    else {
	At = abs( ln(A) );
	r1 = muld(r,Inf(At));
	if (sign(r1)==0)
	{
	    cerr << "rel_ln_abs(): Relative error doesn't exist!" 
                 << endl;
	    exit(1);
	}
	r1 = divu(dela,r1);
	r = addu(r1,1.0);
	r = mulu(r,eps_ln);
	rel = addu(r1,r);
    }
} // rel_ln_abs

void rel_ln_rel(const interval& A, const real& epsa,
		interval& R, real& rel)
{
    real r,r1;
    r = subd(1.0,epsa);
    interval At;
    At = A * (1+interval(-epsa,epsa));
    if (Inf(At)<=0 || epsa!=0 && (1<=At || r<=0)) 
    {
	cerr << "rel_ln_rel(): Relative error doesn't exist!" << endl;
	exit(1);
    }
    R = ln(A);
    if (epsa==0)
	rel = A==interval(1) ? 0 : eps_ln;
    else {
	At = abs( ln(A) );
	r1 = muld(r,Inf(At));
	if (sign(r1)==0)
	{
	    cerr << "rel_ln_rel(): Relative error doesn't exist!" 
                 << endl;
	    exit(1);
	}
	r1 = divu(epsa,r1);
	r = addu(r1,1.0);
	r = mulu(r,eps_ln);
	rel = addu(r1,r);
    }
}

//***************************************************************************
//*                          Function Ln(1+a)                               *
//***************************************************************************	
void abs_lnp1_abs(const interval& A, const real& dela,
		  interval& R, real& abs)
{
    interval At,D,z=interval(-dela,dela); 
    At = A+z;   D = At+z;  // At = A_tilde
    if (Inf(D)<=-1) 
    {
	cerr << "abs_lnp1_abs(): Absolute error doesn't exist!" 
             << endl;  	exit(1);
    }
    real r,r1;
    if (Disjoint(At,MR_l)) // intersection is empty:
    { 
	abs = lnp1_abs4Max(Inf(A),dela);
	r   = lnp1_abs4Max(Sup(A),dela);
	if (r>abs) abs = r;
    }
    else // intersection not empty:
    {
    // Calculating an absolute upper bound abs only for the exact 
    // function values ln(1+wa) lying in the denormalized range:
	D = At & MR_l;  // D: not empty intersection of At and MR_l;
	r = subd(Inf(D),2*dela);
	r = addd(1,r);  // r>0 is guaranteed by: if (Inf(D)<=-1) { } 
	r = divu(dela,r);
	abs = addu(r,minreal);  // abs: the desired error bound

	if (Inf(At) <= -MinReal)
	{
	    r  = lnp1_abs4Max(Inf(A),dela);
	    r1 = addu(-MinReal,dela);
	    r1 = lnp1_abs4Max(r1,dela);
	    if (r>r1) r1 = r;
	    if (r1>abs) abs = r1;
	}

	if (Sup(At)>MinReal)
	{
	    r  = lnp1_abs4Max(Sup(A),dela);
	    r1 = subd(succ(MinReal),dela);
	    r1 = lnp1_abs4Max(r1,dela);
	    if (r>r1) r1 = r;
	    if (r1>abs) abs = r1;
	}
    } 
    R = lnp1(A); // R: inclusion of ln(1+A);
}

void abs_lnp1_rel(const interval& A, const real& epsa,
		  interval& R, real& abs)
// abs is an upper bound of the absolute error |lnp1(wa)-ln(1+a)| for
// all exact function arguments a in A. epsa is a relative error bound 
// of the erroneous function arguments wa; |wa-a|<=|a|*epsa for all a 
// in A. R is an inclusion of all function values ln(1+a) with a in A.
{
    real one_peps = addu(1,epsa), one_meps = subd(1,epsa);
    interval At = A * interval(one_meps,one_peps); // At: A_tilde
    if (epsa>=1 || Inf(At)<=-1)
    {
	cerr << "abs_lnp1_rel(): Absolute error doesn't exist!" 
             << endl;
	exit(1);
    }
    real r,r1,IA=Inf(A),SA=Sup(A);
    if (Disjoint(At,MR_l)) // empty intersection:
    { 
	if (IA>=0)
	    abs = abslnp1_4Max(SA,epsa,one_meps,one_peps);
	else if (SA<=0)
	    abs = abslnp1_4Max(IA,epsa,one_meps,one_peps);
	else 
	{
	    abs = abslnp1_4Max(IA,epsa,one_meps,one_peps);
	    r   = abslnp1_4Max(SA,epsa,one_meps,one_peps);
	    if (r>abs) abs = r;
	}
    }
    else // no empty intersection:
    {
	interval D = At & MR_l; // D: not empty intersection 
                                // of At and MR_l;
	r = Inf(D);
	r1 = (sign(r)<0) ? divd(r,one_meps) : divd(r,one_peps);
	r1 = Delta_4Max(r1,epsa,one_meps,one_peps); // left value
	r = Sup(D);
	r = (sign(r)<0) ? divd(r,one_peps) : divd(r,one_meps);
	r = Delta_4Max(r,epsa,one_meps,one_peps); // right value
	if (r>r1) r1 = r;  // r1: maximum value
	abs = addu(r1,minreal); // upper bound of the absolute error
                                // for all machine values of D
	if (Inf(At) <= -MinReal)
	{
	    r = abslnp1_4Max(IA,epsa,one_meps,one_peps);
	    if (r>abs) abs = r;
	}

	if (Sup(At)>MinReal)
	{
	    r = abslnp1_4Max(SA,epsa,one_meps,one_peps);
	    if (r>abs) abs = r;
	}
    }
    R = lnp1(A);
}  

void rel_lnp1_abs(const interval& A, const real& dela,
		  interval& R, real& rel)
// rel is an upper bound of the relative error 
// |(lnp1(wa)-ln(1+a))/ln(1+a)| for all exact function arguments a 
// in A. dela is the absolute error bound of the erroneous function 
// arguments wa; |wa-a|<=dela for all a in A.
// R is an inclusion of all function values ln(1+a) with a in A.
{
    interval At,D,z=interval(-dela,dela); 
    At = A+z;  D = At+z;  // At = A_tilde
    if (Inf(D)<=-1 || 0<=D) 
    {
	cerr << "rel_lnp1_abs(): Relative error doesn't exist!" 
             << endl;  	exit(1);
    }
    // Now for any erroneous wa in At the corresponding exact 
    // value a lies in D and it holds:  a <> 0, and 1+a-dela > 0;
    real r,r1,r2,In,Su,one_pe;
    one_pe = addu(1,eps_lnp1);  // (1+eps(f))
    if (Disjoint(At,MR_l)) // intersection is empty:
    {  // ln(1+wa) only in the normalized range;
	In = Inf(At);  Su = Sup(At);
	if (sign(In) > 0)
	{
	    r = subd(In,dela); 
	    r1 = Help_lnp1_1(r,dela);
	    r = mulu(r1,one_pe);
	    rel = addu(r,eps_lnp1);
	}
	else // Sup(At)<0:
	{
	    r = addu(Su,dela); 
	    r1 = Help_lnp1_1(r,dela);
	    r = mulu(r1,one_pe);
	    rel = addu(r,eps_lnp1);

	    r = subd(In,dela); // r: smallest possible argument a
	    r1 = Help_lnp1_1(r,dela);
	    r = mulu(r1,one_pe);
	    r1 = addu(r,eps_lnp1);
	    if (r1>rel) rel = r1;
	}
    }
    else // intersection not empty:
    {
    // Calculating an relative upper bound rel only for the exact 
    // function values ln(1+wa) lying in the denormalized range:
	D = At & MR_l;  // D: not empty intersection of At and MR_l;
	In = Inf(D);  Su = Sup(D); 
	if (sign(In)>0) 
	{ // eps_dat,f + minreal/ln(1+a) is monotonic decreasing
	    r = subd(In,dela);
            // r: smallest possible exact function argument;
	    r1 = Help_lnp1_1(r,dela);
	    z = lnp1( interval(r) );
	    r2 = Inf(z);
	    r2 = divu(minreal,r2);
	    rel = addu(r1,r2); // rel: upper bound if Inf(D)>0;

	    if (Sup(At)>MinReal) 
	    {
		r = subd(succ(MinReal),dela);
		r1 = Help_lnp1_1(r,dela);
		r = mulu(r1,one_pe);
		r1 = addu(r,eps_lnp1);
		if (r1>rel) rel = r1;
	    }
	}
	else  
	{
	    r = addu(Su,dela); // r: greatest possible argument;
	    r1 = Help_lnp1_1(r,dela);  
	    z = abs( lnp1(interval(r)) );
	    r2 = Inf(z);   
	    r2 = divu(minreal,r2);
	    rel = addu(r1,r2);  // rel: right upper bound;

	    r = subd(In,dela);  
            // r: smallest possible exact function argument;
	    r1 = Help_lnp1_1(r,dela); 
	    z = abs( lnp1(interval(r)) );
	    r2 = Inf(z);  
	    r2 = divu(minreal,r2);
	    r2 = addu(r1,r2);  // r2: left upper bound; 
	    if (r2>rel) rel = r2;

	    In = Inf(At);
	    if (In <= -MinReal)
	    {
		r1 = Help_lnp1_1(In,dela);
		r = mulu(r1,one_pe);
		r1 = addu(r,eps_lnp1);

		r = addu(-MinReal,dela);
		r2 = Help_lnp1_1(r,dela);
		r = mulu(r2,one_pe);
		r2 = addu(r,eps_lnp1);
		if (r2>r1) r1 = r2;
		if (r1>rel) rel = r1;
	    }
	}
    } // else
    R = lnp1(A);
}

void rel_lnp1_rel(const interval& A, const real& epsa,
		  interval& R, real& rel)
// rel is an upper bound of the relative error 
// |(lnp1(wa)-ln(1+a))/ln(1+a)| for all exact function arguments a 
// in A. epsa is the relative error bound of the erroneous function 
// arguments wa; |wa-a|<=|a|*epsa for all a in A.
// R is an inclusion of all function values ln(1+a) with a in A.
{
    real one_peps = addu(1,epsa), 
         one_meps = subd(1,epsa),r,r1,abs_a,ar;
    interval At=A*interval(one_meps,one_peps); // At: A_tilde
    if (epsa>0.5 || Inf(At)<=-1 || Disjoint(At,MR_l)==0)
    {
	cerr << "rel_lnp1_rel(): Relative error doesn't exist!" 
             << endl;
	exit(1);
    }
    // All function values ln(1+wa) lie in the normalized range
    r = addu(1,eps_lnp1);  
    r = mulu(r,epsa);  // r = epsa*(1+eps_lnp1)
    ar = Inf(A);
    if (sign(ar)>0) { r1 = one_meps;  abs_a = ar; }
    else { r1 = one_peps;  abs_a = -ar; }

    r = mulu(r,abs_a); // r: numerator
    r1 = muld(ar,r1);
    r1 = addd(1,r1);
    At = abs( lnp1(interval(ar)) );
    ar = Inf(At);
    ar = muld(r1,ar);
    ar = divu(r,ar); 
    rel = addu(ar,eps_lnp1); // relative error bound
    R = lnp1(A);
} // rel_lnp1_rel

void rel_acosh_abs(const interval& A, const real& dela,
		  interval& R, real& rel)
{
    real r(Inf(A)),edat;
    interval b;
    if (r<=1)
    {
	cerr << "rel_acosh_abs: relative error doesn't exist!" << endl;
	exit(1);
    } // Now it holds:  Inf(A) > 1;
    b = interval(r) - dela;
    edat = Sup( dela / ( sqrtx2m1(b)*acosh(interval(r)) ) );
    r = addu(1.0,edat);
    r = mulu(r,eps_acosh);
    rel = addu(edat,r); 
    R = acosh(A);
}

void rel_acosh_rel(const interval& A, const real& epsa,
		  interval& R, real& rel)
{
    real r(Inf(A)),edat;
    interval b,a1;
    if (r<=1)
    {
	cerr << "rel_acosh_rel: relative error doesn't exist!" << endl;
	exit(1);
    } // Now it holds:  Inf(A) > 1;
    a1 = interval(r);
    b = a1 * (interval(1.0)-epsa);
    edat = Sup( a1*epsa / (sqrtx2m1(b)*acosh(a1)) ); 
    r = addu(1.0,edat);
    r = mulu(r,eps_acosh);
    rel = addu(edat,r); 
    R = acosh(A);
}

//****************************************************************************
//*               Some functions for a simple error calculus                 *
//****************************************************************************

real eps_2(const real& eps_a, const real& eps_b)
    // returns an upperbound for |eps_ab| with 
    // (1+eps_a)*(1+eps_b) = 1+eps_ab
{
    real t,res,Eps_a(eps_a),Eps_b(eps_b);
    if (Eps_a > Eps_b) {
	t = Eps_a; Eps_a = Eps_b; Eps_b = t; 
    };      
    t = addu(1.0,Eps_b);
    res = mulu(Eps_a,t);
    res = addu(res,Eps_b);
    return res;
}

real eps_3(const real& eps_a, const real& eps_b, const real& eps_k)
    // returns an upperbound for |eps_abk| with 
    // (1+eps_a)*(1+eps_b)*(1+eps_k) = 1+eps_abk
{
    real eps,res;
    eps = eps_2(eps_a,eps_b);
    res = eps_2(eps,eps_k);
    return res;
}

real eps_4(const real& eps_a, const real& eps_b,
           const real& eps_c, const real& eps_d)
    // returns an upperbound for |eps_abcd| with 
    // (1+eps_a)*(1+eps_b)*(1+eps_c)*(1+eps_c) = 1+eps_abcd
{
    real t1,t2,res;
    t1 = eps_2(eps_a,eps_b);
    t2 = eps_2(eps_c,eps_d);
    res = eps_2(t1,t2);
    return res;
}

real eps_pow(const real& eps, int n) throw()
    // returns an upperbound for |eps_p| with 
    // (1+eps)^n = (1+eps_p), n >= 1;
{
    real ep;
    if (n<1) 
	cxscthrow(STD_FKT_OUT_OF_DEF("real eps_pow(const real& eps; int n)"));
    if (n==1)  ep = eps;
    else 
    {          
	ep = eps_2(eps,eps); 
	n = n-2;
	for (int k=1; k<=n; k++) ep = eps_2(ep,eps);
    }
    return ep;
}

real eps_rezi(const real& eps) throw()
    // 1/(1+eps) = 1+eps_1; eps<=0.5 !
{
    real t,res;
    if (eps>0.5) 
	cxscthrow(STD_FKT_OUT_OF_DEF("real eps_rezi(const real& eps)")); 
    t = addu(1.0,2*eps);
    res = mulu(eps,t);
    return res;
}

real eps_ZdN(const real& eps_Z, const real& eps_N)
    // Error bound of a fraction; 
    // The rel. error bound for the division is used with eps_h = 2^{-52} 
    // for only high accuracy! 
{
    real eps,res;
    eps = eps_rezi(eps_N);
    res = eps_3(eps_Z,eps,eps_h);
    return res;
}

} // namespace cxsc











