/*
**  CXSC is a C++ library for eXtended Scientific Computing
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2001 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "abs_relm.hpp"
#include <interval.hpp>

using namespace std;
namespace cxsc {

static const real eps_m = comp(0.5,-52);   // eps_m = 2^(-53) 
                                           // for maximum accuracy
    static real ForMaxS4(const real& epsa, const real& epsb,
                         const real& a, const real& b)
    // (|a|*epsa + |b|*epsb)/|a+b|
    {
	real d1=0, d2;
	if (sign(epsa)!=0 || sign(epsb)!=0)
	{
	    d1 = mulu(abs(a),epsa), d2 = mulu(abs(b),epsb);
	    d1 = addu(d1,d2);  // Numerator ready
	    interval ai = interval(a), bi = interval(b);
	    ai = abs(ai+bi);   // '+' because of sum
	    d2 = Inf(ai);      // Denominator ready
	    if (d2<=0)
	    {
	        cerr << "Realtive error doesn't exist!" << endl;
	        exit(1);
	    } 
	    d1 = divu(d1,d2);
	}
	return d1;
    }

    static real ForMaxD4(const real& epsa, const real& epsb,
                         const real& a, const real& b)
    // (|a|*epsa + |b|*epsb)/|a-b|
    {
	real d1=0, d2;
	if (sign(epsa)!=0 || sign(epsb)!=0)
	{
	    d1 = mulu(abs(a),epsa), d2 = mulu(abs(b),epsb);
	    d1 = addu(d1,d2);  // Numerator ready
	    interval ai = interval(a), bi = interval(b);
	    ai = abs(ai-bi);   // '-' because of difference
	    d2 = Inf(ai);      // Denominator ready
	    if (d2<=0)
	    {
	        cerr << "Realtive error doesn't exist!" << endl;
	        exit(1);
	    } 
	    d1 = divu(d1,d2);
	}
	return d1;
    }

    static real ForMaxS2(const real& dela,  const real& epsb,
                         const interval& A, const real& b)
    // (dela + |b|*epsb)/Inf(|A+b|)
    {
	real d1=0, d2;
	if (sign(dela)!=0 || sign(epsb)!=0)
	{
	    d2 = mulu(abs(b),epsb);
	    d1 = addu(dela,d2);  // Numerator ready
	    interval bi = interval(b);
	    bi = abs(A+bi);     // '+' because of sum
	    d2 = Inf(bi);       // Denominator ready
	    if (d2<=0)
	    {
	        cerr << "Realtive error doesn't exist!" << endl;
	        exit(1);
	    } 
	    d1 = divu(d1,d2);
	}
	return d1;
    }

    static real ForMaxD2(const real& dela,  const real& epsb,
                         const interval& A, const real& b)
    // (dela + |b|*epsb)/Inf(|A-b|)
    {
	real d1=0, d2;
	if (sign(dela)!=0 || sign(epsb)!=0)
	{
	    d2 = mulu(abs(b),epsb);
	    d1 = addu(dela,d2);  // Numerator ready
	    interval bi = interval(b);
	    bi = abs(A-bi);     // '-' because of difference
	    d2 = Inf(bi);       // Denominator ready
	    if (d2<=0)
	    {
	        cerr << "Realtive error doesn't exist!" << endl;
	        exit(1);
	    } 
	    d1 = divu(d1,d2);
	}
	return d1;
    }

    static real ForMaxS3(const real& epsa,  const real& delb,
                         const interval& B, const real& a)
    // (|a|*epsa + delb)/Inf(|a+B|)
    {
	real d1=0, d2;
	if (sign(epsa)!=0 || sign(delb)!=0)
	{
	    d2 = mulu(abs(a),epsa);
	    d1 = addu(delb,d2);  // Numerator ready
	    interval ai = interval(a);
	    ai = abs(ai+B);     // '+' because of sum
	    d2 = Inf(ai);       // Denominator ready
	    if (d2<=0)
	    {
	        cerr << "Realtive error doesn't exist!" << endl;
	        exit(1);
	    } 
	    d1 = divu(d1,d2);
	}
	return d1;
    }

    static real ForMaxD3(const real& epsa,  const real& delb,
                         const interval& B, const real& a)
    // (|a|*epsa + delb)/Inf(|a-B|)
    {
	real d1=0, d2;
	if (sign(epsa)!=0 || sign(delb)!=0)
	{
	    d2 = mulu(abs(a),epsa);
	    d1 = addu(delb,d2);  // Numerator ready
	    interval ai = interval(a);
	    ai = abs(ai-B);     // '-' because of difference
	    d2 = Inf(ai);       // Denominator ready
	    if (d2<=0)
	    {
	        cerr << "Realtive error doesn't exist!" << endl;
	        exit(1);
	    } 
	    d1 = divu(d1,d2);
	}
	return d1;
    }

// ---------------------------------------------------------------------------
// -------------------- Calculating the absolur errors -----------------------
// ---------------------------------------------------------------------------
// ------------------------- Delta(a), Delta(b) ------------------------------
// ---------------------------------------------------------------------------

void abs_addm1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A+B and an upper bound del of the absolute error by 
               // calculating (a+b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    del = addu(dela,delb); // del = reproduced data error
    R=A+B;
    bool bl1,bl2;
    interval A1,B1;
    bl1 = (A!=0 || dela!=0) && (B!=0 || delb!=0);
    A1 = A+interval(-dela,dela);
    B1 = B+interval(-delb,delb);
    bl2 = A1/B1 <= interval(-2,-0.5); 
    if (bl1 && !bl2) 
    {
	real r = Sup( abs(R) );
        if (r>=MinReal) // Results in the normalized range
        {
	    r = addu(r,del);
	    r = mulu(r,eps_m);  // r: absolute rounding error
	    del = addu(del,r);
        }
    }
} // abs_addm1
     
void abs_subm1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A-B and an upper bound del of the absolute error by 
               // calculating (a-b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1;
    bool bl1,bl2;
    del = addu(dela,delb); // del = reproduced data error
    R=A-B;
    bl1 = (A!=0 || dela!=0) && (B!=0 || delb!=0);
    A1 = A + interval(-dela,dela);
    B1 = B + interval(-delb,delb);
    bl2 = A1/B1 <= interval(0.5,2);
    if (bl1 && !bl2)
    {
	real r = Sup( abs(R) );
        if (r>=MinReal) // Results in the normalized range
        {
	    r = addu(r,del);
	    r = mulu(r,eps_m);  // r: absolute rounding error
	    del = addu(del,r);
        } 
    }
} // abs_subm1

void abs_mulm1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A*B and an upper bound del of the absolute error by 
               // calculating (a*b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    real d = mulu(dela,delb);
    real SA=Sup( abs(A) ), SB=Sup( abs(B) );
    real d1 = mulu(SA,delb), d2 = mulu(SB,dela);
    d = addu(d,d1); 
    d = addu(d,d2);  // d: reproduced data error

    R=A*B;
    interval aR = abs(R);
    real S = Sup(aR), I = Inf(aR);

    if (I>=MinReal) // Products only in the normalized range
    {
	if (dela==0 && diam(A)==0 && mant(SA)==0.5) d1=0;
        else
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_m,d1);
        }   // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (S<MinReal) // Products only in the denormalized range U
    {
	del = addu(d,minreal);
    } else // Products in normalized and denormalized range --> maximum!:
    {
	del = addu(d,minreal); // del: absolute error for the underflow range
        d1 = addu(d,S);
	d1 = mulu(eps_m,d1);  // d1:  absolute rounding error
	d2 = addu(d,d1);      // d2: absolute error for the normalized range
	if (d2>del) del = d2;
    }
} // abs_mulm1

void abs_divm1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A/B and an upper bound del of the absolute error by 
               // calculating (a/b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    real SA=Sup( abs(A) ), IB=Inf( abs(B) ), d=0, d1,d2;
    if (sign(dela)!=0 || sign(delb)!=0)
    {
	d1 = mulu(SA,delb);
        d1 = divu(d1,IB);
        d1 = addu(d1,dela);       // numerator ready
        d2 = subd(IB,delb);  // denominator ready
        if (d2<=0) 
        {
	    cerr << "abs_divm1(...): Absolute error doesn't exist." << endl;
	    exit(1);
        }
        d = divu(d1,d2);  // d: reproduced data error
    }
    R=A/B;
    interval aR = abs(R);
    real S = Sup(aR), I = Inf(aR);

    if (I>=MinReal) // All quotients only in the normalized range
    {
	if (delb==0 && diam(B)==0 && mant(IB)==0.5) d1=0;
        else
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_m,d1);
	}   // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (S<MinReal) // Quotients only in the denormalized range U
    {
	del = addu(d,minreal);
    } else // Quotients in normalized and denormalized range --> maximum!:
    {
	del = addu(d,minreal); // del: absolute error for the underflow range
        d1 = addu(d,S);
	d1 = mulu(eps_m,d1);  // d1:  absolute rounding error
	d2 = addu(d,d1);      // d2: absolute error for the normalized range
	if (d2>del) del = d2; // calculating the maximum 
    }
} // abs_divm1

// ---------------------------------------------------------------------------
// --------------------------- Delta(a), Eps(b) ------------------------------
// --------------------------------------------------------------------------- 

void abs_addm2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A+B and an upper bound del of the absolute error by 
               // calculating (a+b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1;
    real r;
    r = Sup( abs(B) );
    r = mulu(r,epsb);
    del = addu(dela,r);  // del = reproduced data error
    R=A+B;
    bool bl1,bl2;
    bl1 = (A!=0 || dela!=0) && (B!=0 || epsb!=0);
    A1 = A+interval(-dela,dela);
    B1 = B + interval(-r,r);
    bl2 = A1/B1 <= interval(-2,-0.5);
    if (bl1 && !bl2)
    {
    r = Sup( abs(R) );
    if (r>=MinReal) // sums in the normalized range
    {
	r = addu(r,del);
	r = mulu(r,eps_m);  // r: absolute rounding error
	del = addu(del,r);
    } }
} // abs_addm2
     
void abs_subm2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A-B and an upper bound del of the absolute error by 
               // calculating (a-b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1;
    real r;
    bool bl1,bl2;
    r = Sup( abs(B) );  
    r = mulu(r,epsb);
    del = addu(dela,r);  // del = reproduced data error

    R=A-B;
    bl1 = (A!=0 || dela!=0) && (B!=0 || epsb!=0);
    A1 = A + interval(-dela,dela);
    B1 = B + interval(-r,r);
    bl2 = A1/B1 <= interval(0.5,2);
    if (bl1 && !bl2)
    {
	r = Sup( abs(R) );
        if (r>=MinReal) // Differences in the normalized range
        {
	    r = addu(r,del);
	    r = mulu(r,eps_m);  // r: absolute rounding error
	    del = addu(del,r);
        } 
    }
} // abs_subm2

void abs_mulm2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A*B and an upper bound del of the absolute error by 
               // calculating (a*b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    real d = mulu(dela,epsb), d1,d2;
    real SA=Sup( abs(A) ), SB=Sup( abs(B) );
    d1 = mulu(SA,epsb);
    d = addu(d,d1); 
    d = addu(d,dela);
    d = mulu(SB,d);  // d: reproduced data error

    R=A*B;
    interval aR = abs(R);
    real S = Sup(aR), I = Inf(aR);

    if (I>=MinReal) // Products only in the normalized range
    {
	if (dela==0 && diam(A)==0 && mant(SA)==0.5) d1=0;
        else
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_m,d1);
	}   // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (S<MinReal) // Products only in the denormalized range U
    {
	del = addu(d,minreal);
    } else // Products in normalized and denormalized range --> maximum!:
    {
	del = addu(d,minreal); // del: absolute error for the underflow range
        d1 = addu(d,S);
	d1 = mulu(eps_m,d1);  // d1:  absolute rounding error
	d2 = addu(d,d1);      // d2: absolute error for the normalized range
	if (d2>del) del = d2;
    }
} // abs_mulm2

void abs_divm2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A/B and an upper bound del of the absolute error by 
               // calculating (a/b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    real IB=Inf( abs(B) ), d=0,d1,d2;
    if (sign(dela)!=0 || sign(epsb)!=0) 
    {
	real SA=Sup( abs(A) );
        d1 = mulu(SA,epsb);
        d1 = addu(d1,dela);  // numerator ready
        d2 = subd(1,epsb);   
	d2 = muld(IB,d2);    // denominator ready
        if (d2<=0) 
        {   
	    cerr << "abs_divm2(...): Absolute error doesn't exist." << endl;
	    exit(1);
        }
        d = divu(d1,d2);  // d: reproduced data error
    }


    R=A/B;
    interval aR = abs(R);
    real S = Sup(aR), I = Inf(aR);

    if (I>=MinReal) // All quotients only in the normalized range
    {
	if (epsb==0 && diam(B)==0 && mant(IB)==0.5) d1=0;
        else
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_m,d1);
	}   // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (S<MinReal) // Quotients only in the denormalized range U
    {
	del = addu(d,minreal);
    } else // Quotients in normalized and denormalized range --> maximum!:
    {
	del = addu(d,minreal); // del: absolute error for the underflow range
        d1 = addu(d,S);
	d1 = mulu(eps_m,d1);  // d1:  absolute rounding error
	d2 = addu(d,d1);      // d2: absolute error for the normalized range
	if (d2>del) del = d2; // calculating the maximum 
    }
} // abs_divm2

// ---------------------------------------------------------------------------
// --------------------------- Eps(a), Delta(b) ------------------------------
// --------------------------------------------------------------------------- 

void abs_addm3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A+B and an upper bound del of the absolute error by 
               // calculating (a+b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1;
    real r;
    bool bl1,bl2;
    r = Sup( abs(A) );  
    r = mulu(r,epsa);
    del = addu(delb,r);  // del = reproduced data error
    R=A+B;
    bl1 = (A!=0 || epsa!=0) && (B!=0 || delb!=0);
    A1 = A + interval(-r,r);
    B1 = B + interval(-delb,delb);
    bl2 = A1/B1 <= interval(-2,-0.5); 
    if (bl1 && !bl2)
    {
        r = Sup( abs(R) );
        if (r>=MinReal) // sums in the normalized range
        {
	    r = addu(r,del);
	    r = mulu(r,eps_m);  // r: absolute rounding error
	    del = addu(del,r);
        } 
    }
} // abs_addm3

void abs_subm3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A-B and an upper bound del of the absolute error by 
               // calculating (a-b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1;
    real r;
    bool bl1,bl2;
    r = Sup( abs(A) );  
    r = mulu(r,epsa);
    del = addu(delb,r);  // del = reproduced data error

    R=A-B;
    bl1 = (A!=0 || epsa!=0) && (B!=0 || delb!=0);
    A1 = A + interval(-r,r);
    B1 = B + interval(-delb,delb);
    bl2 = A1/B1 <= interval(0.5,2);
    if (bl1 && !bl2)
    {
	r = Sup( abs(R) );
        if (r>=MinReal) // Differences in the normalized range
        {
	    r = addu(r,del);
	    r = mulu(r,eps_m);  // r: absolute rounding error
	    del = addu(del,r);
        } 
    }
} // abs_subm3

void abs_mulm3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A*B and an upper bound del of the absolute error by 
               // calculating (a*b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    real d = mulu(epsa,delb), d1,d2;
    real SA=Sup( abs(A) ), SB=Sup( abs(B) );
    d1 = mulu(SB,epsa);
    d = addu(d,d1); 
    d = addu(d,delb);
    d = mulu(SA,d);  // d: reproduced data error

    R=A*B;
    interval aR = abs(R);
    real S = Sup(aR), I = Inf(aR);

    if (I>=MinReal) // Products only in the normalized range
    {
	if (epsa==0 && diam(A)==0 && mant(SA)==0.5) d1=0;
        else
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_m,d1); 
	}    // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (S<MinReal) // Products only in the denormalized range U
    {
	del = addu(d,minreal);
    } else // Products in normalized and denormalized range --> maximum!:
    {
	del = addu(d,minreal); // del: absolute error for the underflow range
        d1 = addu(d,S);
	d1 = mulu(eps_m,d1);  // d1:  absolute rounding error
	d2 = addu(d,d1);      // d2: absolute error for the normalized range
	if (d2>del) del = d2;
    }
} // abs_mulm3

void abs_divm3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A/B and an upper bound del of the absolute error by 
               // calculating (a/b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    real IB=Inf( abs(B) ), d=0,d1,d2;
    if (sign(epsa)!=0 || sign(delb)!=0) 
    {
	real SA=Sup( abs(A) );
        d1 = divu(delb,IB);
        d1 = addu(d1,epsa);  // d1: numerator ready
        d2 = subd(IB,delb);  // d2: denominator ready 
        if (d2<=0) 
        {   
	    cerr << "abs_divm3(...): Absolute error doesn't exist." << endl;
	    exit(1);
        }
        d = divu(d1,d2); 
        d = mulu(d,SA);  // d: reproduced data error
    }

    R=A/B;
    interval aR = abs(R);
    real S = Sup(aR), I = Inf(aR);

    if (I>=MinReal) // All quotients only in the normalized range
    {
	if (delb==0 && diam(B)==0 && mant(IB)==0.5) d1=0;
        else
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_m,d1); 
	}   // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (S<MinReal) // Quotients only in the denormalized range U
    {
	del = addu(d,minreal);
    } else // Quotients in normalized and denormalized range --> maximum!:
    {
	del = addu(d,minreal); // del: absolute error for the underflow range
        d1 = addu(d,S);
	d1 = mulu(eps_m,d1);  // d1:  absolute rounding error
	d1 = addu(d,d1);      // d1:  absolute error for the normalized range
	if (d1>del) del = d1; // calculating the maximum 
    }
} // abs_divm3

// ---------------------------------------------------------------------------
// ---------------------------- Eps(a), Eps(b) -------------------------------
// ---------------------------------------------------------------------------

void abs_addm4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A+B and an upper bound del of the absolute error by 
               // calculating (a+b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1;
    real SA = Sup( abs(A) ), SB = Sup( abs(B) );
    real r = mulu(SA,epsa), r1 = mulu(SB,epsb);
    del = addu(r1,r);  // del = reproduced data error
    R=A+B;
    bool bl1,bl2;
    bl1 = (A!=0 || epsa!=0) && (B!=0 || epsb!=0);
    A1 = A + interval(-r,r);
    B1 = B + interval(-r1,r1);
    bl2 = A1/B1 <= interval(-2,-0.5);
    if (bl1 && !bl2)
    {
	r = Sup( abs(R) );
        if (r>=MinReal) // Some sums in the normalized range
        {
	    r = addu(r,del);
	    r = mulu(r,eps_m);  // r: absolute rounding error
	    del = addu(del,r);
        } 
    }
} // abs_addm4

void abs_subm4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A-B and an upper bound del of the absolute error by 
               // calculating (a-b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1;
    bool bl1,bl2;
    real SA = Sup( abs(A) ), SB = Sup( abs(B) );
    real r = mulu(SA,epsa), r1 = mulu(SB,epsb);
    del = addu(r1,r);  // del = reproduced data error
    R=A-B;
    bl1 = (A!=0 || epsa!=0) && (B!=0 || epsb!=0);
    A1 = A + interval(-r,r);
    B1 = B + interval(-r1,r1);
    bl2 = A1/B1 <= interval(0.5,2);
    if (bl1 && !bl2)
    {
	r = Sup( abs(R) );
        if (r>=MinReal) // Some sums in the normalized range
        {
	    r = addu(r,del);
	    r = mulu(r,eps_m);  // r: absolute rounding error
	    del = addu(del,r);
        } 
    }
} // abs_subm4

void abs_mulm4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A*B and an upper bound del of the absolute error by 
               // calculating (a*b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    real d = mulu(epsa,epsb);
    d = addu(d,epsb);
    d = addu(d,epsa);
    real SA=Sup( abs(A) ), SB=Sup( abs(B) ), d1;
    d = mulu(SB,d);
    d = mulu(SA,d);  // d: reproduced data error

    R=A*B;
    interval aR = abs(R);
    real S = Sup(aR), I = Inf(aR);

    if (I>=MinReal) // Products only in the normalized range
    {
	if (epsa==0 && diam(A)==0 && mant(SA)==0.5) d1=0;
        else
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_m,d1); 
	}   // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (S<MinReal) // Products only in the denormalized range U
    {
	del = addu(d,minreal);
    } else // Products in normalized and denormalized range --> maximum!:
    {
	del = addu(d,minreal); // del: absolute error for the underflow range
        d1 = addu(d,S);
	d1 = mulu(eps_m,d1);  // d1: absolute rounding error
	d1 = addu(d,d1);      // d1: absolute error for the normalized range
	if (d1>del) del = d1; // calculating the maximum
    }
} // abs_mulm4

void abs_divm4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& del) throw()  // Returns the interval 
               // R=A/B and an upper bound del of the absolute error by 
               // calculating (a/b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    real SA=Sup( abs(A) ), IB=Inf( abs(B) ), d=0, d1,d2;
    if (sign(epsa)!=0 || sign(epsb)!=0)
    {
	d1 = addu(epsa,epsb);
        d1 = mulu(d1,SA);   // numerator ready
        d2 = subd(1,epsb);
        d2 = muld(d2,IB);  
        if (d2<=0) 
        {
	    cerr << "abs_divm4(...): Absolute error doesn't exist." << endl;
	    exit(1);
        }
        d = divu(d1,d2);  // d: reproduced data error
    }
    R=A/B;
    interval aR = abs(R);
    real S = Sup(aR), I = Inf(aR);

    if (I>=MinReal) // All quotients only in the normalized range
    {
	if (epsb==0 && diam(B)==0 && mant(IB)==0.5) d1=0;
        else
	{
	    d1 = addu(d,S);
	    d1 = mulu(eps_m,d1);
	}   // d1:  absolute rounding error
	del = addu(d,d1);
    } else if (S<MinReal) // Quotients only in the denormalized range U
    {
	del = addu(d,minreal);
    } else // Quotients in normalized and denormalized range --> maximum!:
    {
	del = addu(d,minreal); // del: absolute error for the underflow range
        d1 = addu(d,S);
	d1 = mulu(eps_m,d1);  // d1:  absolute rounding error
	d2 = addu(d,d1);      // d2: absolute error for the normalized range
	if (d2>del) del = d2; // calculating the maximum 
    }
} // abs_divm4


// ---------------------------------------------------------------------------
// ------------------- Calculating the relative errors -----------------------
// ---------------------------------------------------------------------------
// -------------------------- Delta(a), Delta(b) -----------------------------
// ---------------------------------------------------------------------------

void rel_addm1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A+B and an upper bound eps of the relative error by 
               // calculating (a+b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1;
    bool bl1,bl2;
    eps = 0;
    R = A+B;
    interval aR = abs(R);
    real S = Sup(aR), I = Inf(aR),d;
    if (sign(dela)!=0 || sign(delb)!=0)
    {
	d = addu(dela,delb);  // d != 0  
        if (sign(I)==0)
        {
	    cerr << "rel_addm1(...): Relative error doesn't exist." << endl;
	    exit(1);
        }
        eps = divu(d,I);  // eps: reproduced relative data error
    }
    bl1 = (A!=0 || dela!=0) && (B!=0 || delb!=0);
    A1 = A + interval(-dela,dela);
    B1 = B + interval(-delb,delb);
    bl2 = A1/B1 <= interval(-2,-0.5);
    if (bl1 && !bl2)
        if (S>=MinReal) // Some sums are in the normalized range
        {
	    d   = addu(1,eps);
	    d   = mulu(eps_m,d);  // d: relative rounding error
	    eps = addu(d,eps);
        }
} // rel_addm1

void rel_subm1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A-B and an upper bound rel of the relative error by 
               // calculating (a-b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1;
    bool bl1,bl2;
    eps = 0;
    R = A-B;
    interval aR = abs(R);
    real S = Sup(aR), I = Inf(aR),d;
    if (sign(dela)!=0 || sign(delb)!=0)
    {
	d = addu(dela,delb); // d != 0 
        if (sign(I)==0)
        {
	    cerr << "rel_subm1(...): Relative error doesn't exist." << endl;
	    exit(1);
        }
        eps = divu(d,I);  // eps: reproduced relative data error
    }
    bl1 = (A!=0 || dela!=0) && (B!=0 || delb!=0);
    A1 = A + interval(-dela,dela);
    B1 = B + interval(-delb,delb);
    bl2 = A1/B1 <= interval(0.5,2);
    if (bl1 && !bl2)
        if (S>=MinReal) // Some differences are in the normalized range
        {
	    d   = addu(1,eps);
	    d   = mulu(eps_m,d);  // d: relative rounding error
	    eps = addu(d,eps);
        }
} // rel_subm1

void rel_mulm1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A*B and an upper bound eps of the relative error by 
               // calculating (a*b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    R = A*B;
    interval aR = abs(R);
    real S = Sup(aR), I = Inf(aR), IA = Inf( abs(A) ), d=0,d1;
    if (sign(dela)!=0 || sign(delb)!=0)
    {
	real IB = Inf( abs(B) );
        d = divu(dela,IA);
        d1 = divu(delb,IB);
        IA = addu(d1,d);
        d = mulu(d,d1);
        d = addu(d,IA); // d:  reproduced relative data error
    }

    if (I>=MinReal) // All products only in the normalized range
    {
	if (dela==0 && diam(A)==0 && mant(IA)==0.5) d1=0;
        else
	{
	    d1  = addu(1,d);
	    d1  = mulu(eps_m,d1);
	}    // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (S<MinReal) // All products only in the denormalized range U
    {
	d1 = divu(minreal,I);
	eps = addu(d,d1);
    } else // products in normalized and denormalized range --> maximum!:
    {
	d1  = divu(minreal,I);
	eps = addu(d,d1);     // eps: relative error in U
	d1  = addu(1,d);
	d1  = mulu(eps_m,d1); // d1:  relative rounding error
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    }
} // rel_mulm1

void rel_divm1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A/B and an upper bound eps of the relative error by 
               // calculating (a/b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    R = A/B;
    interval aR = abs(R);
    real S = Sup(aR), I = Inf(aR), IB = Inf( abs(B) ), d=0, d1;
    if (sign(dela)!=0 || sign(delb)!=0)
    {
	real IA = Inf( abs(A) );
        d = divu(dela,IA);
        d1 = divu(delb,IB);
        d = addu(d,d1);
        d1 = subd(1,d1);
        if (d1<=0) 
        {
	    cerr << "rel_divm1(...): Absolute error doesn't exist." << endl;
	    exit(1);
        }
        d = divu(d,d1);  // d:  reproduced relative data error
    }
    if (I>=MinReal) // All quotients only in the normalized range
    {
	if (delb==0 && diam(B)==0 && mant(IB)==0.5) d1=0;
        else
	{
	    d1 = addu(1,d);
	    d1 = mulu(eps_m,d1);
	}   // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (S<MinReal) // All quotients only in the denormalized range U
    {
	d1 = divu(minreal,I);
	eps = addu(d,d1);
    } else // quotients in normalized and denormalized range --> maximum!:
    {
	d1  = divu(minreal,I);
	eps = addu(d,d1);     // eps: relative error in U
	d1  = addu(1,d);
	d1  = mulu(eps_m,d1); // d1:  relative rounding error
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    }
} // rel_divm1

// ---------------------------------------------------------------------------
// ----------------------------- Delta(a), Eps(b) ----------------------------
// ---------------------------------------------------------------------------

void rel_addm2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A+B and an upper bound eps of the relative error by 
               // calculating (a+b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1;
    R = A+B;
    real S = Sup( abs(R) ), d;
    eps  = ForMaxS2(dela,epsb,A,Inf(B));
    d    = ForMaxS2(dela,epsb,A,Sup(B));
    if (d>eps) eps = d;  // eps: reproduced relative data error
    bool bl1,bl2;
    bl1 = (A!=0 || dela!=0) && (B!=0 || epsb!=0);
    A1 = A + interval(-dela,dela);
    d = Sup( abs(B) );
    d = mulu(d,epsb);
    B1 = B + interval(-d,d);
    bl2 = A1/B1 <= interval(-2,-0.5);
    if (bl1 && !bl2)
        if (S>=MinReal) // Some sums are in the normalized range
        {
	    d   = addu(1,eps);
	    d   = mulu(eps_m,d);  // d: relative rounding error
	    eps = addu(d,eps);
        }
} // rel_addm2

void rel_subm2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A-B and an upper bound eps of the relative error by 
               // calculating (a-b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1;
    bool bl1,bl2;
    R = A-B;
    real S = Sup( abs(R) ), d;

    eps  = ForMaxD2(dela,epsb,A,Inf(B));
    d    = ForMaxD2(dela,epsb,A,Sup(B));
    if (d>eps) eps = d;  // eps: reproduced relative data error
    bl1 = (A!=0 || dela!=0) && (B!=0 || epsb!=0);
    A1 = A + interval(-dela,dela);
    d = Sup( abs(B) );
    d = mulu(d,epsb);
    B1 = B + interval(-d,d);
    bl2 = A1/B1 <= interval(0.5,2);
    if (bl1 && !bl2)
        if (S>=MinReal) // Some differences are in the normalized range
        {
	    d   = addu(1,eps);
	    d   = mulu(eps_m,d);  // d: relative rounding error
	    eps = addu(d,eps);
        }
} // rel_subm2

void rel_mulm2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A*B and an upper bound eps of the relative error by 
               // calculating (a*b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    R = A*B;
    interval aR = abs(R);
    real S = Sup(aR), I = Inf(aR), IA = Inf( abs(A) ), d=0,d1;
    if (sign(dela)!=0 || sign(epsb)!=0)
    {
        d = divu(dela,IA);
        d1 = mulu(d,epsb);
	d1 = addu(d1,epsb);
        d = addu(d1,d); // d:  reproduced relative data error
    }
    if (I>=MinReal) // All products only in the normalized range
    {
	if (dela==0 && diam(A)==0 && mant(IA)==0.5) d1=0;
        else
	{
	    d1  = addu(1,d);
	    d1  = mulu(eps_m,d1);
	}   // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (S<MinReal) // All products only in the denormalized range U
    {
	d1 = divu(minreal,I);
	eps = addu(d,d1);
    } else // products in normalized and denormalized range --> maximum!:
    {
	d1  = divu(minreal,I);
	eps = addu(d,d1);     // eps: relative error in U
	d1  = addu(1,d);
	d1  = mulu(eps_m,d1); // d1:  relative rounding error
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    }
} // rel_mulm2

void rel_divm2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A/B and an upper bound eps of the relative error by 
               // calculating (a/b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    R = A/B;
    interval aR = abs(R);
    real S = Sup(aR), I = Inf(aR), IB = Inf( abs(B) ), d=0,d1;
    if (sign(dela)!=0 || sign(epsb)!=0)
    {
	real IA = Inf( abs(A) );
        d  = divu(dela,IA);
        d  = addu(d,epsb);
        d1 = subd(1,epsb);
        if (d1<=0) 
        {
	    cerr << "rel_divm2(...): Relative error doesn't exist." << endl;
	    exit(1);
        }
        d = divu(d,d1);  // d:  reproduced relative data error
    }
    if (I>=MinReal) // All quotients only in the normalized range
    {
	if (epsb==0 && diam(B)==0 && mant(IB)==0.5) d1=0;
        else
	{
	    d1 = addu(1,d);
	    d1 = mulu(eps_m,d1);
	}   // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (S<MinReal) // All quotients only in the denormalized range U
    {
	d1 = divu(minreal,I);
	eps = addu(d,d1);
    } else // quotients in normalized and denormalized range --> maximum!:
    {
	d1  = divu(minreal,I);
	eps = addu(d,d1);     // eps: relative error in U
	d1  = addu(1,d);
	d1  = mulu(eps_m,d1); // d1:  relative rounding error
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    }
} // rel_divm2

// ---------------------------------------------------------------------------
// --------------------------- Eps(a), Delta(b) ------------------------------
// --------------------------------------------------------------------------- 

void rel_addm3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A+B and an upper bound eps of the relative error by 
               // calculating (a+b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1;
    bool bl1,bl2;
    R = A+B;
    real S = Sup( abs(R) ), d;
    eps  = ForMaxS3(epsa,delb,B,Inf(A));
    d    = ForMaxS3(epsa,delb,B,Sup(A));
    if (d>eps) eps = d;  // eps: reproduced relative data error
    bl1 = (A!=0 || epsa!=0) && (B!=0 || delb!=0);
    d = Sup( abs(A) );
    d = mulu(d,epsa);
    A1 = A + interval(-d,d);
    B1 = B + interval(-delb,delb);
    bl2 = A1/B1 <= interval(-2,-0.5);
    if (bl1 && !bl2)
        if (S>=MinReal) // Some sums are in the normalized range
        {
	    d   = addu(1,eps);
	    d   = mulu(eps_m,d);  // d: relative rounding error
	    eps = addu(d,eps);
        }
} // rel_addm3

void rel_subm3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A+B and an upper bound eps of the relative error by 
               // calculating (a+b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1;
    bool bl1,bl2;
    R = A+B;
    real S = Sup( abs(R) ), d;

    eps  = ForMaxD3(epsa,delb,B,Inf(A));
    d    = ForMaxD3(epsa,delb,B,Sup(A));
    if (d>eps) eps = d;  // eps: reproduced relative data error

    bl1 = (A!=0 || epsa!=0) && (B!=0 || delb!=0);
    d = Sup( abs(A) );
    d = mulu(d,epsa);
    A1 = A + interval(-d,d);
    B1 = B + interval(-delb,delb);
    bl2 = A1/B1 <= interval(0.5,2);
    if (bl1 && !bl2)
        if (S>=MinReal) // Some sums are in the normalized range
        {
	    d   = addu(1,eps);
	    d   = mulu(eps_m,d);  // d: relative rounding error
	    eps = addu(d,eps);
        }
} // rel_subm3

void rel_mulm3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A*B and an upper bound eps of the relative error by 
               // calculating (a*b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    R = A*B;
    interval aR = abs(R);
    real S = Sup(aR), I = Inf(aR), IA = Inf( abs(A) ), d=0,d1;
    if (sign(epsa)!=0 || sign(delb)!=0)
    {
	real IB = Inf( abs(B) );
        d = divu(delb,IB);
        d1 = mulu(d,epsa);
	d1 = addu(d1,epsa);
        d = addu(d1,d); // d:  reproduced relative data error
    }
    if (I>=MinReal) // All products only in the normalized range
    {
	if (epsa==0 && diam(A)==0 && mant(IA)==0.5) d1=0;
        else
	{
	    d1  = addu(1,d);
	    d1  = mulu(eps_m,d1);
	}   // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (S<MinReal) // All products only in the denormalized range U
    {
	d1 = divu(minreal,I);
	eps = addu(d,d1);
    } else // products in normalized and denormalized range --> maximum!:
    {
	d1  = divu(minreal,I);
	eps = addu(d,d1);     // eps: relative error in U
	d1  = addu(1,d);
	d1  = mulu(eps_m,d1); // d1:  relative rounding error
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    }
} // rel_mulm3

void rel_divm3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A/B and an upper bound eps of the relative error by 
               // calculating (a/b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    R = A/B;
    interval aR = abs(R);
    real S = Sup(aR), I = Inf(aR), IB = Inf( abs(B) ), d=0,d1;
    if (sign(epsa)!=0 || sign(delb)!=0)
    {
        d  = mulu(IB,epsa);
        d  = addu(d,delb);   // d:  nominator ready
        d1 = subd(IB,delb);  // d1: denominator ready
        if (d1<=0) 
        {
	    cerr << "rel_divm3(...): Relative error doesn't exist." << endl;
	    exit(1);
        }
        d = divu(d,d1);  // d:  reproduced relative data error
    }
    if (I>=MinReal) // All quotients only in the normalized range
    {
	if (delb==0 && diam(B)==0 && mant(IB)==0.5) d1=0;
        else
	{
	d1 = addu(1,d);
	d1 = mulu(eps_m,d1);
	}   // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (S<MinReal) // All quotients only in the denormalized range U
    {
	d1 = divu(minreal,I);
	eps = addu(d,d1);
    } else // quotients in normalized and denormalized range --> maximum!:
    {
	d1  = divu(minreal,I);
	eps = addu(d,d1);     // eps: relative error in U
	d1  = addu(1,d);
	d1  = mulu(eps_m,d1); // d1:  relative rounding error
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    }
} // rel_divm3

// ---------------------------------------------------------------------------
// ------------------------------ Eps(a), Eps(b) -----------------------------
// ---------------------------------------------------------------------------

void rel_addm4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A+B and an upper bound eps of the relative error by 
               // calculating (a+b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1;
    bool bl1,bl2;
    real IA = Inf(A), SA = Sup(A), IB = Inf(B), SB = Sup(B),d,d1;
    eps  = ForMaxS4(epsa,epsb,IA,IB);
    d1 = ForMaxS4(epsa,epsb,IA,SB);
    if (d1>eps) eps = d1;
    d1 = ForMaxS4(epsa,epsb,SA,IB);
    if (d1>eps) eps = d1;
    d1 = ForMaxS4(epsa,epsb,SA,SB);
    if (d1>eps) eps = d1;  // eps: reproduced relative data error
    R = A+B;
    bl1 = (A!=0 || epsa!=0) && (B!=0 || epsb!=0);
    d = Sup( abs(A) );
    d = mulu(d,epsa);
    A1 = A + interval(-d,d);
    d = Sup( abs(B) );
    d = mulu(d,epsb);
    B1 = B + interval(-d,d);
    bl2 = A1/B1 <= interval(-2,-0.5);
    if (bl1 && !bl2)
    {   
	real S = Sup( abs(R) );
        if (S>=MinReal) // Some sums are in the normalized range
        {
	    d1  = addu(1,eps);
	    d1  = mulu(eps_m,d1);  // d1: relative rounding error
	    eps = addu(d1,eps);
        }
    }
} // rel_addm4

void rel_subm4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A-B and an upper bound eps of the relative error by 
               // calculating (a-b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    interval A1,B1;
    bool bl1,bl2;
    real IA = Inf(A), SA = Sup(A), IB = Inf(B), SB = Sup(B),d1;
    eps  = ForMaxD4(epsa,epsb,IA,IB);
    d1 = ForMaxD4(epsa,epsb,IA,SB);
    if (d1>eps) eps = d1;
    d1 = ForMaxD4(epsa,epsb,SA,IB);
    if (d1>eps) eps = d1;
    d1 = ForMaxD4(epsa,epsb,SA,SB);
    if (d1>eps) eps = d1;  // eps: reproduced relative data error
 
    R = A-B;
    bl1 = (A!=0 || epsa!=0) && (B!=0 || epsb!=0);
    d1 = Sup( abs(A) );  d1 = mulu(d1,epsa);
    A1 = A + interval(-d1,d1);
    d1 = Sup( abs(B) );  d1 = mulu(d1,epsb);
    B1 = B + interval(-d1,d1);
    bl2 = A1/B1 <= interval(0.5,2);
    if (bl1 && !bl2)
    {
	real S = Sup( abs(R) );
        if (S>=MinReal) // Some differences are in the normalized range
        {
	    d1   = addu(1,eps);
	    d1   = mulu(eps_m,d1);  // d1: relative rounding error
	    eps = addu(d1,eps);
        } 
    }
} // rel_subm4

void rel_mulm4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A*B and an upper bound eps of the relative error by 
               // calculating (a*b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    real d = mulu(epsa,epsb); 
    d = addu(d,epsa);
    d = addu(d,epsb);  // d:  reproduced relative data error

    R = A*B;
    interval aR = abs(R);
    real S = Sup(aR), I = Inf(aR), IA = Inf( abs(A) ), d1;

    if (I>=MinReal) // All products only in the normalized range
    {
	if (epsa==0 && diam(A)==0 && mant(IA)==0.5) d1=0;
        else
	{
	    d1  = addu(1,d);
	    d1  = mulu(eps_m,d1);
	}   // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (S<MinReal) // All products only in the denormalized range U
    {
	d1 = divu(minreal,I);
	eps = addu(d,d1);
    } else // products in normalized and denormalized range --> maximum!:
    {
	d1  = divu(minreal,I);
	eps = addu(d,d1);     // eps: relative error in U
	d1  = addu(1,d);
	d1  = mulu(eps_m,d1); // d1:  relative rounding error
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    }
} // rel_mulm4

void rel_divm4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw()  // Returns the interval 
               // R=A/B and an upper bound eps of the relative error by 
               // calculating (a/b) on the machine with maximum accuracy, 
               // where a and b are erroneous maschine values.
{
    real d = addu(epsa,epsb), d1 = subd(1,epsb);
    if (d1<=0) 
    {
	cerr << "rel_divm4(...): Relative error doesn't exist." << endl;
	exit(1);
    }
    d = divu(d,d1);  // d:  reproduced relative data error

    R = A/B;
    interval aR = abs(R);
    real S = Sup(aR), I = Inf(aR), IB = Inf( abs(B) );

    if (I>=MinReal) // All quotients only in the normalized range
    {
	if (epsb==0 && diam(B)==0 && mant(IB)==0.5) d1=0;
        else
	{
	    d1 = addu(1,d);
	    d1 = mulu(eps_m,d1);
	}   // d1:  relative rounding error
	eps = addu(d,d1);
    } else if (S<MinReal) // All quotients only in the denormalized range U
    {
	d1 = divu(minreal,I);
	eps = addu(d,d1);
    } else // quotients in normalized and denormalized range --> maximum!:
    {
	d1  = divu(minreal,I);
	eps = addu(d,d1);     // eps: relative error in U
	d1  = addu(1,d);
	d1  = mulu(eps_m,d1); // d1:  relative rounding error
	d1  = addu(d,d1);     // d1:  relative error in the normalized range
	if (d1>eps) eps = d1; // calculating the maximum 
    }
} // rel_divm4

} // namespace cxsc











