/*
**  CXSC is a C++ library for eXtended Scientific Computing
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2001 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*  Calculating absolute and relative error bounds for the basic operations
**  (+,-,*,/)  in maximum accuracy. The relative error for the basic 
**  operations is assumed as:   eps_m = 2^(-53) = 1.11022302...E-16  if the 
**  operation results are normalized IEEE-numbers. 
**
**  Functions to calculate the absolute error are named: 
**     abs_addm1(...)   for the basic operation +
**     abs_subm1(...)   for the basic operation -
**     abs_mulm1(...)   for the basic operation *
**     abs_divm1(...)   for the basic operation / 
**  The last sign '1' stands for:  absolute error of the two operands.
**  The last sign '2' stands for:  absolute error of the first operand and
**                                 relative error of the second operand.
**  The last sign '3' stands for:  relative error of the first operand and
**                                 absolute error of the second operand.
**  The last sign '4' stands for:  relative error of the two operands.
**
**  Functions to calculate the relative error are named: 
**     rel_addm1(...)   for the basic operation +
**     rel_subm1(...)   for the basic operation -
**     rel_mulm1(...)   for the basic operation *
**     rel_divm1(...)   for the basic operation / 
**  The meaning of last sign is the same as by calculating the absolute error.
**  The sign 'm' stands for:  maximum accuracy 
**
**  Blomquist, 28.01.03;
*/

#ifndef abs_relm_hpp_included
#define abs_relm_hpp_included

#include <interval.hpp>

namespace cxsc
{

//---------------------------------------------------------------------------
// ----------- Calculating the upper bounds of the absolute errors ----------
//---------------------------------------------------------------------------
// -------------------------- Delta(a), Delta(b) ----------------------------
//---------------------------------------------------------------------------
void abs_addm1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& del) throw(); // Returns the interval R=A+B
         // and an upper bound del of the absolute error by calculating (a+b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

void abs_subm1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& del) throw(); // Returns the interval R=A-B
         // and an upper bound del of the absolute error by calculating (a-b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

void abs_mulm1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& del) throw(); // Returns the interval R=A*B
         // and an upper bound del of the absolute error by calculating (a*b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.  

void abs_divm1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& del) throw(); // Returns the interval R=A/B
         // and an upper bound del of the absolute error by calculating (a/b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

//---------------------------------------------------------------------------
// ---------------------------- Delta(a), Eps(b) ----------------------------
//---------------------------------------------------------------------------

void abs_addm2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& del) throw(); // Returns the interval R=A+B
         // and an upper bound del of the absolute error by calculating (a+b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

void abs_subm2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& del) throw(); // Returns the interval R=A-B
         // and an upper bound del of the absolute error by calculating (a-b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

void abs_mulm2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& del) throw(); // Returns the interval R=A*B
         // and an upper bound del of the absolute error by calculating (a*b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.  

void abs_divm2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& del) throw(); // Returns the interval R=A/B
         // and an upper bound del of the absolute error by calculating (a/b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

//---------------------------------------------------------------------------
// ---------------------------- Eps(a), Delta(b) ----------------------------
//---------------------------------------------------------------------------

void abs_addm3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& del) throw(); // Returns the interval R=A+B
         // and an upper bound del of the absolute error by calculating (a+b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

void abs_subm3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& del) throw(); // Returns the interval R=A-B
         // and an upper bound del of the absolute error by calculating (a-b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

void abs_mulm3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& del) throw(); // Returns the interval R=A*B
         // and an upper bound del of the absolute error by calculating (a*b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.  

void abs_divm3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& del) throw(); // Returns the interval R=A/B
         // and an upper bound del of the absolute error by calculating (a/b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

//---------------------------------------------------------------------------
// ----------------------------- Eps(a), Eps(b) -----------------------------
//---------------------------------------------------------------------------

void abs_addm4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& del) throw(); // Returns the interval R=A+B
         // and an upper bound del of the absolute error by calculating (a+b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

void abs_subm4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& del) throw(); // Returns the interval R=A-B
         // and an upper bound del of the absolute error by calculating (a-b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

void abs_mulm4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& del) throw(); // Returns the interval R=A*B
         // and an upper bound del of the absolute error by calculating (a*b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.  

void abs_divm4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& del) throw(); // Returns the interval R=A/B
         // and an upper bound del of the absolute error by calculating (a/b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

//---------------------------------------------------------------------------
// ----------- Calculating the upper bounds of the relative errors ----------
//---------------------------------------------------------------------------
// -------------------------- Delta(a), Delta(b) ----------------------------
//---------------------------------------------------------------------------

void rel_addm1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& eps) throw(); // Returns the interval R=A+B
         // and an upper bound eps of the relative error by calculating (a+b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

void rel_subm1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& eps) throw(); // Returns the interval R=A-B
         // and an upper bound eps of the relative error by calculating (a-b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

void rel_mulm1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& eps) throw(); // Returns the interval R=A*B
         // and an upper bound eps of the relative error by calculating (a*b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.  

void rel_divm1(const interval& A, const real& dela,
               const interval& B, const real& delb,
               interval& R, real& eps) throw(); // Returns the interval R=A/B
         // and an upper bound eps of the relative error by calculating (a/b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

//---------------------------------------------------------------------------
// ---------------------------- Delta(a), Eps(b) ----------------------------
//---------------------------------------------------------------------------

void rel_addm2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw(); // Returns the interval R=A+B
         // and an upper bound eps of the relative error by calculating (a+b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

void rel_subm2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw(); // Returns the interval R=A-B
         // and an upper bound eps of the relative error by calculating (a-b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

void rel_mulm2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw(); // Returns the interval R=A*B
         // and an upper bound eps of the relative error by calculating (a*b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.  

void rel_divm2(const interval& A, const real& dela,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw(); // Returns the interval R=A/B
         // and an upper bound eps of the relative error by calculating (a/b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

//---------------------------------------------------------------------------
// ---------------------------- Eps(a), Delta(b) ----------------------------
//---------------------------------------------------------------------------

void rel_addm3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& eps) throw(); // Returns the interval R=A+B
         // and an upper bound del of the absolute error by calculating (a+b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

void rel_subm3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& eps) throw(); // Returns the interval R=A-B
         // and an upper bound del of the absolute error by calculating (a-b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

void rel_mulm3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& eps) throw(); // Returns the interval R=A*B
         // and an upper bound del of the absolute error by calculating (a*b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.  

void rel_divm3(const interval& A, const real& epsa,
               const interval& B, const real& delb,
               interval& R, real& eps) throw(); // Returns the interval R=A/B
         // and an upper bound del of the absolute error by calculating (a/b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

//---------------------------------------------------------------------------
// ----------------------------- Eps(a), Eps(b) -----------------------------
//---------------------------------------------------------------------------

void rel_addm4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw(); // Returns the interval R=A+B
         // and an upper bound eps of the relative error by calculating (a+b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

void rel_subm4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw(); // Returns the interval R=A-B
         // and an upper bound eps of the relative error by calculating (a-b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

void rel_mulm4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw(); // Returns the interval R=A*B
         // and an upper bound eps of the relative error by calculating (a*b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.  

void rel_divm4(const interval& A, const real& epsa,
               const interval& B, const real& epsb,
               interval& R, real& eps) throw(); // Returns the interval R=A/B
         // and an upper bound eps of the relative error by calculating (a/b)
         // on the machine with maximum accuracy, where a and b are erroneous
         // maschine values.

} // namespace cxsc 

#endif











