/*
**  CXSC is a C++ library for eXtended Scientific Computing
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2001 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef bnd_arh_hpp_included
#define bnd_arh_hpp_included

#include <interval.hpp>

namespace cxsc
{
// ---------------------------------------------------------------------------
// --------------------------- T(x) = 1+X ------------------------------------
// ---------------------------------------------------------------------------

    real Rnd_1pxh(const interval& x) throw();   // Returns the upper bound
         // of the absolute rounding error by calculating (1+r) on the 
         // machine with high accuracy; 
         // r is an element of type real of the interval x.

    void abs_1px_delh(const interval& X, const real& del,
		            interval& T, real& bnd) throw();
    // Berechnung des abs. Gesamtfehlers, wenn T(x)=1+x in nur hochgenauer
    // Arithmetik ausgewertet wird. X ist das Intervall der exakten
    // Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
    // T = 1+X ist eine Einschlieung der exakten Werte T(x)=1+x.

    void abs_1px_epsh(const interval& X, const real& eps,
		            interval& T, real& bnd) throw();
    // Berechnung des abs. Gesamtfehlers, wenn T(x)=1+x in nur hochgenauer
    // Arithmetik ausgewertet wird. X ist das Intervall der exakten
    // Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
    // T = 1+X ist eine Einschlieung der exakten Werte T(x)=1+x.

    void rel_1px_delh(const interval& X, const real& del,
		            interval& T, real& bnd) throw();
    // Berechnung des rel. Gesamtfehlers, wenn T(x)=1+x in nur hochgenauer
    // Arithmetik ausgewertet wird. X ist das Intervall der exakten
    // Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
    // T = 1+X ist eine Einschlieung der exakten Werte T(x)=1+x.
    // Bei zu breitem X ist eine Intervallunterteilung notwendig! 

    void rel_1px_epsh(const interval& X, const real& eps,
		            interval& T, real& bnd) throw();
    // Berechnung des rel. Gesamtfehlers, wenn T(x)=1+x in nur hochgenauer
    // Arithmetik ausgewertet wird. X ist das Intervall der exakten
    // Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
    // T = 1+X ist eine Einschlieung der exakten Werte T(x)=1+x. 
    // Bei zu breitem X ist eine Intervallunterteilung notwendig!

// ---------------------------------------------------------------------------
// --------------------------- T(x) = 1-X ------------------------------------
// ---------------------------------------------------------------------------

    void abs_1mx_delh(const interval& X, const real& del,
		            interval& T, real& bnd) throw();
    // Berechnung des abs. Gesamtfehlers, wenn T(x)=1-x in nur hochgenauer
    // Arithmetik ausgewertet wird. X ist das Intervall der exakten
    // Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
    // T = 1-X ist eine Einschlieung der exakten Werte T(x) = 1-x = 1+(-x).

    void abs_1mx_epsh(const interval& X, const real& eps,
		            interval& T, real& bnd) throw();
    // Berechnung des abs. Gesamtfehlers, wenn T(x)=1-x in nur hochgenauer
    // Arithmetik ausgewertet wird. X ist das Intervall der exakten
    // Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
    // T = 1-X ist eine Einschlieung der exakten Werte T(x) = 1-x = 1+(-x).

    void rel_1mx_delh(const interval& X, const real& del,
		            interval& T, real& bnd) throw();
    // Berechnung des rel. Gesamtfehlers, wenn T(x)=1-x in nur hochgenauer
    // Arithmetik ausgewertet wird. X ist das Intervall der exakten
    // Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
    // T = 1-X ist eine Einschlieung der exakten Werte T(x) = 1-x = 1+(-x).

    void rel_1mx_epsh(const interval& X, const real& eps,
		            interval& T, real& bnd) throw();
    // Berechnung des rel. Gesamtfehlers, wenn T(x)=1-x in nur hochgenauer
    // Arithmetik ausgewertet wird. X ist das Intervall der exakten
    // Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
    // T = 1-X ist eine Einschlieung der exakten Werte T(x) = 1-x = 1+(-x). 

// ---------------------------------------------------------------------------
// --------------------------- T(x) = X-1 ------------------------------------
// ---------------------------------------------------------------------------

    void abs_xm1_delh(const interval& X, const real& del,
		            interval& T, real& bnd) throw();
    // Berechnung des abs. Gesamtfehlers, wenn T(x)=x-1 in nur hochgenauer
    // Arithmetik ausgewertet wird. X ist das Intervall der exakten
    // Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
    // T = X-1 ist eine Einschlieung der exakten Werte T(x)=x-1 = -(1+(-x)).

    void abs_xm1_epsh(const interval& X, const real& eps,
		            interval& T, real& bnd) throw();
    // Berechnung des abs. Gesamtfehlers, wenn T(x)=x-1 in nur hochgenauer
    // Arithmetik ausgewertet wird. X ist das Intervall der exakten
    // Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
    // T = X-1 ist eine Einschlieung der exakten Werte T(x)=x-1 = -(1+(-x)).

    void rel_xm1_delh(const interval& X, const real& del,
		            interval& T, real& bnd) throw();
    // Berechnung des rel. Gesamtfehlers, wenn T(x)=x-1 in nur hochgenauer
    // Arithmetik ausgewertet wird. X ist das Intervall der exakten
    // Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
    // T = X-1 ist eine Einschlieung der exakten Werte T(x)=x-1 = -(1+(-x)).

    void rel_xm1_epsh(const interval& X, const real& eps,
		            interval& T, real& bnd) throw();
    // Berechnung des rel. Gesamtfehlers, wenn T(x)=x-1 in nur hochgenauer
    // Arithmetik ausgewertet wird. X ist das Intervall der exakten
    // Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
    // T = X-1 ist eine Einschlieung der exakten Werte T(x)=x-1 = -(1+(-x)). 

// -------------------------------------------------------------------------
// -------------------------------------------------------------------------

    real abs_err_xpah(const real& x, const real& a) throw(); // Returns the
         // upper bound of the absolute rounding error by calculating (x+a)
         // on the machine with high accuracy.

} // namespace cxsc 

#endif










