/*
**  CXSC is a C++ library for eXtended Scientific Computing
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2001 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "bnd_arm.hpp"
#include <dot.hpp>
#include <interval.hpp>

using namespace std;

namespace cxsc {
// ----------------------------------------------------------------------------
// ------------------------------ T(x) = 1+x ----------------------------------
// ----------------------------------------------------------------------------

real Rnd_1pxm(const interval& x) throw()
// Absolute rounding error by calculating r+1 in maximum accuracy;
// r is a machine number of the interval x;
{
    const real c_53 = comp(0.5,-52);  //  2^(-53)
    const real c_52 = comp(0.5,-51);  //  2^(-52)
    const real c53  = comp(0.5,54);   //  2^(+53)
    const real c_11 = comp(0.5,-10);  //  2^(-11)
    real ix = Inf(x), sx = Sup(x),y,t;

    if (sx>=c53 || ix<-c53) return real(1);
    if (ix<=-0.5) ix = succ(-0.5); // ix > -0.5 !!!
    if (sx<=-0.5) return 0;
    if (sx<-c_53) return c_53;
    if (sx>1)  
    {
	y = 0;
	int exsx = expo(sx);
	if (sx>comp(0.5,exsx+1)-1)  y = comp(0.5,exsx-52);
	else if (ix<comp(0.5,exsx)) y = comp(0.5,exsx-53);
	return y;
    }
    if (sx>=c_52) 
    {
	if (ix==1) y = 0;
	else y = (ix>=c_11) ? c_53 : c_52;
	return y;
    } 
    if (sign(sx)<0) y = (ix>=-c_53) ? abs(ix) : c_53;
    else // 0 <= sx < 2^(-52)  
    {
	y = sx;
	t = (ix>=-c_53) ? abs(ix) : c_53;
	if (t>y) y = t;
    }
    return y;
} // Rnd_1pxm

void abs_1px_delm(const interval& X, const real& del,
                        interval& T, real& bnd) throw()
// Berechnung des abs. Gesamtfehlers, wenn T(x)=1+x in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
// T = 1+X ist eine Einschlieung der exakten Werte T(x)=1+x. 
{
    real d = abs(del);
    T = 1+X;  // T: Einschlieung der exakten T(x)-Werte
    interval Xt = X+interval(-d,d);  // Xt: Einschlieung der
                                     // gestrten Maschinenzahlen
    bnd = Rnd_1pxm(Xt);
    bnd = addu(d,bnd); // bnd: Absoluter Gesamtfehler
}

void abs_1px_epsm(const interval& X, const real& eps,
                        interval& T, real& bnd) throw()
// Berechnung des abs. Gesamtfehlers, wenn T(x)=1+x in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
// T = 1+X ist eine Einschlieung der exakten Werte T(x)=1+x.     
{
    T = 1+X;  // T: Einschlieung der exakten T(x)-Werte
    real d = abs(eps);
    d = mulu(d,Sup(abs(X)));  // d: abs. Fehlerschr.     
    interval Xt = X+interval(-d,d);  // Xt: Einschlieung der
                                     // gestrten Maschinenzahlen
    bnd = Rnd_1pxm(Xt);
    bnd = addu(d,bnd); // bnd: Absoluter Gesamtfehler
} 

void rel_1px_delm(const interval& X, const real& del,
                        interval& T, real& bnd) throw()
// Berechnung des rel. Gesamtfehlers, wenn T(x)=1+x in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
// T = 1+X ist eine Einschlieung der exakten Werte T(x)=1+x.
// Bei zu breitem X ist eine Intervallunterteilung notwendig! 
{
    real d = abs(del);
    T = 1+X;  // T: Einschlieung der exakten T(x)-Werte
    interval Xt = X+interval(-d,d);  // Xt: Einschlieung der
                                     // gestrten Maschinenzahlen
    bnd = Rnd_1pxm(Xt);
    bnd = addu(d,bnd); // bnd: Absoluter Gesamtfehler
    if (sign(bnd)!=0) 
    {
	d = Inf( abs(T) );
	if (sign(d)==0)
	{
	    cerr << "rel_1px_delm(): Relative error doesn't exist!" << endl;
	    exit(1);
	}
        else bnd = divu(bnd,d); // bnd: Rel. error calculated
    }
} // rel_1px_delm

void rel_1px_epsm(const interval& X, const real& eps,
                        interval& T, real& bnd) throw()
// Berechnung des rel. Gesamtfehlers, wenn T(x)=1+x in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
// T = 1+X ist eine Einschlieung der exakten Werte T(x)=1+x. 
// Bei zu breitem X ist eine Intervallunterteilung notwendig!    
{
    T = 1+X;  // T: Einschlieung der exakten T(x)-Werte
    real d = abs(eps);
    d = mulu(d,Sup(abs(X)));  // d: abs. Fehlerschr.     
    interval Xt = X+interval(-d,d);  // Xt: Einschlieung der
                                     // gestrten Maschinenzahlen
    bnd = Rnd_1pxm(Xt);
    bnd = addu(d,bnd); // bnd: Absoluter Gesamtfehler
    if (sign(bnd)!=0) 
    {
	d = Inf( abs(T) );
	if (sign(d)==0)
	{
	    cerr << "rel_1px_epsm(): Relative error doesn't exist!" << endl;
	    exit(1);
	}
        else bnd = divu(bnd,d); // bnd: Rel. error calculated
    }
}

// ----------------------------------------------------------------------------
// ------------------------------ T(x) = 1-x ----------------------------------
// ----------------------------------------------------------------------------

void abs_1mx_delm(const interval& X, const real& del,
                        interval& T, real& bnd) throw()
// Berechnung des abs. Gesamtfehlers, wenn T(x)=1-x in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
// T = 1-X ist eine Einschlieung der exakten Werte T(x) = 1-x = 1+(-x). 
{   
    abs_1px_delm(-X,del,T,bnd);
} // abs_1mx_delm

void abs_1mx_epsm(const interval& X, const real& eps,
                        interval& T, real& bnd) throw()
// Berechnung des abs. Gesamtfehlers, wenn T(x)=1-x in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
// T = 1-X ist eine Einschlieung der exakten Werte T(x) = 1-x = 1+(-x). 
{   
    abs_1px_epsm(-X,eps,T,bnd);
} // abs_1mx_epsm

void rel_1mx_delm(const interval& X, const real& del,
                        interval& T, real& bnd) throw()
// Berechnung des rel. Gesamtfehlers, wenn T(x)=1-x in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
// T = 1-X ist eine Einschlieung der exakten Werte T(x) = 1-x = 1+(-x). 
{   
    rel_1px_delm(-X,del,T,bnd);
} // abs_1mx_delm

void rel_1mx_epsm(const interval& X, const real& eps,
                        interval& T, real& bnd) throw()
// Berechnung des rel. Gesamtfehlers, wenn T(x)=1-x in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
// T = 1-X ist eine Einschlieung der exakten Werte T(x) = 1-x = 1+(-x). 
{   
    rel_1px_epsm(-X,eps,T,bnd);
} // rel_1mx_epsm

// ----------------------------------------------------------------------------
// ------------------------------ T(x) = x-1 ----------------------------------
// ----------------------------------------------------------------------------

void abs_xm1_delm(const interval& X, const real& del,
                        interval& T, real& bnd) throw()
// Berechnung des abs. Gesamtfehlers, wenn T(x)=x-1 in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
// T = X-1 ist eine Einschlieung der exakten Werte T(x)=x-1 = -(1+(-x)). 
{   
    abs_1px_delm(-X,del,T,bnd);
    T = -T;
} // abs_xm1_delm

void abs_xm1_epsm(const interval& X, const real& eps,
                        interval& T, real& bnd) throw()
// Berechnung des abs. Gesamtfehlers, wenn T(x)=x-1 in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
// T = X-1 ist eine Einschlieung der exakten Werte T(x)=x-1 = -(1+(-x)). 
{   
    abs_1px_epsm(-X,eps,T,bnd);
    T = -T;
} // abs_xm1_epsm

void rel_xm1_delm(const interval& X, const real& del,
                        interval& T, real& bnd) throw()
// Berechnung des rel. Gesamtfehlers, wenn T(x)=x-1 in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
// T = X-1 ist eine Einschlieung der exakten Werte T(x)=x-1 = -(1+(-x)). 
{   
    rel_1px_delm(-X,del,T,bnd);
    T = -T;
} // abs_xm1_delm

void rel_xm1_epsm(const interval& X, const real& eps,
                        interval& T, real& bnd) throw()
// Berechnung des rel. Gesamtfehlers, wenn T(x)=x-1 in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
// T = X-1 ist eine Einschlieung der exakten Werte T(x)=x-1 = -(1+(-x)). 
{   
    rel_1px_epsm(-X,eps,T,bnd);
    T = -T;
} // rel_xm1_epsm

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

real abs_err_xpam(const real& x, const real& a) throw()
// Returns the absolute error by calculating (x+a) on the machine in maximum
// accuracy.
{
    dotprecision dot;
    interval z;
    real xp1,delx;
    dot = a;
    dot += x;    // dot = a+x;
    xp1 = a+x;
    dot -= xp1;  // dot includes the absolute error;
    if (sign(dot)==0) delx = 0;
    else 
    {
	rnd(dot,z);
	z = abs(z);
	delx = Sup(z);
    }
    return delx;
} // abs_err_xpam


} // namespace cxsc







