/*
**  CXSC is a C++ library for eXtended Scientific Computing
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2001 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*  Calculating error bounds for the following arithmetic expressions:
**
**     1.   r+1,  r of type real in a given interval x;
**     2.   r-1,  r of type real in a given interval x;
**     3.   r+a,  r,a of type real;
**
**  in maximum accuracy. The relative error for the basic operations +,-,*,/
**  is assumed as:   eps = 2^(-53) = 1.11022302...E-16    if the operation
**  results are normalized IEEE-numbers.
**
**  Calculating the relative error, large intervals x may produce such
**  overestimations!! In this cases you should subdivide the given interval x
**  into a sufficient large number of subintervals and calculate the maximum
**  of the relative errors of all these subintervals as the relative error 
**  bound for the given interval x. 
**
**  Blomquist;
*/

#ifndef bnd_arm_hpp_included
#define bnd_arm_hpp_included

#include <interval.hpp>

namespace cxsc
{
//-----------------------------------------------------------------------------
//------------------------------ T(x) = 1+x -----------------------------------
//-----------------------------------------------------------------------------
    real Rnd_1pxm(const interval& x) throw(); // Returns the upper bound
         // of the absolute rounding error by calculating (r+1) on the 
         // machine with maximum accuracy; r is an element of type real 
         // of the interval x.

    void abs_1px_delm(const interval& X, const real& del,
		            interval& T, real& bnd) throw();
    // Berechnung des abs. Fehlers, wenn T(x)=1+x in maximal genauer
    // Arithmetik ausgewertet wird. X ist das Intervall der exakten
    // Werte, und del ist die Fehlerschranke der gestrten Maschinenwerte.
    // T = 1+X ist eine Einschlieung der exakten Werte T(x)=1+x.

    void abs_1px_epsm(const interval& X, const real& eps,
		            interval& T, real& bnd) throw();
    // Berechnung des abs. Gesamtfehlers, wenn T(x)=1+x in maximal genauer
    // Arithmetik ausgewertet wird. X ist das Intervall der exakten
    // Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
    // T = 1+X ist eine Einschlieung der exakten Werte T(x)=1+x. 

    void rel_1px_delm(const interval& X, const real& del,
		            interval& T, real& bnd) throw();
    // Berechnung des rel. Gesamtfehlers, wenn T(x)=1+x in maximal genauer
    // Arithmetik ausgewertet wird. X ist das Intervall der exakten
    // Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
    // T = 1+X ist eine Einschlieung der exakten Werte T(x)=1+x.
    // Bei zu breitem X ist eine Intervallunterteilung notwendig!

    void rel_1px_epsm(const interval& X, const real& eps,
		            interval& T, real& bnd) throw();
    // Berechnung des rel. Gesamtfehlers, wenn T(x)=1+x in maximal genauer
    // Arithmetik ausgewertet wird. X ist das Intervall der exakten
    // Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
    // T = 1+X ist eine Einschlieung der exakten Werte T(x)=1+x. 
    // Bei zu breitem X ist eine Intervallunterteilung notwendig!

// ----------------------------------------------------------------------------
// ---------------------------- T(x) = 1-x ------------------------------------
// ----------------------------------------------------------------------------

void abs_1mx_delm(const interval& X, const real& del,
		        interval& T, real& bnd) throw();
// Berechnung des abs. Gesamtfehlers, wenn T(x)=1-x in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
// T = 1-X ist eine Einschlieung der exakten Werte T(x) = 1-x = 1+(-x). 

void abs_1mx_epsm(const interval& X, const real& eps,
		        interval& T, real& bnd) throw();
// Berechnung des abs. Gesamtfehlers, wenn T(x)=1-x in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
// T = 1-X ist eine Einschlieung der exakten Werte T(x) = 1-x = 1+(-x). 

void rel_1mx_delm(const interval& X, const real& del,
		        interval& T, real& bnd) throw();
// Berechnung des rel. Gesamtfehlers, wenn T(x)=1-x in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
// T = 1-X ist eine Einschlieung der exakten Werte T(x) = 1-x = 1+(-x). 

void rel_1mx_epsm(const interval& X, const real& eps,
		        interval& T, real& bnd) throw();
// Berechnung des rel. Gesamtfehlers, wenn T(x)=1-x in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
// T = 1-X ist eine Einschlieung der exakten Werte T(x) = 1-x = 1+(-x). 

// ----------------------------------------------------------------------------
// ---------------------------- T(x) = x-1 ------------------------------------
// ----------------------------------------------------------------------------

void abs_xm1_delm(const interval& X, const real& del,
		        interval& T, real& bnd) throw();
// Berechnung des abs. Gesamtfehlers, wenn T(x)=x-1 in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
// T = X-1 ist eine Einschlieung der exakten Werte T(x)=x-1 = -(1+(-x)). 

void abs_xm1_epsm(const interval& X, const real& eps,
		        interval& T, real& bnd) throw();
// Berechnung des abs. Gesamtfehlers, wenn T(x)=x-1 in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
// T = X-1 ist eine Einschlieung der exakten Werte T(x)=x-1 = -(1+(-x)). 

void rel_xm1_delm(const interval& X, const real& del,
		        interval& T, real& bnd) throw();
// Berechnung des rel. Gesamtfehlers, wenn T(x)=x-1 in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und del ist die abs. Fehlerschr. der gestrten Maschinenwerte.
// T = X-1 ist eine Einschlieung der exakten Werte T(x)=x-1 = -(1+(-x)). 

void rel_xm1_epsm(const interval& X, const real& eps,
		        interval& T, real& bnd) throw();
// Berechnung des rel. Gesamtfehlers, wenn T(x)=x-1 in maximal genauer
// Arithmetik ausgewertet wird. X ist das Intervall der exakten
// Werte, und eps ist die rel. Fehlerschr. der gestrten Maschinenwerte.
// T = X-1 ist eine Einschlieung der exakten Werte T(x)=x-1 = -(1+(-x)). 

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

    real abs_err_xpam(const real& x, const real& a) throw(); // Returns the
         // upper bound of the absolute error by calculating (x+a) on the 
         // machine with maximum accuracy.

} // namespace cxsc 

#endif





