/*
**  CXSC is a C++ library for eXtended Scientific Computing
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2001 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "abs_relh.hpp"   // Wegen rel_mulh4(), ...
#include "bnd_arh.hpp"    // Wegen rel_1px_epsh()
#include "bnd_util.hpp"
#include <l_interval.hpp>
#include <intvector.hpp>
#include <ivector.hpp>

using namespace std;
namespace cxsc {

static void Koeff_tools(const intvector& Za, const intvector& Ne,
                        ivector& ia, rvector& Next, rvector& del, rvector& eps)
// Mit Hilfe der Vektoren Za,Ne, welche die Zhler und Nenner der 
// Polynomkoeffizienten a[j] enthalten, werden fr j=1 bis j=Ub(Za), d.h.
// fr alle a[j] mit ia,Next,del und eps Vektoren berechnet, die optimale 
// Einschlieungen, zur nchsten Rasterzahl gerundete a[j] und die mit 
// dieser Rundung verbundenen absoluten und relative Datenfehler liefern.
{
    int N = Ub(Za);
    for (int j=1; j<=N; j++)
    {
	frac_tools(Za[j],Ne[j],ia[j],Next[j],del[j],eps[j]);
    }
} // Koeff_tools

static real Horn_relh1_Xi(const interval& Xi, const real& epsx, 
                          const ivector& ia, const rvector& del, const int& k)
// Berechnung der relativen Fehlerschranke ber einem Teilintervall Xi,
// welches die exakten Polynom-Argumente enthlt.
// Vorausgesetzt wird eine nur hochgenauer Arithmetik, mit u(x) = x^k. 
// epsx ist die relative Fehlerschranke der gestrten Polynomargumente.
// ia importiert die optimalen Einschlieungen der exakten Polynom-Koeffztn. 
// del importiert die absoluten Fehlerschranken bezglich der Rundung der 
// exakten Polynomkoeffizienten zur jeweils nchsten IEEE-Rasterzahl.
// Das Zeichen '1' im Funktionsnamen bedeutet, dass der Polynomkoeffizient
// a[0] == 1 sein muss!!
{
    real del1,del2,delu;
    interval u,h1,h2;
    delu = Sup( abs(Xi) );
    delu = mulu(delu,epsx);
    // Auswertung von u(x) = x^k:
    switch (k)
    {
	case 1:  // u(x) = x^1
	    u = Xi;
	    break;
	case 2:  // u(x) = x^2
	    abs_mulh4(Xi,epsx,Xi,epsx,u,delu);
	    break;
	case 3:
	    abs_mulh4(Xi,epsx,Xi,epsx,h1,del1);  // h1: Einschlieg. fr Xi*Xi
	    abs_mulh2(h1,del1,Xi,epsx,u,delu);   // u: Einschg. fr (Xi*Xi)*Xi
	    break;
	case 4:  // u(x) = x^4
	    abs_mulh4(Xi,epsx,Xi,epsx,h1,del1);  // h1: Einschlieg. fr Xi*Xi
	    abs_mulh1(h1,del1,h1,del1,u,delu); // u:  Einschlg. fr (Xi*Xi)^2
	    break;
	case 5:  // u(x) = x^5
	    abs_mulh4(Xi,epsx,Xi,epsx,h1,del1); // h1: Einschlieg. fr Xi*Xi
	    abs_mulh1(h1,del1,h1,del1,h2,del2); // h2: Einschlg. fr (Xi*Xi)^2
            abs_mulh2(h2,del2,Xi,epsx,u,delu);  // u:  Einschlg. fr Xi^5
	    break;
	case 6:  // u(x) = x^6
	    abs_mulh4(Xi,epsx,Xi,epsx,h1,del1); // h1: Einschlieg. fr Xi*Xi
	    abs_mulh1(h1,del1,h1,del1,h2,del2); // h2: Einschlg. fr (Xi*Xi)^2
            abs_mulh1(h2,del2,h1,del1,u,delu);  // u:  Einschlg. fr Xi^6
	    break;
	default:
	    cerr << "Horn_relh1_Xi(): Unzulssige Potenz x^k" << endl;
	    exit(1);
    } // switch
    // u ist Einschlieung von Xi^k; 
    // delu ist die zugehrige absolute Fehlerschranke; 
    h1 = ia[Ub(ia)];     // h1: Startwert fr Hornerschema
    del1 = del[Ub(ia)];  // del1: absolute Fehlerschranke
    int m = Ub(ia)-1;
    for (int j=m; j>=0; j--)
    {   
	abs_mulh1(h1,del1,u,delu,h2,del2);            // h2 = y*u
	if (j==0) rel_1px_delh(h2,del2,h1,del1);
	else abs_addh1(ia[j],del[j],h2,del2,h1,del1); // h1 = y*u + a[j]
    } // del1: Relativer Fehler des Hornerschemas wegen rel_1px_delh()
    return del1;            // Rckgabe der relativen Fehlerschranke 
} // Horn_relh1_Xi

static real Rat_relh_Xi(const interval& Xi, const real& epsx, 
                        const ivector& ia, const ivector& ib, 
                        const int& k, const real& x0)
// Berechnung der relativen Fehlerschranke ber einem Teilintervall Xi,
// welches die exakten Polynom-Argumente enthlt.
// Vorausgesetzt wird eine nur hochgenauer Arithmetik, mit u(x) = (x-x0)^k. 
// epsx ist die relative Fehlerschranke der gestrten Polynomargumente.
// ia,ib sind Punktintervalle und importieren die optimalen Einschlieungen
// der exakten Polynomkoeffizienten. 

{
    real del1,del2,delu,delK;
    interval Kl,u,h1,h2;
    abs_subh3(Xi,epsx,interval(x0),0,Kl,delK); 
    // Kl  : Einschlieung von (x-x0) fr alle x aus Xi;
    // delK: Absolute Fehlerschranke bei Auswertung von x-x0, wobei die 
    //       relative Fehlerschranke epsx der gestrten Polynomargumente 
    //       bercksichtigt wird. 
    // Auswertung von u(x) = (x-x0)^k:
    switch (k)
    {
	case 1:  // u(x) = (x-x0)^1
	    u = Kl;  delu = delK;
	    break;
	case 2:  // u(x) = (x-x0)^2
	    abs_mulh1(Kl,delK,Kl,delK,u,delu);
	    break;
	case 3:
	    abs_mulh1(Kl,delK,Kl,delK,h1,del1);  // h1: Einschlieg. fr Kl*Kl
	    abs_mulh1(h1,del1,Kl,delK,u,delu);   // u: Einschg. fr (Kl*Kl)*Kl
	    break;
	case 4:  // u(x) = x^4
	    abs_mulh1(Kl,delK,Kl,delK,h1,del1);  // h1: Einschlieg. fr Kl*Kl
	    abs_mulh1(h1,del1,h1,del1,u,delu);  // u:  Einschlg. fr (Kl*Kl)^2
	    break;
	case 5:  // u(x) = x^5
	    abs_mulh1(Kl,delK,Kl,delK,h1,del1); // h1: Einschlieg. fr Kl*Kl
	    abs_mulh1(h1,del1,h1,del1,h2,del2); // h2: Einschlg. fr (Kl*Kl)^2
            abs_mulh1(h2,del2,Kl,delK,u,delu);  // u:  Einschlg. fr Kl^5
	    break;
	case 6:  // u(x) = x^6
	    abs_mulh1(Kl,delK,Kl,delK,h1,del1); // h1: Einschlieg. fr Kl*Kl
	    abs_mulh1(h1,del1,h1,del1,h2,del2); // h2: Einschlg. fr (Kl*Kl)^2
            abs_mulh1(h2,del2,h1,del1,u,delu);  // u:  Einschlg. fr Kl^6
	    break;
	default:
	    cerr << "Horn_relh_Xi(): Unzulssige Potenz x^k" << endl;
	    exit(1);
    } // switch
    // u ist Einschlieung von (Xi-x0)^k; 
    // delu ist die zugehrige absolute Fehlerschranke; 

    h1 = ia[Ub(ia)]; // h1: Startwert fr Hornerschema des Zhlerpolynoms
    del1 = 0;  // Die Polynomkoeffizienten sind rundungsfehlerfrei!
    int m = Ub(ia)-1;
    for (int j=m; j>=0; j--)
    {   
	abs_mulh1(h1,del1,u,delu,h2,del2);          // h2 = y*u
	if (ia[j]==1) abs_1px_delh(h2,del2,h1,del1);
	else abs_addh1(ia[j],0,h2,del2,h1,del1); // h1 = y*u + a[j]
    } 
    delK = del1; // delK: Absoluter Auswertefehler des Zhlerpolynoms
    Kl = h1;     // Kl  : Einschlieung des Zhlerpolynomwertes

    h1 = ib[Ub(ib)]; // h1: Startwert fr Hornerschema des Nennerpolynoms
    del1 = 0;  // Die Polynomkoeffizienten sind rundungsfehlerfrei!
    m = Ub(ib)-1;
    for (int j=m; j>=0; j--)
    {   
	abs_mulh1(h1,del1,u,delu,h2,del2);          // h2 = y*u
	if (ib[j]==1) abs_1px_delh(h2,del2,h1,del1);
	else abs_addh1(ib[j],0,h2,del2,h1,del1); // h1 = y*u + a[j]
    } // del1: absoluter Fehler;  h1: Einschlieung des Nennerpolynoms

    // Berechnung des relativen Fehlers des Polynom-Quotienten:
    rel_divh1(Kl,delK,h1,del1,h2,del2);

    return del2;  // Rckgabe der relativen Fehlerschranke del2
} // Rat_relh_Xi

static real Horn_relh_Xi(const interval& Xi, const real& epsx, 
                         const ivector& ia, const int& k, const real& x0)
// Berechnung der relativen Fehlerschranke ber einem Teilintervall Xi,
// welches die exakten Polynom-Argumente enthlt.
// Vorausgesetzt wird eine nur hochgenauer Arithmetik, mit u(x) = (x-x0)^k. 
// epsx ist die relative Fehlerschranke der gestrten Polynomargumente.
// ia importiert die optimalen Einschlieungen der exakten Polynom-Koeffztn. 

{
    real del1,del2,delu,delK;
    interval Kl,u,h1,h2;
    abs_subh3(Xi,epsx,interval(x0),0,Kl,delK); 
    // Kl  : Einschlieung von (x-x0) fr alle x aus Xi;
    // delK: Absolute Fehlerschranke bei Auswertung von x-x0, wobei die 
    //       relative Fehlerschranke epsx der gestrten Polynomargumente 
    //       bercksichtigt wird. 
    // Auswertung von u(x) = (x-x0)^k:
    switch (k)
    {
	case 1:  // u(x) = (x-x0)^1
	    u = Kl;  delu = delK;
	    break;
	case 2:  // u(x) = (x-x0)^2
	    abs_mulh1(Kl,delK,Kl,delK,u,delu);
	    break;
	case 3:
	    abs_mulh1(Kl,delK,Kl,delK,h1,del1);  // h1: Einschlieg. fr Kl*Kl
	    abs_mulh1(h1,del1,Kl,delK,u,delu);   // u: Einschg. fr (Kl*Kl)*Kl
	    break;
	case 4:  // u(x) = x^4
	    abs_mulh1(Kl,delK,Kl,delK,h1,del1);  // h1: Einschlieg. fr Kl*Kl
	    abs_mulh1(h1,del1,h1,del1,u,delu);  // u:  Einschlg. fr (Kl*Kl)^2
	    break;
	case 5:  // u(x) = x^5
	    abs_mulh1(Kl,delK,Kl,delK,h1,del1); // h1: Einschlieg. fr Kl*Kl
	    abs_mulh1(h1,del1,h1,del1,h2,del2); // h2: Einschlg. fr (Kl*Kl)^2
            abs_mulh1(h2,del2,Kl,delK,u,delu);  // u:  Einschlg. fr Kl^5
	    break;
	case 6:  // u(x) = x^6
	    abs_mulh1(Kl,delK,Kl,delK,h1,del1); // h1: Einschlieg. fr Kl*Kl
	    abs_mulh1(h1,del1,h1,del1,h2,del2); // h2: Einschlg. fr (Kl*Kl)^2
            abs_mulh1(h2,del2,h1,del1,u,delu);  // u:  Einschlg. fr Kl^6
	    break;
	default:
	    cerr << "Horn_relh_Xi(): Unzulssige Potenz x^k" << endl;
	    exit(1);
    } // switch
    // u ist Einschlieung von (Xi-x0)^k; 
    // delu ist die zugehrige absolute Fehlerschranke; 

    h1 = ia[Ub(ia)];     // h1: Startwert fr Hornerschema
    del1 = 0;
    int m = Ub(ia)-1;
    for (int j=m; j>=0; j--)
    {   
	abs_mulh1(h1,del1,u,delu,h2,del2);          // h2 = y*u
	if (ia[j]==1) abs_1px_delh(h2,del2,h1,del1);
	else abs_addh1(ia[j],0,h2,del2,h1,del1); // h1 = y*u + a[j]
    } // del1: Absoluter Fehler des Hornerschemas;
    del2 = Inf( abs(h1) );
    if (sign(del2)==0)
    {
	cerr << "Horn_relh_Xi: The relative error doesn't exist" << endl;
	exit(1);
    }
    del1 = divu(del1,del2); // Relative Fehlerschranke des Hornerschemas
    return del1;            // Rckgabe der relativen Fehlerschranke 
} // Horn_relh_Xi

static void Max_Horn_Xi(real (*FPt)(const interval&, const real&, 
                                    const ivector&,  const rvector&,
                                    const int&), // Funktions-Zeiger
                 const interval& X,   // X: Zu unterteilendes Intervall der
		                      //    exakten Argumente
		 const real& bndx,    // bndx: abs. oder rel. Fehlerschranke
                                      //       der gestrten Argumente
                 const real& diam,    // diam: Mantissen-Durchmesser bei der
                                      //       Intervallunterteilung
		 const ivector& ia,   // Optimale Koeffizienteneinschlieung
		 const rvector& bnda, // bnda: Abs. oder rel. Fehler bei der
                                      //       Koeffz.-Rundung zur nchsten 
                                      //       Rasterzahl.
		 const int& k,        // k:    u(x) = x^k
	               real& bnd)     // bnd: Maximum aller Fehlerschranken.
// Max_Horn_Xi unterteilt X mittels diam in Teilintervalle Xi, berechnet fr
// alle Xi mit der Funktion FPt() die gesuchten Fehlerschranken und liefert 
// mit bnd das Maximum dieser Fehlerschranken.
// Bedeutung der Funktionsparameter von FPt:
// 1. Param.: Teilintervall der exakten Argumente fr die Fehlerabschtzung;
//            Dieses Teilintervall wird in dieser Funktion als Xi berechnet.
// 2. Param.: Absolute oder rel. Fehlerschranke der gestrten Argumente.
//            Dieser Parameter wird beim Aufruf von FPt durch bndx ersetzt.
// 3. Param.: Intervallvektor, mit den optimalen Koeffztn.-Einschlieungen.
//            Dieser Parameter wird beim Aufruf von FPt durch  ia  ersetzt.
// 4. Param.: Vektor mit abs. oder rel. Fehlern bez. der Koeffz.-Rundungen.
//            Dieser Parameter wird beim Aufruf von FPt durch bnda ersetzt.
// 5. Param.: Integer-Gre k bezglich u(x) = x^k. 
//            Dieser Parameter wird beim Aufruf von FPt durch  k  ersetzt. 
{
    int ex;
    real r1 = Inf(X), m = mant(r1), SX = Sup(X), d = 0, r2;
    bnd = 0;
    if (sign(m)==0) r2 = minreal;
    else
    {
        ex = expo(r1);
        r2 = m + diam;
        times2pown(r2,ex);
	r2 = succ(r2);
    }
    if (r2>SX) r2 = SX;
    interval Xi = interval(r1,r2); // Xi: first subinterval
    do 
    {
	d = FPt(Xi,bndx,ia,bnda,k); // d: Fehlerschr. ber Teilintervall Xi 
	if (d>bnd) bnd = d;         // Maximumbildung
        // New subinterval:
	r1 = r2; // Neue Intervalluntergrenze = alte Intervallobergrenze
        m = mant(r1);
	if (sign(m)==0) r2 = minreal;
        else
	{
            ex = expo(r1);
	    r2 = m + diam;
            times2pown(r2,ex);
	    r2 = succ(r2);
	}
        if (r2>SX) r2 = SX;    // Neue Obergrenze r2 darf X nicht nach 
                               // oben berschreiten!
	Xi = interval(r1,r2);  // Xi: Neues Teilintervall
    } while (r1 < SX);
} // Max_Horn_Xi

static void Max_Horn_Xi(real (*FPt)(const interval&, const real&, 
                                    const ivector&,  const int&,
                                    const real&),    // Funktions-Zeiger
                 const interval& X,   // X: Zu unterteilendes Intervall der
		                      //    exakten Argumente
		 const real& bndx,    // bndx: abs. oder rel. Fehlerschranke
                                      //       der gestrten Argumente
                 const real& diam,    // diam: Mantissen-Durchmesser bei der
                                      //       Intervallunterteilung
		 const ivector& ia,   // Optimale Koeffizienteneinschlieung
		 const int& k,        // k:    u(x) = x^k
		 const real& x0,      // x0:   Entwicklingspunkt
	               real& bnd)     // bnd: Maximum aller Fehlerschranken.
// Max_Horn_Xi unterteilt X mittels diam in Teilintervalle Xi, berechnet fr
// alle Xi mit der Funktion FPt() die gesuchten Fehlerschranken und liefert 
// mit bnd das Maximum dieser Fehlerschranken.
// Bedeutung der Funktionsparameter von FPt:
// 1. Param.: Teilintervall der exakten Argumente fr die Fehlerabschtzung;
//            Dieses Teilintervall wird in dieser Funktion als Xi berechnet.
// 2. Param.: Absolute oder rel. Fehlerschranke der gestrten Argumente.
//            Dieser Parameter wird beim Aufruf von FPt durch bndx ersetzt.
// 3. Param.: Intervallvektor, mit den optimalen Koeffztn.-Einschlieungen.
//            Dieser Parameter wird beim Aufruf von FPt durch  ia  ersetzt.
// 4. Param.: Integer-Gre k bezglich u(x) = x^k. 
//            Dieser Parameter wird beim Aufruf von FPt durch  k  ersetzt. 
// 5. Param.: real-Gre x0 ist Entwicklungspunkt
{
    int ex;
    real r1 = Inf(X), m = mant(r1), SX = Sup(X), d = 0, r2;
    bnd = 0;
    if (sign(m)==0) r2 = minreal;
    else
    {
        ex = expo(r1);
        r2 = m + diam;
        times2pown(r2,ex);
	r2 = succ(r2);
    }
    if (r2>SX) r2 = SX;
    interval Xi = interval(r1,r2); // Xi: first subinterval
    do 
    {
	d = FPt(Xi,bndx,ia,k,x0);  // d: Fehlerschr. ber Teilintervall Xi 
	if (d>bnd) bnd = d;        // Maximumbildung
        // New subinterval:
	r1 = r2; // Neue Intervalluntergrenze = alte Intervallobergrenze
        m = mant(r1);
	if (sign(m)==0) r2 = minreal;
        else
	{
            ex = expo(r1);
	    r2 = m + diam;
            times2pown(r2,ex);
	    r2 = succ(r2);
	}
        if (r2>SX) r2 = SX;    // Neue Obergrenze r2 darf X nicht nach 
                               // oben berschreiten!
	Xi = interval(r1,r2);  // Xi: Neues Teilintervall
    } while (r1 < SX);
} // Max_Horn_Xi

void Max_bnd_Xi(real (*FPt)(const interval&, const real&), // Funkt.-Zeiger
                const interval& X, // X: Zu unterteilendes Intervall
	        const real& bndx,  // bndx: Fehlerschranke der Elemente aus X
                const real& diam,  // diam: Mantissen-Durchmesser
	              real& bnd)   // bnd: Maximum aller Fehlerschranken
{
    int ex,exr2;
    real r1 = Inf(X), m = mant(r1), SX = Sup(X), d = 0, r2;
    bnd = 0;
    if (sign(m)==0) r2 = minreal;
    else
    {
        ex = expo(r1);
        r2 = m + diam;
        times2pown(r2,ex);
	r2 = succ(r2);
    }
    if (r2>SX) r2 = SX;
    interval Xi = interval(r1,r2); // Xi: first subinterval
    do 
    {
	d = FPt(Xi,bndx);
	if (d>bnd) bnd = d;
        // New subinterval:
	r1 = r2;
        m = mant(r1);
	if (sign(m)==0) r2 = minreal;
        else
	{
            ex = expo(r1);
	    r2 = m + diam;
	    exr2 = expo(r2); 
	    if (ex+exr2>1024) r2 = SX;
	    else { 
		times2pown(r2,ex);
		r2 = succ(r2); 
	    }
	}
        if (r2>SX) r2 = SX;
	Xi = interval(r1,r2);
    } while (r1 < SX);
} // Max_bnd_Xi

void Max_Rat_Xi(real (*FPt)(const interval&, const real&, 
                            const ivector&, const ivector&, const int&,
                            const real&),    // Funktions-Zeiger
                const interval& X,   // X: Zu unterteilendes Intervall der
		                     //    exakten Argumente
		const real& bndx,    // bndx: abs. oder rel. Fehlerschranke
                                     //       der gestrten Argumente
                const real& diam,    // diam: Mantissen-Durchmesser bei der
                                     //       Intervallunterteilung
		const ivector& ia,   // Optimale Koeffizienteneinschlieung
                const ivector& ib,   // durch Punktintervalle
		const int& k,        // k:    u(x) = x^k
		const real& x0,      // x0:   Entwicklingspunkt
	              real& bnd)     // bnd: Maximum aller Fehlerschranken.
// Max_Rat_Xi unterteilt X mittels diam in Teilintervalle Xi, berechnet fr
// alle Xi mit der Funktion FPt() die gesuchten Fehlerschranken und liefert 
// mit bnd das Maximum dieser Fehlerschranken.
// Bedeutung der Funktionsparameter von FPt:
// 1. Param.: Teilintervall der exakten Argumente fr die Fehlerabschtzung;
//            Dieses Teilintervall wird in dieser Funktion als Xi berechnet.
// 2. Param.: Absolute oder rel. Fehlerschranke der gestrten Argumente.
//            Dieser Parameter wird beim Aufruf von FPt durch bndx ersetzt.
// 3. Param.: Intervallvektor, mit den optimalen Koeffztn.-Einschlieungen.
//            Dieser Parameter wird beim Aufruf von FPt durch  ia  ersetzt.
// 4. Param.: Intervallvektor, mit den optimalen Koeffztn.-Einschlieungen.
//            Dieser Parameter wird beim Aufruf von FPt durch  ib  ersetzt.
// 5. Param.: Integer-Gre k bezglich u(x) = x^k. 
//            Dieser Parameter wird beim Aufruf von FPt durch  k  ersetzt. 
// 6. Param.: real-Gre x0 ist Entwicklungspunkt
{
    int ex;
    real r1 = Inf(X), m = mant(r1), SX = Sup(X), d = 0, r2;
    bnd = 0;
    if (sign(m)==0) r2 = minreal;
    else
    {
        ex = expo(r1);
        r2 = m + diam;
        times2pown(r2,ex);
	r2 = succ(r2);
    }
    if (r2>SX) r2 = SX;
    interval Xi = interval(r1,r2); // Xi: first subinterval
    do 
    {
	d = FPt(Xi,bndx,ia,ib,k,x0);  // d: Fehlerschr. ber Teilintervall Xi 
	if (d>bnd) bnd = d;  // Maximumbildung
        // New subinterval:
	r1 = r2; // Neue Intervalluntergrenze = alte Intervallobergrenze
        m = mant(r1);
	if (sign(m)==0) r2 = minreal;
        else
	{
            ex = expo(r1);
	    r2 = m + diam;
            times2pown(r2,ex);
	    r2 = succ(r2);
	}
        if (r2>SX) r2 = SX;    // Neue Obergrenze r2 darf X nicht nach 
                               // oben berschreiten!
	Xi = interval(r1,r2);  // Xi: Neues Teilintervall
    } while (r1 < SX);
} // Max_Rat_Xi

static void Max_Term_Xi(real (*FPt)(const interval& Xi), // Funktions-Zeiger
                        const interval& X, // X: Zu unterteilendes Intervall
		                           // der exakten Argumente
                        const real& diam,  // diam: Mantissen-Durchmesser bei
                                           // der Intervallunterteilung
	                real& bnd)   // bnd: Maximum aller Fehlerschranken.
// Max_Term_Xi unterteilt X mittels diam in Teilintervalle Xi, berechnet fr
// alle Xi mit der Funktion FPt() die gesuchten Fehlerschranken und liefert 
// mit bnd das Maximum dieser Fehlerschranken.
// Bedeutung der Funktionsparameter von FPt:
// 1. Param.: Teilintervall der exakten Argumente fr die Fehlerabschtzung;
//            Dieses Teilintervall wird in dieser Funktion als Xi berechnet.
{
    int ex;
    real r1 = Inf(X), m = mant(r1), SX = Sup(X), d = 0, r2;
    bnd = 0;
    if (sign(m)==0) r2 = minreal;
    else
    {
        ex = expo(r1);
        r2 = m + diam;
        times2pown(r2,ex);
	r2 = succ(r2);
    }
    if (r2>SX) r2 = SX;
    interval Xi = interval(r1,r2); // Xi: first subinterval
    do 
    {
	d = FPt(Xi);         // d: Fehlerschranke ber Teilintervall Xi 
	if (d>bnd) bnd = d;  // Maximumbildung
        // New subinterval:
	r1 = r2; // Neue Intervalluntergrenze = alte Intervallobergrenze
        m = mant(r1);
	if (sign(m)==0) r2 = minreal;
        else
	{
            ex = expo(r1);
	    r2 = m + diam;
            times2pown(r2,ex);
	    r2 = succ(r2);
	}
        if (r2>SX) r2 = SX;    // Neue Obergrenze r2 darf X nicht nach 
                               // oben berschreiten!
	Xi = interval(r1,r2);  // Xi: Neues Teilintervall
    } while (r1 < SX);
} // Max_Term_Xi

real Term_relh(real (*FPt)(const interval& Xi),
               const interval& X, const real& diam)
// Liefert mit bnd eine Oberschranke des relativen Fehlers, wenn der
// Term auszuwerten ist. Die Berechnung der relativen Fehler-
// schranke muss mit der Funktion FPt() erfolgen.
// X ist das Intervall der exakten Termargumente x aus X.
// X wird in Teilintervalle Xi zerlegt, um bnd zu minimieren; mit diam wird
// der Durchmesser der X1 gesteuert. diam = 10^(-5) ist i.a. eine gute Wahl.
{
    real bnd;
    Max_Term_Xi(FPt,X,diam,bnd);
    return bnd;
} // Term_relh

void eps2fractions(const string& eps, 
		   real& Z1p, real& N1p, real& Z1m, real& N1m)
// For a given relative error bound eps this function delivers the 
// numerators and the denominators for the terms 1+eps and 1-eps.
// Example:  With eps = "2.97768e-15" we get:
// 1-eps = Z1m / N1m = 9007199254740962.0 / 9007199254740992.0
// 1+eps = Z1p / N1p = 4503599627370513.0 / 4503599627370496.0
{
    N1m = 9007199254740992.0; // denominator of 1-eps
    N1p = 4503599627370496.0; // denominator of 1+eps
    string str = eps;
    real epsup, r;
    str >> RndUp >> epsup;  // epsup = eps, rounded upwards!

    // calculation of:   1 - eps:
    r = subd(1,epsup);
    r = muld(r,N1m);
    Z1m = pred(pred(pred(r)));

    // calculation of:   1 + eps:
    r = addu(1,epsup);
    r = mulu(r,N1p);
    Z1p = succ(succ(succ(r)));
} // eps2fractions

void IEEE2frac_int(const real& x, real& z, real& n)
// For a given IEEE-number x the two whole-numbered IEEE-variables
// z,n are calculated, so that x = z/n;
// Notice, that the machine division z/n will be performed errorfree, 
// because the exact quotient z/n is the IEEE-number x. 
// The programm output of the results z,n should be done with:
// cout << SetPrecision(1,1) << Fixed << "z = " << z << endl;
// cout << "n = " << n << endl;
{
    if (sign(x)==0) {z=0; n=1;}
    else {
	int ex = 54 - expo(x);
	z = mant(x);
	if (ex<=0) { z = x;  n = 1; }
	else { // ex > 0:
	    n = comp(0.5,ex);  // n = 2^ex
	    times2pown(z,53); 
	}
    }
} // IEEE2frac_int

void frac_tools(const int& Z, const int& N,
		interval& x, real& Next, real& del, real& eps)
// Z is the numerator and N is the denominator of the fraction Z/N;
// x is the optimal inclusion of Z/N in the IEEE double format;
// Next is the nearest real number of the fraction Z/N;
// del is an upper bound of the absolute error |Z/N - Next|
// eps is an upper bound of the relative error |Z/N - Next|/|Z/N|
{
    if (N==0)
    {
	cerr << "frac_tools(): Z/N is not a fraction!" << endl;
	exit(1);
    }
    real Ix,Sx,r1,r2,r;
    l_interval iZ=l_interval(Z), iN=l_interval(N), q;
    interval d1,d2;
    stagprec = 2;  del=0;  eps=0;
    do 
    {
	stagprec++;
	q = iZ/iN;
	x = q;
	Ix = Inf(x);  Sx = Sup(x);
    } while(Sx > succ(Ix));
    // x is now a point interval or it holds: succ(Inf(x))==Sup(x)
    if (diam(x)==0) Next = Ix; 
    else
    {
	r = Inf( abs(x) );
	d1 = q-Ix;   r1 = Sup(d1);
	d2 = Sx-q;   r2 = Sup(d2);
	if (r2<r1)
	{
	    Next = Sx;  del = r2;  eps = divu(r2,r);
	}
        else
	{
	    Next = Ix;  del = r1;  eps = divu(r1,r);
	}
    }
} // frac_tools

void Polynomgrad(int& N, int& k)
// Interaktive Eingabe von:
// N: Polynomgrad
// k: Exponent bezglich u=x^k, 1<=k<=6
{
    cout << "********************************************************" << endl;
    cout << "*    Eingabe von Polynomdaten zur Polynomauswertung    *" << endl;
    cout << "********************************************************" << endl;
    cout << endl;
    do
    { 
	cout << "   Polynomgrad N >= 1  ? "; cin >> N;
    } while(N<1);
    cout << endl;
    do
    {
	cout << "   u = x^k, 1<= k <=6  ? "; cin >> k;
    } while(k<1 || k>6);
    cout << endl;
} // Polynomgrad

void Koeff_Rat(intvector& Za, intvector& Ne)
// Der Anwender wird aufgefordert, in die Vektoren Za,Ne die ganzzahligen
// Werte fr Zhler und Nenner der Polynomkoeffizienten einzulesen.
// Vorausgesetzt wird:  a[0]=1;  
// Lb(Za)=Lb(Ne)=1;  Polynomgrad = Ub(Za) = Ub(Ne) >= 1;
{
    int N = Ub(Za);
    for (int j=1; j<=N; j++)
    {
	cout << "Zhler von a[" << j << "] = ? ";  
        cin >> Za[j];
	cout << "Nenner von a[" << j << "] = ? ";  
        cin >> Ne[j];   
        cout << endl;
    }
} // Koeff_Rat

void Poly_Dat_rel(interval& X, real& epsx, real& diam)
// Interaktive Eingabe von:
// X:    Intervall der exakten Polynomargumente
// epsx: Relativer Fehler der gesrten Polynomargumente
// diam: Mit diam wird die Intervallunterteilung gesteuert
{
    string str;
    cout << "   Intervall der exakten Polynom-Argumente [a,b] = ? ";
    cin >> str;
    str >> X;
    cout << endl;
    cout << "   Relative Fehlerschranke der gestrten Argumente ? "; 
    cin >> epsx; 
    if (epsx<0) epsx = -epsx;
    cout << endl;
    cout << "   Mantissenbreite der Teilintervalle (1e-5) ? "; cin >> diam; 
    if (diam<0 || diam > 0.125) diam = 1e-5;
    cout << endl;
} // Poly_Dat_rel

void Term_Dat(interval& X, real& diam)
// Interaktive Eingabe von:
// X   : Intervall der Term-Maschinen-Agumente
// diam: Mit diam wird die Intervallunterteilung gesteuert
{
    string str;
    cout << "   Intervall der Maschinen-Argumente [a,b] = ? "; cin >> str;
    str >> X;
    cout << endl;
    cout << "   Mantissenbreite der Teilintervalle (1e-5) ? "; cin >> diam; 
    if (diam<0 || diam>0.125) diam = 1e-5;
    cout << endl;
} // Term_Dat

real Horn_relh1(const intvector& Za, const intvector& Ne, const interval& X,
                const real& epsx, const int& k, const real& diam)
// Liefert mit bnd eine Oberschranke des relativen Fehlers, wenn das Polynom
// P_N(x) = 1 + a[1]*u^1 +...+a[N]*u^N,  u=x^k, 1<= k <=6,  nach dem Horner-
// schema auszuwerten ist. Die Vektoren Za und Ne, mit Ub(Za)=Ub(Ne)=N, mssen 
// fr jeden Koeffizienten a[j], j=1,..,N, die entsprechenden integer-Zahlen 
// fr Zhler und Nenner enthalten. Die exakten rationalen Koeffizienten
// a[j]=Za[j]/Ne[j] werden mit der Funktion Koeff_tools() zur jeweils nchsten
// Rasterzahl Next[j] aus S(2,53) gerundet, und mit diesen Koeffizienten
// Next[j] ist dann das Polynom nach dem Hornerschema in nur hochgenauer
// Arithmetik auszuwerten.
// X ist das Intervall der exakten Polynomargumente x aus X. Wurden die 
// Polynomargumente fehlerhaft berechnet, so ist epsx die entsprechende 
// relative Fehlerschranke. 
// X wird in Teilintervalle Xi zerlegt, um bnd zu minimieren; mit diam wird
// der Durchmesser der X1 gesteuert. diam = 10^(-5) ist i.a. eine gute Wahl.
{
    real bnd;
    ivector ia(Ub(Za));   // Optimale Einschlieung der Koeffizienten
    rvector Next(Ub(Za)), // Zur nchsten Rasterzahl gerundete Koeffizienten
            del(Ub(Za)),  // Vektor der abs. Koeffizienten-Rundungsfehler
            eps(Ub(Za));  // Vektor der rel. Koeffizienten-Rundungsfehler
    Koeff_tools(Za,Ne,ia,Next,del,eps);   // realisiert ia,Next,del,eps;
    // ia:   Optimale Einschlieung der exakten Polynomkoeffizienten a[j]
    // Next: Next[j] ist die zu a[j] nchstgelegene Rasterzahl, 1<=j<=N.
    // del:  del[j] ist die bei der Rundung nach Next[j] aufgetretene
    //       absolute Fehlerschranke.
    // eps:  eps[j] ist die bei der Rundung nach Next[j] aufgetretene
    //       relative Fehlerschranke. 
    Max_Horn_Xi(Horn_relh1_Xi,X,epsx,diam,ia,del,k,bnd);
    return bnd;
} // Horn_relh1

real Horn_relh(const rvector& a, const real& x0, const interval& X,
               const real& epsx, const int& k, const real& diam)
// Liefert mit bnd eine Oberschranke des relativen Fehlers, wenn das Polynom
// P_N(u) = a[0] + a[1]*u^1 +...+a[N]*u^N,  u=(x-x0)^k, 1<= k <=6,  nach dem
// Hornerschema auszuwerten ist. 
// a ist der Vektor der exakten Polynomkoeffizienten a[j], j=0,1,...,N;
// x0 ist der Polynom-Entwicklungspunkt; x0 in S(2,53) muss erfllt sein!
// X ist das Intervall der exakten Polynomargumente x aus X. Wurden die 
// Polynomargumente fehlerhaft berechnet, so ist epsx die entsprechende 
// relative Fehlerschranke.
// X wird in Teilintervalle Xi zerlegt, um bnd zu minimieren; mit diam wird
// der Durchmesser der X1 gesteuert. diam = 10^(-5) ist i.a. eine gute Wahl.
{
    real bnd;
    ivector ia(0,Ub(a));  // Einschlieung der exakten Polynom-Koeffizienten
                          // durch Punktintervalle
    for (int j=0; j<=Ub(a); j++) ia[j] = interval(a[j]);
    Max_Horn_Xi(Horn_relh_Xi,X,epsx,diam,ia,k,x0,bnd);
    return bnd;
} // Horn_relh

real Rat_relh(const rvector& a, const rvector& b, const real& x0, 
              const interval& X, const real& epsx, const int& k, 
              const real& diam)
// Liefert mit bnd eine Oberschranke des relativen Fehlers, wenn die
// rationale Funktion   R(u)=P_N(u)/Q_M(u), u=(x-x0)^k, 1<=k<=6, mit den
// Polynomen:  P_N(u) = a[0] + a[1]*u^1 +...+a[N]*u^N,  
//             Q_M(u) = b[0] + b[1]*u^1 +...+b[M]*u^M,
// nach dem Hornerschema ausgewertet wird. 
// a ist der Vektoren der exakten Koeffizienten a[j], j=0,1,...,N;
// b ist der Vektoren der exakten Koeffizienten b[j], j=0,1,...,M;
// x0 ist der Polynom-Entwicklungspunkt; x0 in S(2,53) muss erfllt sein!
// X ist das Intervall der exakten Polynomargumente x aus X. Wurden die 
// Polynomargumente fehlerhaft berechnet, so ist epsx die entsprechende 
// relative Fehlerschranke.
// X wird in Teilintervalle Xi zerlegt, um bnd zu minimieren; mit diam wird
// der Durchmesser der X1 gesteuert. diam = 10^(-5) ist i.a. eine gute Wahl.
{
    real bnd;
    ivector ia(0,Ub(a)), ib(0,Ub(b));  // Einschlieungen der exakten 
                       // Polynom-Koeffizienten durch Punktintervalle
    for (int j=0; j<=Ub(a); j++) ia[j] = interval(a[j]);
    for (int j=0; j<=Ub(b); j++) ib[j] = interval(b[j]);
    Max_Rat_Xi(Rat_relh_Xi,X,epsx,diam,ia,ib,k,x0,bnd);
    return bnd;
} // Rat_relh

	
} // namespace cxsc













