// Programm expmx2.cpp zur Auswertung der Funktion
// e^(-x^2) fuer Punktargumente x aus [-MaxReal,+MaxReal]
// und fuer Intervallargumente. 

#include <rmath.hpp>  // Wegen exp(...)
#include <imath.hpp>  // Wegen abs(x)
#include "hilfe.hpp"  // Wegen sqr2uv(...)
#include <iostream>   // Wegen cout ...

using namespace std;
using namespace cxsc;

// real-Konstante  exp_x2_x0  fuer die Funktion e^(-x^2);
// Diese Konstante wird auch fuer die Intervallfunktion benoetigt!
const real exp_x2_x0 = 7491658466053896.0 / 281474976710656.0;
// Fuer x > exp_x2_x0 werden die Funktionswerte auf Null gesetzt.
// Die relative Fehlerschranke e(f) := 4.618919E-16 gilt fuer
// alle |x| <= exp_x2_x0 = 26.61571750925....

real Expmx2(const real& x) throw()
// e^(-x^2);  rel. Fehlerschranke:  eps = 4.618919E-16 = e(f) gilt
// fuer alle |x| <= exp_x2_x0 = 26.61571750925....
// exp_x2_x0 = 7491658466053896.0 / 281474976710656.0;
// Fuer |x| > exp_x2_x0 --> expmx2(x) = 0;
// Ausfuehrlich getestet;  Blomquist, 05.07.04;
{
    real t=x,u,v,res=0;
    int ex;
    if (t<0) t = -t;  // t >= 0;
    ex = expo(t);
    if (ex<=-26) res = 1;  // t < 2^(-26)
    else if (ex<=-6)       // t < 2^(-6)
    {
	u = t*t;  v = u;
	times2pown(v,-1);  // v: 0.5*x^2
	res = 1-u*( 1-v*(1-u/3) );
    } else if (t<=exp_x2_x0) {
	sqr2uv(x,u,v);  // u:= x*x,v aus S(2,53); x^2 = u+v (exakt!)
	res = exp(-u);
	if (v!=0) {
	    times2pown(res,500);  // Die Skalierung verhindert, dass
	    v *= res; // v*exp(-u) in den denormal. Bereich faellt.
	    res -= v;
	    times2pown(res,-500); // Rueckskalierung
	}
    }
    return res;
} // Expmx2

// Konstanten fuer die Intervallfunktion exp(-x^2):
real q_exp_x2p = 4503599627370502.0 / 4503599627370496.0;  // (1+e(f))
real q_exp_x2m = 9007199254740984.0 / 9007199254740992.0;  // (1-e(f))
// Oberschranke exp_x2_UB fuer |x| >exp_x2_x0:
real exp_x2_UB = 2.225073858507447856659E-308;

interval Expmx2(const interval& x)
// e^(-x^2);  Blomquist, 05.07.04;
{
    real y,r1,r2,Sz,Iz;
    interval z = abs(x);
    // Berechnung einer Unterschranke:
    Sz = Sup(z);  Iz = Inf(z);
    y = Expmx2(Sz);
    r1 = y;
    if (Sz<=exp_x2_x0) r1 = r1 * q_exp_x2m; // Abrunden
    if (Sz==0) r1 = y;
    // Berechnung einer Oberschranke:
    if (Iz>exp_x2_x0) r2 = exp_x2_UB;
    else r2 = (Sz==Iz)? y * q_exp_x2p : Expmx2(Iz) * q_exp_x2p;
    if (r2>1) r2 = 1.0;
    return interval(r1,r2);

} // Expmx2 (Intervallargumente)

int main() {
    real x,y;
    interval z,yz;

    while (1) {
	cout << "real x = ?" << endl;       cin >> RndNext >> x;
	y = Expmx2(x);
        cout << "Expmx2(x) = " << Scientific << SetPrecision(15,15)
	     << y << endl;

	cout << "interval z = ?" << endl;   cin >> z;
	yz = Expmx2(z);
        cout << "Expmx2(z) = " << yz << endl << endl;
    };
}







