/*
**  CXSC is a C++ library for eXtended Scientific Computing
**
**  Copyright (C) 1990-2000 Institut fuer Angewandte Mathematik,
**                          Universitaet Karlsruhe, Germany
**            (C) 2000-2001 Wiss. Rechnen/Softwaretechnologie
**                          Universitaet Wuppertal, Germany   
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Library General Public
**  License as published by the Free Software Foundation; either
**  version 2 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
**  Library General Public License for more details.
**
**  You should have received a copy of the GNU Library General Public
**  License along with this library; if not, write to the Free
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "hilfe.hpp"
#include <real.hpp>

namespace cxsc {

int uint_trail(const unsigned int& n)
//  Liefert die Anzahl aufeinanderfolgender Null-Bits am Ende von n;
//  n=0 --> 32;   n=1 --> 0;   n=2 --> 1;   n=4 --> 2;   n=1024 --> 10;
{
    int m1,p=0;
    unsigned int m=n;
    if (m==0) p = 32;
    else { // m != 0, d.h. die folgende while-Schleife bricht ab!
       do
       {
	   m1 = m & 1;  // m1=0(1), wenn letztes Bit von m gleich 0(1) ist.
	   if (m1==0) p++;
	   m = m >> 1;  // Bit-Muster um 1 nach rechts schieben.
       } while(m1==0);
    }
    return p;
} // uint_trail

typedef union { double f; char intern[7]; } help_real;

real Cut24(const real& x)
// y = Cut24(x) liefert ein y, das mit den ersten 24 Mantissenbits
// von x bereinstimmt, das hidden bit ist dabei mitgezhlt!
// Die restlichen 53-24=29 Mantissenbits werden auf Null gesetzt.
{
  help_real y;
  y.f = _double(x);
#if INTEL
  y.intern[3] &= 224;
  y.intern[0] = y.intern[1] = y.intern[2] = 0;
#else
  y.intern[4] &= 224;
  y.intern[7] = y.intern[6] = y.intern[5] = 0;
#endif
  return real(y.f);
}

real Cut25(const real& x)
// y = Cut25(x) liefert ein y, das mit den ersten 25 Mantissenbits
// von x bereinstimmt, das hidden bit ist dabei mitgezhlt!
// Die restlichen 53-25=28 Mantissenbits werden auf Null gesetzt.
{
  help_real y;
  y.f = _double(x);
#if INTEL
  y.intern[3] &= 240;
  y.intern[0] = y.intern[1] = y.intern[2] = 0;
#else
  y.intern[4] &= 240;
  y.intern[7] = y.intern[6] = y.intern[5] = 0;
#endif
  return real(y.f);
}

real Cut26(const real& x)
// y = Cut26(x) liefert ein y, das mit den ersten 26 Mantissenbits
// von x bereinstimmt, das hidden bit ist dabei mitgezhlt!
// Die restlichen 53-26=27 Mantissenbits werden auf Null gesetzt.
{   
  help_real y;
  y.f = _double(x);
#if INTEL
  y.intern[3] &= 248;
  y.intern[0] = y.intern[1] = y.intern[2] = 0;
#else
  y.intern[4] &= 248;
  y.intern[7] = y.intern[6] = y.intern[5] = 0;
#endif
  return real(y.f);
}

void sqr2uv(const real& x, real& u, real& v)
// Liefert u,v fr: x^2 = u + v; EXAKTE Darstellung, falls kein overflow 
// auftritt und v im normalisierten Bereich liegt. u > |v|
// Vorsicht: Funktioniert zunchst nur auf INTEL-Systemen!!!
{
    real a,b,t,y1,y2;
    a = Cut26(x);
    b = x-a;  // x = a+b;
    u = x*x;
    t = u - a*a; // exakte Auswertung!
    y2 = a*b;
    times2pown(y2,1); // y2 = 2*a*b, exakt!
    t -= y2;
// Jetzt fehlt noch:  t2 - b*b, aber b*b wird nicht immer korrekt berechnet,
// daher nochmalige Aufspaltung von b in y1+y2!!
    y1 = Cut25(b);
    y2 = b - y1;   // b = y1+y2, exakt;
    t -= y1*y1;
    if (sign(y2)!=0)
    {
	a = y1*y2;
	times2pown(a,1); // a = 2*y1*y2, exakt!
	t -= a;
	t -= y2*y2;
    }
    v = -t;
} // sqr2uv

} // namespace cxsc







